/*
 * holdings.c --
 *
 *	This file implements functions for keeping track of
 *	tokens held by this client.
 */

static char rcsid[] =
"$Id: holdings.c,v 1.6 2001/08/31 20:51:27 gmt Exp $";

#include "tokInt.h"

/*
 * Private (per-service) data structure:
 */
typedef struct TokHoldingsData {
    Tok_Token bin[TOK_NBINS];		/* hash bins */
    TokMutex binMutex[TOK_NBINS];	/* mutexes for hash bin access */
} *HoldingsData;

/*
 * Prototypes for procedures referenced only in this file:
 */

/*
 * TokInitHoldings(svc) -- initialize data structures.
 */
void
TokInitHoldings(Tok_Service svc) {
    int i;
    HoldingsData hd;
    
    hd = svc->holdingsData = TokAlloc(sizeof(struct TokHoldingsData));
    for (i = 0; i < TOK_NBINS; i++) {
	hd->bin[i] = NULL;
	TokInitMutex(&hd->binMutex[i]);
    }
}
/*
 * TokTermHoldings(svc) -- free storage.
 */
void
TokTermHoldings(Tok_Service svc) {
    int i;
    HoldingsData hd = svc->holdingsData;
    Tok_Token t;

    for (i = 0; i < TOK_NBINS; i++) {
	hd->bin[i] = NULL;
	while ((t = hd->bin[i]) != NULL) {
	    TokDelete(t);
	}
	TokDestroyMutex(&hd->binMutex[i]);
    }
    TokFree(hd);
    svc->holdingsData = NULL;
}

/*
 * TokLookup(svc, name, create) -- find token in registry.
 *
 * If the token does not exist, it is created if create != 0.
 */
Tok_Token
TokLookup(Tok_Service svc, char *name, int create) {
    HoldingsData hd = svc->holdingsData;
    Tok_Token p, *q;
    int h = TokHash(name);
    int i = h % TOK_NBINS;

    TokLockMutex(&hd->binMutex[i]);

    q = &hd->bin[i];
    while ((p = *q) != NULL && (p->hashcode!=h || strcmp(p->name,name)!=0)) {
	q = &p->next;
    }

    if (p == NULL && create) {
	p = TokZalloc(sizeof(struct TokToken));
	*q = p;
	p->service = svc;
	p->name = TokSalloc(name);
	p->hashcode = h;
	TokInitMutex(&p->mutex);
	TokInitCond(&p->cv);
    }

    TokUnlockMutex(&hd->binMutex[i]);
    return p;
}

/*
 *  TokDelete(t) -- remove token from registry, without freeing it.
 */
void
TokDelete(Tok_Token t) {
    Tok_Service svc = t->service;
    HoldingsData hd = svc->holdingsData;
    Tok_Token p, *q;
    int h = t->hashcode;
    int i = h % TOK_NBINS;

    TokLockMutex(&hd->binMutex[i]);
    q = &hd->bin[i];
    while ((p = *q) != NULL && p != t) {
	q = &p->next;
    }
    if (p != NULL) {
	*q = p->next;
    }
    TokUnlockMutex(&hd->binMutex[i]);
}

/*
 * TokSendCatalog(svc, dest) -- send CATALOG message.
 *
 * We need to pickle up a count "n" followed by n tokens; however,
 * we don't know the count in advance.  A two-pass approach would
 * be risky due to difficulties in keeping the two passes consistent;
 * so we gather the tokens into a scratch pickle first.
 */
void TokSendCatalog(Tok_Service svc, int dest) {
    int i, n;
    Tok_Token t;
    TokPickle p;
    HoldingsData hd = svc->holdingsData;

    p = TokEmptyPickle();
    n = 0;
    for (i = 0; i < TOK_NBINS; i++) {
	TokLockMutex(&hd->binMutex[i]);
	t = hd->bin[i];
	while (t != NULL) {
	    TokLockMutex(&t->mutex);
	    if (TokServer(svc, t->name) == dest
	    && t->state >= TOK_RECEIVED && t->state != TOK_RELEASED) {
		n++;
		TokPickleString(p, t->name);
		TokPickleBytes(p, t->data, t->datalen);
	    }
	    TokUnlockMutex(&t->mutex);
	    t = t->next;
	}
	TokUnlockMutex(&hd->binMutex[i]);
    }

    TokSendVals(svc, TSVC_CATALOG, dest, "ip", n, p);
    TokFreePickle(p);
}
