/*
 * pickle.c --
 *
 *	This file implements functions for encoding and decoding values.
 */

static char rcsid[] =
"$Id: pickle.c,v 1.5 2001/07/03 22:08:29 gmt Exp $";

#include "tokInt.h"

/*
 * The Pickle structure:
 */
struct TokPickle {
    unsigned char *buffer;
    int allo;		/* number of characters allocated */
    int size;		/* number of characters stored */
    int next;		/* index of next read */
};

/*
 * Manifest constants:
 */
#define DEFAULT_SIZE 120	/* default size for a new pickle */
#define INCREMENT_SIZE 20	/* increment for growing a pickle */

/*
 * Prototypes for procedures referenced only in this file:
 */
static void Pneed(TokPickle p, int n);
static void PutByte(TokPickle p, int b);
static unsigned char GetByte(TokPickle p);

/*
 * TokEmptyPickle() -- create new pickle for encoding values.
 */
TokPickle
TokEmptyPickle() {
    TokPickle p = TokZalloc(sizeof(struct TokPickle));
    p->buffer = TokAlloc(DEFAULT_SIZE);
    p->allo = DEFAULT_SIZE;
    return p;
}

/*
 * TokFullPickle(buf, len) -- initialize read-only pickle by copying buffer.
 */
TokPickle
TokFullPickle(void *buf, int len) {
    TokPickle p = TokZalloc(sizeof(struct TokPickle));
    p->buffer = TokAlloc(len);
    p->allo = len;
    p->size = len;
    memcpy(p->buffer, buf, len);
    return p;
}

/*
 * TokFreePickle(p) -- release space allocated for a pickle.
 */
void
TokFreePickle(TokPickle p) {
    TokFree(p->buffer);
    TokFree(p);
}

/*
 * TokPickleSize(p) -- return number of bytes stored in pickle.
 */
int
TokPickleSize(TokPickle p) {
    return p->size;
}

/*
 * TokPickleData(p) -- return pointer to data stored in pickle.
 */
void *
TokPickleData(TokPickle p) {
    return p->buffer;
}

/*
 * TokResetPickle(p) -- reinitialize to reread all stored data.
 */
void
TokResetPickle(TokPickle p) {
    p->next = 0;
}

/*
 * TokPickleInt(p, value) -- add integral value to pickle.
 *
 *  0snn nnnn			7-bit signed value (in one byte)
 *  1bbb snnn  nnnn nnnn  ...	12 + 8 * bbb bits (in (bbb + 2) bytes)
 */
void
TokPickleInt(TokPickle p, long v) {
    int b;
    long x;

    if (-0x40 <= v && v < 0x40) {
	PutByte(p, v & 0x7F);
    } else {
	b = 0;
	for (x = v >> 11; x != 0 && x != -1; x >>= 8)
	    b++;
	if (b == (sizeof(long) - 1))
	    PutByte(p, 0x80 + (b << 4) + ((v >> (8 * b + 7)) & 0xF));
	else
	    PutByte(p, 0x80 + (b << 4) + ((v >> (8 * (b + 1))) & 0xF));
	while (b >= 0)
	    PutByte(p, v >> (8 * b--));
    }
}

/*
 * TokUnpickleInt(p) -- read integral value from a pickle.
 */
long
TokUnpickleInt(TokPickle p) {
    long v;
    int b, n;

    b = GetByte(p);			/* get first byte */
    if (b < 0x40)
	return b;			/* 6-bit positive value */
    if (b < 0x80)
	return b - 0x80;		/* 6-bit negative value */

    n = (b >> 4) & 0x7;			/* get bytecount-1 */
    v = (- (b & 8)) | (b & 7);		/* get 3 data bits and extended sign */

    while (n-- >= 0)			/* while more bits */
	v = (v << 8) | GetByte(p);	/* squeeze them in */

    return v;
}

/*
 * TokPickleString(p, s) -- add string to pickle.
 */
void
TokPickleString(TokPickle p, char *s) {
    TokPickleBytes(p, s, strlen(s));
}

/*
 * TokUnpickleString(p) -- read from pickle, returning allocated string.
 */
char *
TokUnpickleString(TokPickle p) {
    int n = TokUnpickleInt(p);
    char *s = TokAlloc(n + 1);
    memcpy(s, p->buffer + p->next, n);
    p->next += n;
    s[n] = '\0';
    return s;
}

/*
 * TokPickleBytes(p, buf, n) -- add n-byte char array to pickle p.
 */

void
TokPickleBytes(TokPickle p, void *buf, int n) {
    TokPickleInt(p, n);
    TokAppendBytes(p, buf, n);
}

/*
 * TokUnpickleBytes(p, buf, n) -- read byte array, maxlen n, returning len.
 */
int
TokUnpickleBytes(TokPickle p, void *buf, int n) {
    int len = TokUnpickleInt(p);
    memcpy(buf, p->buffer + p->next, (len <= n) ? len : n);
    p->next += len;
    return len;
}

/*
 * TokAppendBytes(p, buf, n) -- append n raw bytes to pickle.
 */
void
TokAppendBytes(TokPickle p, void *buf, int n) {
    if (n != 0) {
	Pneed(p, n);
	memcpy(p->buffer + p->size, buf, n);
	p->size += n;
    }
}

/*
 * TokExtractBytes(p, buf, n) -- extract up to n raw bytes, returning n.
 */
int
TokExtractBytes(TokPickle p, void *buf, int n) {
    int max = p->size - p->next;
    if (n > max) {
	n = max;
    }
    memcpy(buf, p->buffer + p->next, n);
    p->next += n;
    return n;
}

/*
 * Pneed(p, n) -- preallocate room to add n bytes to pickle p.
 */
static void
Pneed(TokPickle p, int n) {
    if (p->size + n <= p->allo)
	return;
    p->allo = p->size + n + INCREMENT_SIZE;
    p->buffer = TokRealloc(p->buffer, p->allo);
}

/*
 * PutByte(p, b) -- write byte b into pickle p.
 */
static void
PutByte(TokPickle p, int b) {
    if (p->size >= p->allo)
	Pneed(p, 1);
    p->buffer[p->size++] = b;
}

/*
 * GetByte(p) -- read next byte from pickle p
 */
static unsigned char
GetByte(TokPickle p) {
    if (p->next >= p->size)
	TokAbort("reading beyond end of pickle");
    return p->buffer[p->next++];
}
