/*
 * reissue.c --
 *
 *	This file implements functions for reissuing messages
 *	until they have been acknowledged.
 */

static char rcsid[] =
"$Id: reissue.c,v 1.5 2003/10/15 23:55:11 gmt Exp $";

#include "tokInt.h"

/*
 * This structure represents an unacknowledged message when its
 * "pickle" field is non-null.
 */
typedef struct MsgRef {
    Tok_Token token;		/* associated token */
    TokPickle pickle;		/* pickled message, if not yet acknowledged */
    struct MsgRef *next;	/* next entry in hash chain */
    long seqnum;		/* message sequence number */
    int timer;			/* time until reissue, in seconds */
} *MsgRef;

/*
 * Private (per-service) data structure:
 */
typedef struct TokReissueData {
    int active;				/* are we still running? */
    int checking;			/* is checker running? */
    MsgRef bin[ACK_NBINS];		/* hash bins (based on seq number) */
    TokMutex binMutex[ACK_NBINS];	/* mutexes for hash bin access */
} *ReissueData;

/*
 * Prototypes for procedures referenced only in this file:
 */
static void *Checker(void *arg);
static void Shutdown(Tok_Service svc);

/*
 * TokInitReissue(svc) -- initialize data structures for reissuing messages.
 */
void
TokInitReissue(Tok_Service svc) {
    int i;
    ReissueData rd;

    rd = svc->reissueData = TokAlloc(sizeof(struct TokReissueData));
    for (i = 0; i < ACK_NBINS; i++) {
	rd->bin[i] = NULL;
	TokInitMutex(&rd->binMutex[i]);
    }
    rd->active = TRUE;
    rd->checking = TRUE;
    TokSpawn(Checker, svc);
}

/*
 * TokTermReissue(svc) -- terminate reissue service, freeing storage.
 */
void
TokTermReissue(Tok_Service svc) {
    ReissueData rd = svc->reissueData;

    rd->active = FALSE;			/* signal Checker to clean up */

    while (rd->checking) {
	sleep(1);			/* wait for Checker to exit */
    }

    Shutdown(svc);			/* finish shutdown actions */
}

/*
 * TokPersist(token, p) -- remember message, reissue until ack'd.
 *
 * When the message is acknowledged, the pickle is freed.
 */
void
TokPersist(Tok_Token t, TokPickle p) {
    int i;
    MsgRef m;
    Tok_Service svc = t->service;
    long seqnum = t->msgnum;
    ReissueData rd = svc->reissueData;

    i = seqnum % ACK_NBINS;
    TokLockMutex(&rd->binMutex[i]);
    /*
     * Search for an empty slot, indicated by m->pickle == null.
     */
    for (m = rd->bin[i]; m != NULL; m = m->next) {
	if (m->pickle == NULL) {
	    break;
	}
    }
    if (m == NULL) {
	/*
	 * No empty slot found, so allocate one and link it in.
	 */
	m = TokAlloc(sizeof(struct MsgRef));
	m->next = rd->bin[i];
	rd->bin[i] = m;
    }
    /*
     * Set up the slot to remember this pickled message.
     */
    m->token = t;
    m->pickle = p;
    m->seqnum = seqnum;
    m->timer = RECHECK_INTERVAL + 1;
    TokUnlockMutex(&rd->binMutex[i]);
}

/*
 * TokForget(svc, seqnum) -- forget a sent message.
 *
 * Forgets and frees a message that was awaiting acknowledgment.
 * It is not an error to forget an already forgotten message.
 *
 * The associated token is returned, unless none can be found.
 *
 * The hash bin slot is saved for reuse by another later request.
 */
Tok_Token
TokForget(Tok_Service svc, long seqnum) {
    ReissueData rd = svc->reissueData;
    int i;
    MsgRef m;
    TokPickle p;
    Tok_Token t;

    t = NULL;
    i = seqnum % ACK_NBINS;
    TokLockMutex(&rd->binMutex[i]);
    for (m = rd->bin[i]; m != NULL; m = m->next) {
	p = m->pickle;
	if (p != NULL && m->seqnum == seqnum) {
	    t = m->token;
	    TokFreePickle(p);		/* deallocate pickled message */
	    m->pickle = NULL;		/* indicate that slot is free again */
	}
    }
    TokUnlockMutex(&rd->binMutex[i]);
    return t;
}

/*
 * Checker(svc) -- wakes up periodically to reissue unack'd messages.
 */
static void *
Checker(void *arg) {
    Tok_Service svc = arg;
    ReissueData rd = svc->reissueData;
    int i;
    MsgRef m;
    TokPickle p;

    while (rd->active) {			/* run until terminated */
	for (i = 0; i < ACK_NBINS; i++) {
	    TokLockMutex(&rd->binMutex[i]);
	    for (m = rd->bin[i]; m != NULL; m = m->next) {
		p = m->pickle;
		if (p != NULL) {
		    /*
		     * This is an active entry for an unacknowledged message.
		     * Decrement the timer.  If expired, resend message and
		     * reset the timer.
		     */
		    m->timer -= RECHECK_INTERVAL;
		    if (m->timer <= 0) {
			TokSend(svc->net, TokServer(svc, Tok_GetName(m->token)),
				TokPickleData(p), TokPickleSize(p));
			m->timer = RESEND_INTERVAL;
		    }
		}
	    }
	    TokUnlockMutex(&rd->binMutex[i]);
	}
	sleep(RECHECK_INTERVAL);		/* check once a second */
    }

    rd->checking = FALSE;			/* acknowledge exit signal */
    return NULL;
}

/*
 * Shutdown(svc) -- free memory used by reissue service.
 */
static void
Shutdown(Tok_Service svc) {
    int i;
    MsgRef m, mnext;
    ReissueData rd = svc->reissueData;

    for (i = 0; i < ACK_NBINS; i++) {
	TokDestroyMutex(&rd->binMutex[i]);
	mnext = rd->bin[i];
	rd->bin[i] = NULL;
	while ((m = mnext) != NULL) {
	    if (m->pickle != NULL)
		TokFreePickle(m->pickle);
	    mnext = m->next;
	    TokFree(m);
	}
    }
    TokFree(rd);
    svc->reissueData = NULL;
}
