/*  control.c -- user interaction and control panel window  */

#ifdef WIN32
  #include <windows.h>
#endif

#include "topo.h"

static void ctlReshapeCB(int w, int h);
static void ctlDisplayCB(void);
static void ctlMouseCB(int button0, int state0, int x0, int y0);
static void mmove(double step, double turn);
static void emitString(int x, int y, char *s);
static void clearTiles(void);



/*  control panel layout  */
#define XLAB		10	/* x position for label */
#define XVAL		100	/* x position for value */
#define LSP		18	/* line spacing (leading) */
#define DSC		4	/* descent below baseline */
#define FONT		GLUT_BITMAP_HELVETICA_12

/*  control panel fields  */
#define CONTROLLABELLENGTH 20

struct {
  unsigned char cl;	/* keyboard character on left mouse click */
  unsigned char cr;	/* keyboard character on right mouse click */
  int variable;		/* is this field variable? */
  int refresh;		/* which windows to refresh when changed */
  int x, y;		/* location in window */
  char label[CONTROLLABELLENGTH+1];	/* current label value */
} controls[] = {
  /*keyL keyR act  refresh     xposn  yposn    string*/
  { 'c', 'C',  0, RC+RO+RP+NR, XLAB, 1 * LSP, "[C]olor by" },
  { 'c', 'C',  1, RC+RO+RP+NR, XVAL, 1 * LSP, "" },
  { '>', '<',  0, RC+RO+RP+NS, XLAB, 2 * LSP, "Error[>0<]" },
  { '>', '<',  1, RC+RO+RP+NS, XVAL, 2 * LSP, "" },
  { '0', '0',  1, RC+RO+RP+NS, XVAL, 2 * LSP, "" },
  { '+', '-',  0, RC+RO+RP+NS, XLAB, 3 * LSP, "Distance[+-]" },
  { '+', '-',  1, RC+RO+RP+NS, XVAL, 3 * LSP, "" },
  { 'e', 'E',  0, RC+RO+RP   , XLAB, 5 * LSP, "[E]ye[zZ]" },
  { 'z', 'Z',  1, RC+RO+RP   , XVAL, 5 * LSP, "" },
  { '2', '@',  0, RC+RO      , XLAB, 6 * LSP, "[2]d Mesh" },
  { '2', '@',  1, RC+RO      , XVAL, 6 * LSP, "" },
  { '3', '#',  0, RC   +RP   , XLAB, 7 * LSP, "[3]d Mesh" },
  { '3', '#',  1, RC   +RP   , XVAL, 7 * LSP, "" },
  { 'v', 'V',  0, RC+RO+RP+NS, XLAB, 8 * LSP, "[V]ert.Exag." },
  { 'v', 'V',  1, RC+RO+RP+NS, XVAL, 8 * LSP, "" },
  { 's', 'S',  0, RC   +RP   , XLAB, 9 * LSP, "[S]mooth" },
  { 's', 'S',  1, RC   +RP   , XVAL, 9 * LSP, "" },
  { 'b', 'B',  0, RC+RO+RP   , XLAB,10 * LSP, "[B]ackface Cull" },
  { 'b', 'B',  1, RC+RO+RP   , XVAL,10 * LSP, "" },
  { 'f', 'F',  0, RC+RO+RP   , XLAB,11 * LSP, "[F]OV Cull" },
  { 'f', 'F',  1, RC+RO+RP   , XVAL,11 * LSP, "" },
  { 'p', 'P',  0,           0, XLAB,12 * LSP, "[^P]rint [^p]s" },
  { 'p', 'P',  0,           0, XLAB,13 * LSP, "[P]lay flight" },
  { 'r', 'R',  0, RC         , XLAB,14 * LSP, "[R]ecord flight" },
  { 'r', 'R',  1, RC         , XVAL,14 * LSP, "" },
  { 'q', 'Q',  0,           0, XLAB,15 * LSP, "[Q]uit" }
};

#define NUMBER_OF_CONTROLS (sizeof(controls) / sizeof(controls[0]))

/*  control panel size  */
#define CONTROL_WIDTH	160
#define CONTROL_HEIGHT	8 + 15 * LSP



/*  openCtlWin() -- open control panel window  */

void openCtlWin(void)
{
  glutInitDisplayMode(GLUT_RGB | GLUT_SINGLE);
  glutInitWindowSize(CONTROL_WIDTH, CONTROL_HEIGHT);
  ctlWin = glutCreateWindow("control");
  glClearColor((GLclampf)1,(GLclampf)1,(GLclampf)1,(GLclampf)1);

  glutMouseFunc(ctlMouseCB);
  glutDisplayFunc(ctlDisplayCB);
  glutReshapeFunc(ctlReshapeCB);
  glutKeyboardFunc(keyCB);
  glutSpecialFunc(specialCB);
}



/*  ctlReshapeCB(w, h) -- handle resize of control window  */

static void ctlReshapeCB(int w, int h)
{
  glViewport((GLint)0,(GLint)0,(GLint)w,(GLint)h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D((GLdouble)0,(GLdouble)w,(GLdouble)0,(GLdouble)h);
  glScaled((GLdouble)1,(GLdouble)-1,(GLdouble)1);
  glTranslated((GLdouble)0,(GLdouble)-h,(GLdouble)0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}



/*  ctlDisplayCb() -- draw control window  */

static void ctlDisplayCB(void)
{
  int i;
  char *s;

  glClear(GL_COLOR_BUFFER_BIT);

  for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
    if (controls[i].x == XVAL) {
      s = controls[i].label;

      switch(controls[i].cl) {

      case '+':
	sprintf(s, "%4.1fm", distThresh);  break;

      case '>':
	sprintf(s, "%4.2fm/km", errThresh*100.);  break;

      case 'b':
	strcpy(s, cullBackface ? "On" : "Off");  break;

      case 'c':
	switch(colorLayer) {
	case T_CONTOURS:strcpy(s, "Contours");  break;
	case T_GRAY:	strcpy(s, "Relief");    break;
	default:
	  strncpy(s, layerName[colorLayer],CONTROLLABELLENGTH);
	  break;
	}
	break;

      case 'z':
	if( eyeIsRelative ) {
	  sprintf(s, "+%d", eyeZRelative / 10);
	} else {
	  sprintf(s, "%d", eyeZAbsolute / 10);
	}
	break;

      case 'f':
	strcpy(s, cullFOV ? "On" : "Off");  break;

      case '2':
	switch(twoDMesh) {
	case NONE2D:  strcpy(s, "Off");		break;
	case WIRE2D:  strcpy(s, "On");		break;
	case FOV2D:   strcpy(s, "Fill FOV");	break;
	}
	break;

      case '3':  strcpy(s, threeDMesh ? "On" : "Off");  break;
      case 'r':  strcpy(s, flightRecord ? "Recording" : "Off"); break;
      case 's':  strcpy(s, smoothSurface ? "On" : "Off");  break;

      case 'v':
	sprintf(s, "%3.1f", vexagger);  break;
      }
    }
  }

  for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
    if (controls[i].variable) {
      glColor3f((GLfloat)1,(GLfloat)0,(GLfloat)0);
    } else {
      glColor3f((GLfloat)0,(GLfloat)0,(GLfloat)0);
    }

    emitString(controls[i].x, controls[i].y, controls[i].label);

#ifdef SHOWCONTROLRECTANGLE
    glColor3f((GLfloat)0,(GLfloat)0,(GLfloat)0);
    glRecti((GLint)2,(GLint)(controls[i].y + DSC),
	    (GLint)(CONTROL_WIDTH - 3),(GLint)(controls[i].y + DSC));
#endif

  }
  glFlush();
}



/*  emitString(x, y, s) -- display a string  */

static void emitString(int x, int y, char *s)
{
  int len, i;

  glRasterPos2i((GLint)x,(GLint)y);
  len = strlen(s);
  for(i = 0; i < len; ++i) {
    glutBitmapCharacter(FONT, s[i]);
  }
}



/*  ctlMouseCB(buttn, state, x, y) -- handle mouse click in control window  */

static void ctlMouseCB(int button0, int state0, int x0, int y0)
{
  int i;

  if (state0 != GLUT_UP)
    return;

  for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
    if (y0 < controls[i].y + DSC && y0 > controls[i].y + DSC - LSP) {
      if (button0 == GLUT_LEFT_BUTTON)
	keyCB(controls[i].cl, x0, y0);
      else if (button0 == GLUT_RIGHT_BUTTON)
	keyCB(controls[i].cr, x0, y0);
      break;
    }
  }
}




#ifdef TESTING

/*  doloop() -- circumnavigate the overhead window, and output timings  */

static void doloop(void)
{
  int x, y, xmax, ymax;
  double starttime,endtime;

  glutSetWindow(ovhWin);
  xmax = glutGet(GLUT_WINDOW_WIDTH) - BORDER - 1;
  ymax = glutGet(GLUT_WINDOW_HEIGHT) - BORDER - 1;
  fprintf(stderr,"Walk dimensions = %d x %d\n",xmax-BORDER, ymax-BORDER);

  doingLoop = 1;
  triPerLoop = 0;

  starttime = getTime();

  mouseto(GLUT_LEFT_BUTTON, x = BORDER, y = ymax);
  while (x < xmax)
    mouseto(GLUT_LEFT_BUTTON, ++x, y);
  while (y > BORDER)
    mouseto(GLUT_LEFT_BUTTON, x, --y);
  while (x > BORDER)
    mouseto(GLUT_LEFT_BUTTON, --x, y);
  while (y < ymax)
    mouseto(GLUT_LEFT_BUTTON, x, ++y);

  endtime = getTime();

  fprintf(stderr,"Loop %s redraw in %f secs (%f tri/frame)\n",
	  (loopRefresh & RP) ? "with":"without",endtime-starttime,
	  (float)triPerLoop/(float)(2*(xmax+ymax-2*BORDER)));
  doingLoop = 0;
}

#endif



/*  keyCB(key, x, y) -- handle keypress in any window  */
 
void keyCB(unsigned char key, int x, int y)
     /*ARGSUSED*/
{
  static double stashedET = 0.0;
  int i, m;
  int sign = 1;
  int refresh = -1;	/* The default refresh forces all views to refresh. */

  char s[200];

  m = glutGetModifiers();

  switch(key) {

#ifdef TESTING
  case 't':
    keyCB('l',0,0);
    keyCB('L',0,0);
    for( i=0; i<30; ++i ) {
      keyCB('>',0,0);
      keyCB('>',0,0);
      keyCB('l',0,0);
      keyCB('L',0,0);
    }
    break;
  case 'L':
    loopRefresh = NS;
    doloop();
    break;
  case 'l':
    loopRefresh = RO+RP+NS;
    doloop();
    break;
#endif

  case '<':
    sign = -sign;
    /*NOBREAK*/
  case '>':
    errThresh  += sign * errThreshInc;
    break;

  case '-':
    sign = -sign;
    /*NOBREAK*/
  case '+':
    distThresh += sign * distThreshInc;
    break;

  case 'B':
  case 'b':
    if( cullBackface ) {
      cullBackface = 0;
      glDisable(GL_CULL_FACE);
    } else {
      cullBackface = 1;
      glEnable(GL_CULL_FACE);
    }
    break;

  case 'F':
  case 'f':
    cullFOV = !cullFOV;
    break;

  case 'c':
    sign = -sign;
    /* NOBREAK */
  case 'C':
    colorLayer += sign;
    if (colorLayer < T_GRAY) {
      colorLayer = nLayers-1;
    } else if (colorLayer >= nLayers) {
      colorLayer = T_GRAY;
    }
    break;

  case '2':
    switch(twoDMesh) {
    case NONE2D:  twoDMesh = WIRE2D;	break;
    case WIRE2D:  twoDMesh = FOV2D;	break;
    case FOV2D:   twoDMesh = NONE2D;	break;
    }
    break;

  case '@':
    switch(twoDMesh) {
    case WIRE2D:  twoDMesh = NONE2D;	break;
    case FOV2D:   twoDMesh = WIRE2D;	break;
    case NONE2D:  twoDMesh = FOV2D;	break;
    }
    break;

  case '0':
    if (stashedET == 0.0) {
      stashedET = errThresh;
      errThresh = 0.0;
    } else {
      errThresh = stashedET;
      stashedET = 0.0;
    }
    break;

  case 'E':
  case 'e':
    eyeIsRelative = !eyeIsRelative;
    setEyeRefHeight();
    break;

  case '#':
  case '3':  threeDMesh = !threeDMesh;		break;

  case 'V':
    sign = -sign;
    /* NOBREAK */
  case 'v':
    vexagger += sign * vexaggerInc;
    break;

  case '\020':	/* control-P */
    if( m & GLUT_ACTIVE_SHIFT ) {
      sprintf(s,"%s.stl",initName);
      printSurface(s);
    } else {
#ifdef DEBUG
      outputEPS(1,NULL);
#else
      sprintf(s,"%s.ps",initName);
      outputEPS(1,s);
#endif
    }
    break;

  case 'P':
  case 'p':
    if( flightStep != 0 ) {
      flightStep = 1;
    } else {
      flightStep = 1;
      followPathCB(0);
    }
    break;

  case 'Q':
  case 'q':  exit(0);				break;

  case 's':
  case 'S':
    smoothSurface = !smoothSurface;
    break;

  case 'Z':
    sign = -sign;
    /* NOBREAK */
  case 'z':
    if( eyeIsRelative ) {
      eyeZRelative += sign * eyeZRelativeInc;
    } else {
      eyeZAbsolute += sign * eyeZAbsoluteInc;
    }
    setEyeRefHeight();
    break;

  case 'r':
  case 'R':
    /* Start/Stop recording flightpath */
    if( flightRecord ) {
      fclose(flightRecord);
      flightRecord = NULL;
    } else {
      sprintf(s,"%s.%d.fp",initName, flightNumber++);
      flightRecord = fopen(s, "w");
      if( flightRecord ) {
	fprintf(flightRecord, "# flightpath\n");
      }
    }
    break;

  /* The following cases do not have refresh flags in the control table. */

  case ' ':			/* SP: move forward */
    mmove(tileSizeM / 5.0, 0.0);
    refresh = RO+RP+NS;
    break;

  case '\b':			/* BS: move backward */
    mmove(-tileSizeM / 5.0, 0.0);
    refresh = RO+RP+NS;
    break;

  case '\n':			/* ENTER: move none (forces tile reload) */
  case '\r':
    mmove(0.0, 0.0);
    break;

  case 033:			/* ESC: halt flight */
    if( flightStep > 0 ) flightStep = -1;
    refresh = 0;
    break;

  case 'D':			/* capital-D: discard tile data */
    clearTiles();
    break;

  default:  return;		/* unrecognized; no further processing */
  }

  /* search control panel table for refresh flags */
  for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
    if (controls[i].cl == key || controls[i].cr == key) {
      refresh = controls[i].refresh;
      break;
    }
  }

  postRedisplays(refresh);
}



/*  specialCB(key, x, y) -- handle non-ASCII keypress  */

void specialCB(int key, int x, int y)
     /*ARGSUSED*/
{
  switch (key) {

  /* arrow keys move or swivel the viewer */
  case GLUT_KEY_UP:		mmove(tileSizeM / 80.0, 0.0);	break;
  case GLUT_KEY_DOWN:		mmove(-tileSizeM / 80.0, 0.0);	break;
  case GLUT_KEY_LEFT:		mmove(0.0, +3.0);	break;
  case GLUT_KEY_RIGHT:		mmove(0.0, -3.0);	break;

  case GLUT_KEY_HOME:		/* HOME key moves to original starting tile */
    eye[0] = eyeInit[0];
    eye[1] = eyeInit[1];
    ref[0] = refInit[0];
    ref[1] = refInit[1];
    setEyeRefHeight();
    ovhHome();
    break;

  case GLUT_KEY_END:		/* END key positions to show all tiles loaded */
    eye[0] = xxmin;
    eye[1] = yymin;
    ref[0] = (xxmin + xxmax) / 2;
    ref[1] = (yymin + yymax) / 2;
    setEyeRefHeight();
    ovhEnd();
    break;

  default:			return;
  }
  postRedisplays(RO+RP+NS);
}



/*  mmove(step, turn) -- step forward and swivel gaze  */

/*  if the shift key is down, scale both argument values by 4  */
/*  (call only from within a callback: it calls glutGetModifiers) */

static void mmove(double step, double turn)
{
  double dx, dy, d, a;

  if (glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
    step *= 4.0;
    turn *= 4.0;
  }

  dx = ref[0] - eye[0];
  dy = ref[1] - eye[1];
  d = sqrt(dx * dx + dy * dy);	/* current distance */
  a = atan2(dy, dx);			/* current view angle */

  eye[0] += step * cos(a);		/* new eye location */
  eye[1] += step * sin(a);
  ref[0] += step * cos(a);		/* also translate reference point */
  ref[1] += step * sin(a);

  a += .017453293 * turn;		/* new view angle (in radians) */
  ref[0] = eye[0] + d * cos(a);		/* new reference point */
  ref[1] = eye[1] + d * sin(a);

  setEyeRefHeight();			/* keep eye above ground */
  ovhVis(ref[0], ref[1]);		/* keep overhead window current */
  ovhVis(eye[0], eye[1]);		/* keep overhead window current */
}



/*  clearTiles() -- discard all currently loaded tiles  */

static void clearTiles(void)
{
  int t;
  Wrap *wrap;
  WrapGen *wg;

  fprintf(stderr, "discarding all tiles\n");
  LOCK(wrapLock);
  wg = genWraps(TYPE_ALL); 
  while( (wrap = nextWrap(wg)) != NULL ) {
    t = wrap->type;
    if( t == TYPE_WANTED || t == TYPE_TILE || t == TYPE_UNAV ) {
      freeTile(wrap->data);
      wrap->data = NULL;
      delWrap(wrap);
    }
  }
  UNLOCK(wrapLock);
  postRedisplays(NR + NS + RO + RP);
}
