/*  defns.h --  structures, types, and macros for TopoVista  */

#ifndef DEFNS_H
#define DEFNS_H


/*
 * ============= Configuration =================
 */
#define TESTING 1		/* enable performance testing */


/*
 * ============= Buffer Sizes =================
 */
/* buffer size sufficient for the basic "xxxUyyyc" tile name */
#define SHORT_TILE_NAME 12
/* max tile name with path, also used as max prefix length */
#define MAX_TILE_NAME 1024


/*
 * ============= Display =================
 */
#define HEIGHT_3D  400 /*618*/	/* height of 3-D (perspective) window */
#define WIDTH_3D   600 /*1000*/	/* width */
#define HEIGHT_2D  256		/* height of 2-D (overhead) window */
				/* (width is calculated) */
#define BORDER 15		/* width of (overhead) border in pixels */
#define REF_HEIGHT 0		/* reference ht above terrain (decimeters) */

/* terrain display modes (these must be -2 and -1) */
#define T_CONTOURS -1	/* color-coded elevation contours */
#define T_GRAY -2	/* grayscale relief only */

/* Added to make all elevations positive */
#define ELEVATION_BIAS		1000 /* decimeters */
/* If LOW_TIDE <= elevation <= HIGH_TIDE, color as ocean  */
#define HIGH_TIDE		1010 /* decimeters */
#define LOW_TIDE		 990 /* decimeters */
/* Indicator of missing elevation data */
#define MISSING_DATA		-1000

/* surface generation modes */
#define NONE2D	0	/* flag for no 2D mesh */
#define WIRE2D	1
#define FOV2D	2
#define WIRE3D	3
#define FILL3D	4

/* Types of field of view */
#define WEDGE    0
#define FOUR_GON 4
#define FIVE_GON 5

/* redisplay flags */
#define RC		 1	/* redisplay control window */
#define RO		 2	/* redisplay overhead view */
#define RP		 4	/* redisplay perspective view */
#define NS		 8	/* create a new surface approximation */
#define NR		16	/* make a new relief map */

typedef struct {		/* rgb spec for coloring */
    float red,green,blue;
} Color;



/*
 * ============== TopoVista Tile Format & Protocol ===============
 */
#define TV_PORT 4742	/* default port number */
#define TV_PHLEN 14	/* protocol header length */
#define TV_RHLEN 8	/* record header length */
#define TV_MAXLEN 65535	/* maximum message length */
#define MAX_ISSUE 2	/* maximum number of outstanding requests */
#define TV_EOF  0x3F	/* EOF message sequence number */

#define TV_HEAD   0x10	/* header record flag */
#define TV_ELEV   0x20	/* first elevation record flag */
#define TV_ERRS   0x40	/* first error record flag */
#define TV_NFLAGS 0x80	/* total number of possible flag values */
#define TV_GZIP   0x80	/* additional flag bit indicating gzipped data */

#define TV_TIMEOUT 2	/* simulate EOF if no response after this many sec */

/* record protocol ("hybrid" protocol) ("COIL") ("AGILE") */

#define RP_IDLE_HBT 60	/* heartbeat interval when idle */
#define RP_WAIT_HBT  5	/* heartbeat interval when waiting for data */
#define RP_STALLED 2	/* render partial tile if stalled for this interval */

#define RP_PORT 4747	/* default port number */
#define RP_PHLEN 12	/* protocol header length */
#define RP_MAXLEN 65535	/* maximum message length */

#define RP_MAXASK 100	/* maximum number of records to ask at once */
#define RP_REFRESH 20	/* reissue maxed-out request after this many receives */


/*
 * ============== Multithreading ===============
 *
 * Use only macros for access; treat a Lock as opaque.
 */

#define MAKELIT(s) #s
#define STRINGIFY(s) MAKELIT(s)

#define INITLOCK(k)	initlock(&(k), __FILE__ "." STRINGIFY(__LINE__))
#define LOCK(k)		lock(&(k), __FILE__ "." STRINGIFY(__LINE__))
#define RATTLE(k)	rattle(&(k), __FILE__ "." STRINGIFY(__LINE__))
#define UNLOCK(k)	unlock(&(k), __FILE__ "." STRINGIFY(__LINE__))

typedef struct Lock {		/* recursive mutual exclusion lock */
  #ifdef UNITHREAD
    int mxlock;
  #else
    pthread_mutex_t mxlock;
  #endif
  pthread_t owner;	/* owning thread */
  int       depth;	/* nesting depth */
  char      *where;	/* line where first locked or last unlocked */
} Lock;

typedef struct MsgQueue MsgQueue;	/* opaque message queue struct */


/*
 * ============== Wrap Structure ===============
 *
 * A Wrap struct associates a name, a token, and data.
 */
#define TYPE_WANTED	0x01	/* tile is wanted but have no data */
#define TYPE_TILE	0x02	/* tile has data and can be displayed */
#define TYPE_UNAV 	0x04	/* tile has been found to be unavailable */

#define TYPE_PNM	0x10	/* grid of colors */
#define TYPE_PATH	0x20	/* route along surface, possibly broken */
#define TYPE_FLIGHTPATH	0x40	/* path for automated viewing */

#define TYPE_ALL	~0	/* mask to select all Wrap types */

typedef struct Wrap {
  char *name;		/* tile name (may be token name, file name, or URL) */
  void *token;		/* token pointer, if held */
  inp file;		/* input file, if opened but not yet loaded */
  void *data;		/* actual data, if loaded */
  short type;		/* TYPE_{WANTED,TILE,PATH,...} */
  short stale;		/* set true if token has been revoked */
  struct Wrap *wnext;	/* next Wrap on linked list (for genWraps use only) */
} Wrap;

typedef struct WrapGen WrapGen;


/*
 * ============== Tile Record Structure ===============
 */
typedef struct TileRec {
  unsigned char tr_interest;	/* interest level (0 to MAX_INTEREST) */
  unsigned char tr_unav;	/* set NZ if known to be unavailable */
  unsigned char tr_done;	/* set NZ after received, decoded, discarded */
  unsigned char tr_flags;	/* flags byte from tile record header */
  int tr_length;		/* length of tile record data */
  unsigned char *tr_data;	/* tile record data (malloc'd memory) */
} TileRec;

#define MAX_INTEREST 63		/* maximum interest level */


/*
 * ============== Tile Structure ===============
 */

typedef struct Tile {		/* a single tile */
  Wrap *wrap;		/* containing Wrap struct */
  char tname[SHORT_TILE_NAME];	/* tile name in "xxxUyyyz" form */
  int tid;		/* 32-bit tile ID */
  int nbytes;		/* number of input bytes read */
  short dlevels;	/* number of data levels decoded */
  short elevels;	/* number of error levels decoded */
  int waiting;		/* idle count (some protocols); >0 if awaiting data */
  int rcvmask;		/* bitmap of records received (TVRP protocol) */
  int l, h;		/* minimum and maximum elevations in tile (dm) */
  float scaleZ;		/* vertical scaling applied to original input */
  unsigned short *err;	/* error of triangles */
  unsigned char *info;	/* surface,neighbor,inFOV of triangles */
  unsigned short *data;	/* elevation data of tile (decimeters above -1000dm) */
  float *normal;	/* array of normal triples */
  int xx, yy;		/* UTM coordinates of SW corner */
  int lx1,ly1,lx2,ly2,lx3,ly3;	/* LCHILD(0) corners (0..rows-1, 0..cols-1) */
  int rx1,ry1,rx2,ry2,rx3,ry3;	/* RCHILD(0) corners (0..rows-1, 0..cols-1) */
  struct Tile *L1, *R1, *L2, *R2;	/* neighboring tiles */
  Rgb **clayer;		/* color layers to paint on surface */
  /* the trec array is dynamically allocated with (3*tileOrder+2) entries */
  /* 1 for header record, 1+tileOrder for elevations, the rest for errors */
  TileRec trec[1];	/* array of tile record structs */
} Tile;

/* tile ID fields */
#define TIDORDER(tid) ((((tid)>>29)&7)+6)
#define TIDZONE(tid) (((tid)>>24)&0x1F)
#define TIDXXX(tid) (((tid)>>14)&0x3FF)
#define TIDYYY(tid) ((tid)&0x3FFF)

/* tile record indexing */
#define RECCOUNT     (3*tileOrder+4)
#define HEADREC(t)   ((t)->trec[0])
#define DATAREC(t,i) ((t)->trec[1+(i)])
#define ERROREC(t,i) ((t)->trec[2+tileOrder+(i)])

/* error field access macro */
#define ERR(t) ((t).tile->err[(t).i])

/* info field access macros */
#define INFO(t) ((t).tile->info[(t).i])
#define PUTONSURFACE(t) INFO(t) = SURFACE
#define ONSURFACE(t) (INFO(t) & SURFACE)
#define INFOV(t) (INFO(t) & (ALLINFOV | PARTINFOV))
#define SMALLN1(t) (INFO(t) & SMALLER1)
#define SMALLN2(t) (INFO(t) & SMALLER2)
#define BIGN3(t) (INFO(t) & BIGGER3)
#define ISLEAF(t) ((((t).i + 1)>>(1 + 2*tileOrder)) > 0)

#define EMNORMAL(tile,i,j,k) ((tile)->normal[((int)(j)*cols + (int)(i))*3+k])

#ifdef FIX_NORMALS
  #define GET_NORMAL {							\
    nx = vexagger / 10.0 * ((y2-y1)*(z3-z1) - (z2-z1)*(y3-y1));		\
    ny = vexagger / 10.0 * ((z2-z1)*(x3-x1) - (x2-x1)*(z3-z1));		\
    nz = (GLfloat)(mPerCell * ((x2-x1)*(y3-y1) - (y2-y1)*(x3-x1)));	\
  }
#else
  #define GET_NORMAL {				\
    nx = ((y2-y1)*(z3-z1) - (z2-z1)*(y3-y1));	\
    ny = ((z2-z1)*(x3-x1) - (x2-x1)*(z3-z1));	\
    nz = ((x2-x1)*(y3-y1) - (y2-y1)*(x3-x1));	\
  }
#endif

#define EMDATA(tile,i,j) ((tile)->data[(int)(j)*cols + (int)(i)])


/*
 * ============== Hierarchy ===============
 */
#define MAXERR USHRT_MAX	/* maximum scaled triangle error */

/* triangle info flag bits */
#define SMALLER1	  1	/* Is neighbor 1 smaller? */
#define SMALLER2	  2	/* Is neighbor 2 smaller? */
#define BIGGER3		  4	/* Is neighbor 3 larger? */
#define SURFACE		  8	/* on surface? */
#define ALLINFOV	 16	/* completely in field of view? */
#define PARTINFOV	 32	/* partially in field of view? */

typedef struct Triangle {
  Tile *tile;
  int i;
} Triangle;

typedef int TriIndex;

#define LCHILD(i) (2*(i)+1)	/* index of left child, given the parent */
#define RCHILD(i) (2*(i)+2)	/* index of right child, given the parent */
#define PARENT(i) (((i)-1)/2)	/* index of parent, given the child */


/*
 * ============== Interests ===============
 */

typedef struct Interest {	/* interest in one record */
  Tile *i_tile;
  short i_recn;			/* record number (bit index) */
  short i_pri;			/* priority */
} Interest;

typedef struct IntList {	/* interest list (dynamically sized) */
  int i_count;			/* number of entries used */
  int i_unused;			/* number of entries allocated but unused */
  Interest i_interest[1];	/* array of Interest entries */ 
} IntList;


/*
 * ============== Paths ===============
 */

/* Lineal paths */
typedef struct Path{
  int n, allocation;		/* number of points and allocated size */
  Color color;			/* color of path */
  struct Path* next;		/* next path in list */
  Coord2f *p;			/* coordinates suitable for display */
} Path;

/* Flight paths */
typedef struct FlightPath{
  int n, allocation;		/* number of points and allocated size */
  int *delay;			/* delay between frames */
  char *isRel;
  int *eyez;			/* altitude of eye (either relative or abs) */
  Coord2f *eye;			/* coordinates suitable for display */
  Coord2f *ref;			/* coordinates suitable for display */
} FlightPath;


/*
 * ============== Miscellany ===============
 */

/* Do integer conversion after adding 500 so truncation is correct */
#define UTM2XXX(xx) ((int)(((xx) - 10) / (float)tileSizeM + 500.0))
/* yy is always positive so we needn't be so tricky */
#define UTM2YYY(yy) ((int)((yy) / tileSizeM))
#define XXX2UTM(xxx) (((xxx) - 500) * tileSizeM + 10)
#define YYY2UTM(yyy) ((yyy) * tileSizeM)

/* initial values */
#define IERRTHRESH	0.01	/* decimeters / meter */
#define IERRTHRESHINC	0.0005
#define IDISTTHRESH	700.0
#define IDISTTHRESHINC	50.0
#define IVEXAGGER	1.0
#define IVEXAGGERINC	0.1


#endif	/* DEFNS_H */
