/*  emit.c -- draw a surface  */

#ifdef WIN32
  #include <windows.h>
#endif

#include "topo.h"

static void emitDFS(Triangle t,
		    int x1, int y1, int x2, int y2, int x3, int y3);
static void emitColor(int x, int y, int z);
static void emitTri(Triangle t,
		    int x1, int y1, int x2, int y2, int x3, int y3);



static int surfType;		/* surface type: WIRE2D, WIRE3D, or FILL3D */
static Rgb *rgb;



/*  emitSurface(hp, type) -- draw surface from triangle hierarchy  */

void emitSurface(int type)
{
  Tile *tile;
  Triangle t;
  Wrap *wrap;
  WrapGen *wg;

  surfType = type;
  if (surfType == WIRE2D) {
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    glColor3f(0.0,0.0,0.0);
  } else if (surfType == FOV2D) {
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  } else if (surfType == WIRE3D) {
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    glColor3f(0.0,0.0,0.0);
  } else {	/* FILL3D */
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  }

  LOCK(wrapLock);
  wg = genWraps(TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    /* let OpenGL scale and translate */
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef((float)tile->xx, (float)tile->yy, 0.0);
    glScalef((float)mPerCell, (float)mPerCell, vexagger / 10.0);
 
    /* set color layer */
    if( colorLayer >= 0 ) {
      rgb = tile->clayer[colorLayer];
    } else {
      rgb = NULL;
    }
    /* draw triangles */
    glBegin(GL_TRIANGLES);
    t.tile = tile;
    t.i = LCHILD(0);
    emitDFS(t, tile->lx1,tile->ly1,tile->lx2,tile->ly2,tile->lx3,tile->ly3);
    t.i = RCHILD(0);
    emitDFS(t, tile->rx1,tile->ry1,tile->rx2,tile->ry2,tile->rx3,tile->ry3);
    glEnd();
  }
  UNLOCK(wrapLock);
}



static void emitDFS(Triangle t, int x1, int y1, int x2, int y2, int x3, int y3)
{
  int xm, ym;

  if (ONSURFACE(t)) {
    emitTri(t, x1, y1, x2, y2, x3, y3);
  } else {

    assert(!ISLEAF(t));

    xm = (x1 + x2) / 2;
    ym = (y1 + y2) / 2;
    t.i = LCHILD(t.i);		/* recurse in left child */
    emitDFS(t, x3, y3, x1, y1, xm, ym);
    t.i = t.i + 1;		/* recurse in right child */
    emitDFS(t, x2, y2, x3, y3, xm, ym);
  }
}



static void emitTri(Triangle t, int x1, int y1, int x2, int y2, int x3, int y3)
{
  GLfloat z1, z2, z3;
  GLfloat nx, ny, nz;

  if (!cullFOV || INFOV(t)) {
    z1 = EMDATA(t.tile, x1, y1);
    z2 = EMDATA(t.tile, x2, y2);
    z3 = EMDATA(t.tile, x3, y3);

    if (surfType == WIRE2D) {
      glVertex2i(x1, y1);
      glVertex2i(x2, y2);
      glVertex2i(x3, y3);
    } else if (surfType == FOV2D) {
      switch(INFOV(t)) {
      case ALLINFOV:
	glColor3f(0.7, 0.7, 0.7);
	break;
      case PARTINFOV:
	glColor3f(0.0, 0.0, 0.8);
	break;
      default:
	glColor3f(0.5, 0.5, 0.5);
      }
      glVertex2i(x1, y1);
      glVertex2i(x2, y2);
      glVertex2i(x3, y3);
    } else if (surfType == WIRE3D) {
      glVertex3f((GLfloat)x1, (GLfloat)y1, z1);
      glVertex3f((GLfloat)x2, (GLfloat)y2, z2);
      glVertex3f((GLfloat)x3, (GLfloat)y3, z3);
    } else {	/* FILL3D */
      if( smoothSurface ) {
	nx = EMNORMAL(t.tile,x1,y1,0);
	ny = EMNORMAL(t.tile,x1,y1,1);
	nz = EMNORMAL(t.tile,x1,y1,2);
	glNormal3f(nx, ny, nz);
	emitColor(x1, y1, z1); glVertex3f((GLfloat)x1, (GLfloat)y1, z1);

	nx = EMNORMAL(t.tile,x2,y2,0);
	ny = EMNORMAL(t.tile,x2,y2,1);
	nz = EMNORMAL(t.tile,x2,y2,2);
	glNormal3f(nx, ny, nz);
	emitColor(x2, y2, z2); glVertex3f((GLfloat)x2, (GLfloat)y2, z2);

	nx = EMNORMAL(t.tile,x3,y3,0);
	ny = EMNORMAL(t.tile,x3,y3,1);
	nz = EMNORMAL(t.tile,x3,y3,2);
	glNormal3f(nx, ny, nz);
	emitColor(x3, y3, z3); glVertex3f((GLfloat)x3, (GLfloat)y3, z3);
      } else {
	GET_NORMAL;
	glNormal3f(nx, ny, nz);
	emitColor(x1, y1, z1); glVertex3f((GLfloat)x1, (GLfloat)y1, z1);
	emitColor(x2, y2, z2); glVertex3f((GLfloat)x2, (GLfloat)y2, z2);
	emitColor(x3, y3, z3); glVertex3f((GLfloat)x3, (GLfloat)y3, z3);
      }
    }
  }
}



/* emit color depending on location and altitude  */

static void emitColor(int x, int y, int z)
{
  Color *c;

  if( rgb != NULL ) {			/* use color from PPM files */
    glColor3ub(RED(rgb, x, y), GREEN(rgb, x, y), BLUE(rgb, x, y));

  } else if (colorLayer == T_CONTOURS){	/* base the color on the elevation */
    if (z == 0) {
      glColor3f(0.4f, 0.0f, 0.0f);		/* dark red for non-data */
    } else if (z <= HIGH_TIDE && z >= LOW_TIDE) {
      glColor3f(0.0f, 0.0f, 0.4f);		/* dark blue for sea level */
    } else {
      c = elevMap + (int)(z * z2level);
      glColor3f(c->red, c->green, c->blue);
    }

} else {					/* gray with slight tan tint */
    glColor3f(0.6f, 0.6f, 0.55f);
  }
}
