/*  errcalc.c -- triangle error calculations  */

#include "topo.h"

typedef struct {	/* triangle error accumulator */
  Tile *tile;		/* containing tile */
  float a, b, c;	/* coefficients of containing plane */
  float recErr;		/* maximum recorded error */
} TEA;



/**
 *  Initializes a TEA for accumulating errors in a triangle.
 */
static void planeInit(Tile *newt, TEA *t,
  int x1, int y1, int x2, int y2, int x3, int y3)
{
  int z1, z2, z3;
  float det;

  z1 = EMDATA(newt, x1, y1);
  z2 = EMDATA(newt, x2, y2);
  z3 = EMDATA(newt, x3, y3);
  det = x1*y2 + x2*y3 + x3*y1 - x3*y2 - x2*y1 - x1*y3;
  t->a = (z1*y2 + z2*y3 + z3*y1 - z3*y2 - z2*y1 - z1*y3) / det;
  t->b = (x1*z2 + x2*z3 + x3*z1 - x3*z2 - x2*z1 - x1*z3) / det;
  t->c = (x1*y2*z3 + x3*y1*z2 + x2*y3*z1 - x3*y2*z1 - x1*y3*z2 - x2*y1*z3) /det;
  t->recErr = 0.0;
  t->tile = newt;
}



/**
 *  Calculates the error at (x,y) and updates the maximum error in the TEA.
 */
static void recordErr(TEA *t, int x, int y)
{
  float z, v;

  z = t->a * x + t->b * y + t->c;
  v = EMDATA(t->tile, x, y);
  v = z - v;
  v = ABS(v);
  t->recErr = MAX(t->recErr, v);
}



/**
 *  Calculates, stores, and returns the error for a triangle.
 *
 *  @param unknown	height of tree with errs unknown (counting from leaves)
 *  @param hgt		height of this subtree (including levels to be pruned)
 *  @param o		triangle orientation (units of 45deg from 1st quadrant)
 *  @param x1,y1	first counterclockwise vertex
 *  @param x2,y2	second counterclockwise vertex
 *  @param x3,y3	right-angled vertex; origin for orientation measurement
 */
int errDFS(Triangle t, int unknown, int hgt, int o,
		  int x1, int y1, int x2, int y2, int x3, int y3)
{
  int err, e4, e5;
  int x4, y4, x5, y5;
  int v, w;
  int i, j;
  TEA tea;
  Tile *newt;

  int xm = (x1 + x2) / 2;		/* midpoint of hypotenuse */
  int ym = (y1 + y2) / 2;

  if( hgt > 1 ) {			/* if not at deepest level, recurse */
    i = t.i;
    t.i = LCHILD(t.i);
    errDFS(t, unknown, hgt - 1, o + 5, x3, y3, x1, y1, xm, ym);
    t.i += 1;
    errDFS(t, unknown, hgt - 1, o + 3, x2, y2, x3, y3, xm, ym);
    t.i = i;
  }

  if( hgt > unknown )			/* if in region of known error */
    return 0;

  newt = t.tile;

  if( hgt <= 3 ) {			/* if easy special case near bottom */
    if( hgt <= 1 ) {			/* if leaf node */
      return ERR(t) = 0;		/* error is zero */
    }

    /* error at midpoint */
    err = EMDATA(newt, xm, ym) -
      (EMDATA(newt, x1, y1) + EMDATA(newt, x2, y2)) / 2.0;
    err = ABS(err);

    if (hgt == 3) {			/* if two more points to consider */
      x4 = (x1 + x3) / 2;
      y4 = (y1 + y3) / 2;
      e4 = EMDATA(newt,x4,y4) - (EMDATA(newt,x1,y1) + EMDATA(newt,x3,y3)) / 2.0;
      e4 = ABS(e4);
      err = MAX(err, e4);
      x5 = (x2 + x3) / 2;
      y5 = (y2 + y3) / 2;
      e5 = EMDATA(newt,x5,y5) - (EMDATA(newt,x2,y2) + EMDATA(newt,x3,y3)) / 2.0;
      e5 = ABS(e5);
      err = MAX(err, e5);
    }

    return ERR(t) = (unsigned short) (err + 0.5);
  }

  /* general case: find the plane containing the triangle, */
  /* and check its fit at every point within. */

  planeInit(newt, &tea, x1, y1, x2, y2, x3, y3);
  w = x3 - x1;
  w = ABS(w);
  v = x3 - x2;
  v = ABS(v);
  w = MAX(w, v);
  switch (o & 7) {		/* scanning depends on triangle orientation */
  case 0:
    for (i = w; i >= 0; i--)
      for (j = w - i; j >= 0; j--)
	recordErr(&tea, x3 + j, y3 + i);
    break;
  case 1:
    for (i = 0; i <= w; i++)
      for (j = -i; j <= i; j++)
	recordErr(&tea, x3 + j, y3 + i);
    break;
  case 2:
    for (i = w; i >= 0; i--)
      for (j = w - i; j >= 0; j--)
	recordErr(&tea, x3 - j, y3 + i);
    break;
  case 3:
    for (j = 0; j <= w; j++)
      for (i = -j; i <= j; i++)
	recordErr(&tea, x3 - j, y3 + i);
    break;
  case 4:
    for (i = w; i >= 0; i--)
      for (j = w - i; j >= 0; j--)
	recordErr(&tea, x3 - j, y3 - i);
    break;
  case 5:
    for (i = 0; i <= w; i++)
      for (j = -i; j <= i; j++)
	recordErr(&tea, x3 + j, y3 - i);
    break;
  case 6:
    for (i = w; i >= 0; i--)
      for (j = w - i; j >= 0; j--)
	recordErr(&tea, x3 + j, y3 - i);
    break;
  case 7:
    for (j = 0; j <= w; j++)
      for (i = -j; i <= j; i++)
	recordErr(&tea, x3 + j, y3 + i);
    break;
  }
  return ERR(t) = (unsigned short) (tea.recErr + 0.5);
}
