/*  fetch.c -- file reading and token service interface  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "topo.h"

static Wrap *tsopen(Wrap *w);



/*
 *  fetch(name, type) -- locate data source and create Wrap struct.
 *
 *  Creates a new Wrap struct and attempts to open a file for it.
 *  wrap->token is set non-null if the source is located using the
 *  token service.  wrap->file is non-null if the source has been
 *  successfully opened (but not loaded).
 *  
 *  The Wrap is NOT added to the Wrap list.
 *  Do this after initializing it.
 */
Wrap *fetch(char *name, int type) {

  Wrap *w = newWrap(name, type);
  if( ! tsopen(w) )
    w->file = uopen(name);
  return w;

}



/*
 *  destroy(w) -- deallocate Wrap struct.
 *
 *  w->name and w->data are freed.  (No harm if either is null.)
 *  If not null, w->token is released.
 *  If not null, w->file is closed.
 *  The struct w is zeroed (as a precaution) and freed.
 *  NULL is returned.
 */
void *destroy(Wrap *w) {

  free(w->name);
  free(w->data);
  if( w->file )
    w->file->close(w->file);

  #ifdef TSLOOKUP
    if( w->token )
      Tok_Release(w->token);
  #endif

  memset(w, 0, sizeof(*w));
  free(w);
  return NULL;
}



#ifndef TSLOOKUP		/* if no token service, use stubs */

  static Wrap *tsopen(Wrap *w) {
    return NULL;
  }

#else  /* TSLOOKUP defined */



Tok_Service tsvc = NULL;	/* token service handle */



/*
 *  initTSVCHandle -- initialize the token service.
 */
void initTSVCHandle(char *cfg)
{
  char **servers = Tok_Config(cfg);
  if( !servers ) {
    fprintf(stderr, "Invalid token service configuration\n");
    exit(1);
  }
  fprintf(stderr, "connecting to token service... ");
  tsvc = Tok_Open(servers);
  fprintf(stderr, "OK\n");
}



/*
 *  invalidate(tk, wrap) -- release token and mark Wrap as stale.
 *
 *  Called by the token service when it wants us to release a token.
 */
static void invalidate(Tok_Token t, void *arg) {

  Wrap *w = (Wrap *) arg;
  w->stale = 1;
  w->token = NULL;
  Tok_Release(t);
}



/*
 *  tsopen(w) -- open Wrap source using token service.
 *
 *  tsopen obtains (for reading) the token associated with w->name
 *  and opens the associated file, setting w->token and w->file.
 *  It returns w if the token could be obtained, otherwise NULL.
 *  w->file is null if no file could be opened.
 */
static Wrap *tsopen(Wrap *w) {
  Tok_Token t;

  if( tsvc == NULL )
    return NULL;
  t = Tok_Request(tsvc, w->name, TOK_SHARED, invalidate, w);
  if( t == NULL )
    return NULL;
  w->token = t;
  w->file = bopen(Tok_GetData(t), Tok_GetLength(t));
  return w;
}



#endif  /* TSLOOKUP */
