/*  infov.c -- field-of-view (FOV) calculations  */

#include "topo.h"



static void getV(Tile *tile, int p,
		 int *px1, int *py1, int *px2, int *py2, int *px3, int *py3)
/*
 * Don't use getV.  Use getVerts.
 *
 * p is t.i + 1 so that the bit sequence describes a path in the tree.
 */
{
  switch(p) {
  case 0:
  case 1:
  case 2:
    *px1 = tile->lx1;
    *py1 = tile->ly1;
    *px2 = tile->lx2;
    *py2 = tile->ly2;
    *px3 = tile->lx3;
    *py3 = tile->ly3;
    return;
  case 3:
    *px1 = tile->rx1;
    *py1 = tile->ry1;
    *px2 = tile->rx2;
    *py2 = tile->ry2;
    *px3 = tile->rx3;
    *py3 = tile->ry3;
    return;
  default:
    if (p & 1) {
      getV(tile, p>>1, px3, py3, px1, py1, px2, py2);
      *px3 = (*px1 + *px3) / 2;
      *py3 = (*py1 + *py3) / 2;
      return;
    } else {
      getV(tile, p>>1, px2, py2, px3, py3, px1, py1);
      *px3 = (*px2 + *px3) / 2;
      *py3 = (*py2 + *py3) / 2;
      return;
    }
  }
}



void getVerts(Triangle t,
	      int *px1, int *py1, int *px2, int *py2, int *px3, int *py3)
{
  getV(t.tile, t.i + 1, px1, py1, px2, py2, px3, py3);
}



static int inFOVpt(int x, int y)
{
  if( fovType == WEDGE ) {
    return( fovLnorm[0]*(x-eye[0])+fovLnorm[1]*(y-eye[1]) <= 0 &&
	    fovRnorm[0]*(x-eye[0])+fovRnorm[1]*(y-eye[1]) <= 0 );
  } else if( fovType == FOUR_GON ) {
    return( LEFT_TURN(x, y, fovLLpt[0], fovLLpt[1], fovLRpt[0], fovLRpt[1]) &&
	    LEFT_TURN(x, y, fovLRpt[0], fovLRpt[1], fovURpt[0], fovURpt[1]) &&
	    LEFT_TURN(x, y, fovURpt[0], fovURpt[1], fovULpt[0], fovULpt[1]) &&
	    LEFT_TURN(x, y, fovULpt[0], fovULpt[1], fovLLpt[0], fovLLpt[1]) );
  } else {
    return( LEFT_TURN(x, y, fovLLpt[0], fovLLpt[1], eye[0], eye[1]) &&
	    LEFT_TURN(x, y, eye[0], eye[1], fovLRpt[0], fovLRpt[1]) &&
	    LEFT_TURN(x, y, fovLRpt[0], fovLRpt[1], fovURpt[0], fovURpt[1]) &&
	    LEFT_TURN(x, y, fovURpt[0], fovURpt[1], fovULpt[0], fovULpt[1]) &&
	    LEFT_TURN(x, y, fovULpt[0], fovULpt[1], fovLLpt[0], fovLLpt[1]) );
  }
}



static int crossFOV(int x1, int y1, int x2, int y2)
{
  double tin=0.0, tout=1.0;
  double num,den;

#define CLIP(N,R)\
  num = N[0]*(R[0]-x1)+N[1]*(R[1]-y1);\
  den = N[0]*(x2-x1)+N[1]*(y2-y1);\
  if( den < 0 ) {\
    tin = MAX(tin, num/den);\
  } else if( den > 0 ) {\
    tout = MIN(tout, num/den);\
  } else if( num < 0 )\
    return 0;\
  if( tin > tout ) return 0;

  CLIP(fovLnorm, fovLLpt);
  CLIP(fovRnorm, fovLRpt);
  if( fovType == WEDGE ) return 1;
  CLIP(fovTnorm, fovULpt);
  CLIP(fovBLnorm, fovLLpt);
  if( fovType == FOUR_GON ) return 1;
  CLIP(fovBRnorm, fovLRpt);
  return 1;
}



static int inFOV(int x1, int y1, int x2, int y2, int x3, int y3)
{
  int a, b, c;

  a = inFOVpt(x1, y1);
  b = inFOVpt(x2, y2);
  c = inFOVpt(x3, y3);
  if( a && b && c ) return ALLINFOV;
  if( a || b || c ) return PARTINFOV;

  /* All tri points are outside FOV.  Does tri cover part of FOV? */
  if( crossFOV(x1, y1, x2, y2) ) return PARTINFOV;
  if( crossFOV(x2, y2, x3, y3) ) return PARTINFOV;
  if( crossFOV(x3, y3, x1, y1) ) return PARTINFOV;

  /* Perhaps all of FOV is in triangle (check eye position) */
  if( LEFT_TURN(eye[0], eye[1], x1, y1, x2, y2) &&
      LEFT_TURN(eye[0], eye[1], x2, y2, x3, y3) &&
      LEFT_TURN(eye[0], eye[1], x3, y3, x1, y1) )
    return PARTINFOV;

  return 0;
}



/*  triInFOV(t) -- is triangle t in field of view?  */

int triInFOV(Triangle t)
{
  int x1, y1, x2, y2, x3, y3;
  Tile *tile;

  getVerts(t, &x1, &y1, &x2, &y2, &x3, &y3);
  tile = t.tile;
  return inFOV(x1 * mPerCell + tile->xx, y1 * mPerCell + tile->yy,
	       x2 * mPerCell + tile->xx, y2 * mPerCell + tile->yy,
	       x3 * mPerCell + tile->xx, y3 * mPerCell + tile->yy);
}
