/*  in.c -- general-purpose I/O functions  */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "topo.h"



/*
 *  uopen(s) -- open stream specified by a C string giving a URL or filename.
 *
 *  If s is null, use stdin.
 */
inp uopen(char *s)
{
  if (s)
    return bopen(s, strlen(s));
  else {
    SET_BINARY_MODE(stdin);
    return openFP(stdin);
  }
}



/*
 *  bopen(s, len) -- open stream specified by string with explicit length.
 */
inp bopen(char *s, size_t len)
{
  char *t;
  char buf[MAX_URL_LEN + 1];

  /* check first for literal value, which may be of any length */
  if ((t = preskip(s, "literal:")) != NULL)
    return openMem(t, len - (t - s));

  /* all others need to be zero-byte-terminated */
  if (len > MAX_URL_LEN)
    return NULL;
  memcpy(buf, s, len);
  buf[len] = '\0';
  if ((t = preskip(buf, "file:///")) != NULL)	/* standard URL file: form */
    return openFile(t - 1);
  if ((t = preskip(buf, "file:")) != NULL)	/* commonly used shortcut */
    return openFile(t);
  if ((t = preskip(buf, "http://")) != NULL)	/* HTTP URL */
    return openHTTP(buf);
  if ((t = preskip(buf, "tvtp://")) != NULL)	/* TVTP URL */
    return openTVTP(buf);

  /* when all else fails, try opening as unadorned filename */
  return openFile(buf);
}



/*
 *  parseURL(char *s, url *r) -- parse URL s into struct r.
 *
 *  It is assumed that s is no longer than MAX_URL_LEN.
 *
 *  Returns 1 if successful, 0 if not.  Parsing is somewhat lenient.
 */
int parseURL(char *s, url *r) {
  char *p;
  int n;

  r->port = 0;
  r->scheme[0] = r->host[0] = r->path[0] = '\0';

  /* check for "xxxx:..." but do not match "A:..." */
  p = strchr(s, ':');
  if( p != NULL ) {
    n = p - s;
    if( n > 1 && n < sizeof(r->scheme) ) {
      strncpy(r->scheme, s, n);
      s = s + n + 1;
    }
  }

  /* allow "file:path" or just "path" although not a legal URL */
  if( !strcmp(s, "//") ) {
    strcpy(r->path, s);
    return 1;
  }

  /* break apart host and path parts */
  if( sscanf(s, "//%[^/]/%s", r->host, r->path) < 1 )
    return 0;

  /* check for a port number in the host field */
  p = strchr(r->host, ':');
  if( p != NULL ) {
    r->port = atoi(p + 1);
    if( r->port == 0 )
      return 0;
    *p = '\0';
  }

  return 1;
}



/*
 *  readBytes(in, buf, len) -- read data into buffer.
 *
 *  Uses in->rchar to read len bytes of data into buf.
 *  Returns the number of bytes actually stored, which
 *  indicates EOF if less than len.
 */

int readBytes(inp in, void *buf, size_t len)
{
  int c;
  size_t n = 0;
  char *p = buf;

  while (n < len) {
    c = in->rchar(in);
    if (c == EOF)
      break;
    *p++ = c;
    n++;
  }
  return n;
}



/*
 *  readLine(in, buf, len) -- read line into buffer.
 *
 *  Uses in->rchar to read the next line into buffer buf of size len (> 0).
 *  As much of that line as will fit is copied into buf.
 *  The input line may be terminated by \r, \n, or \r\n.
 *  It is stored with only a NUL terminator.
 *
 *  Returns buf, or NULL for EOF.
 */

char * readLine(inp in, char *buf, size_t len)
{
  int c;
  char *p = buf;

  for (;;) {
    c = in->rchar(in);
    if (c == EOF || c == '\r' || c == '\n') {
      *p = '\0';
      if (c == '\r' && in->peek(in) == '\n')
        in->rchar(in);
      if (c == EOF && p == buf)
        return NULL;
      else
        return buf;
    }
    if (len > 0) {			/* if room to store */
      *p++ = c;
      len--;
    }
  }
}



#ifdef INTEST

/* 
 *  main program for testing input functions (#define INTEST to enable)
 *  
 *  usage:  a.out url
 */
int main(int argc, char *argv[])
{
  inp in;
  char buf[1000];
  size_t len;

  if (argc != 2)
    { fprintf(stderr, "usage: %s URL\n", argv[0]); exit(1); }
  in = uopen(argv[1]);
  if (in == NULL)
    { fprintf(stderr, "cannot uopen %s\n", argv[1]); exit(1); }
  in = unzip(in);
  do {
    len = in->rbytes(in, buf, sizeof(buf));
    fwrite(buf, 1, len, stdout);
  } while (len > 0);
  return 0;
}

#endif  /* INTEST */
