/*  interest.c -- calculation of "most needed" tiles  */

#include "topo.h"

static void calcScores(void);
static void setScores(Tile *tile, int base);
static int viewChanged(void);
static int scoreTile(Tile *tile);
static void addintr(IntList **ilp, Tile *tile, TileRec *trec, int bitnum);
static int intrcmp(const void *intr1, const void *intr2);


#define CORNER_WEIGHT 1		/* score for corner in overhead window */
#define DISTANCE_WEIGHT 20	/* score for distance to reference point */
#define ANGLE_WEIGHT 20		/* score for closeness to view angle */
#define TILE_WEIGHT (1 + 4 * CORNER_WEIGHT + DISTANCE_WEIGHT + ANGLE_WEIGHT)

#define RECORD_WEIGHT 18	/* interest variation of records in a tile */
#define LEVEL_INCR 6		/* increment per record level */

#if TILE_WEIGHT + RECORD_WEIGHT > MAX_INTEREST
  #error MAX_INTEREST exceeded
#endif



static int prevCounter;		/* overhead counter at last update */
static double exprev, eyprev;	/* previous eye location */
static double rxprev, ryprev;	/* previous eye location */



/*
 *  mostNeeded() -- return the "most needed" tile.
 *
 *  The "most needed" tile is the TYPE_WANTED entry with waiting=0
 *  having the highest interest score for its header record,
 *  Note that only existing, TYPE_WANTED tiles are considered.
 *  NULL is returned if no tile qualifies.
 */
Tile *mostNeeded(void) {
  Tile *tile;
  Tile *best = NULL;
  Wrap *wrap;
  WrapGen *wg;

  calcScores();
  
  LOCK(wrapLock);
  wg = genWraps(TYPE_WANTED);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    if( tile->waiting != 0 )
      continue;
    if( !best || HEADREC(tile).tr_interest > HEADREC(best).tr_interest )
      best = tile;
  }
  UNLOCK(wrapLock);

  return best;
}



/*
 *  calcScores() -- calculate interest scores for tiles.
 *
 *  Calculates and stores interest scores for every record of every
 *  TYPE_WANTED tile and every TYPE_TILE with waiting > 0.
 */
static void calcScores(void) {
  Tile *tile;
  Wrap *wrap;
  WrapGen *wg;
  int recalc;

  recalc = viewChanged();

  LOCK(wrapLock);
  wg = genWraps(TYPE_WANTED | TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    if( wrap->type == TYPE_TILE && tile->waiting == 0 )
      continue;
    if( recalc || HEADREC(tile).tr_interest == 0 ) {
      setScores(tile, scoreTile(tile));
    }
  }
  UNLOCK(wrapLock);
}



/*
 *  setScores(tile, base) -- store scores in tile records.
 *
 *  Given a base score, saves individual record scores.
 *  If R is RECORD_WEIGHT, the maximum bias,
 *  and N is LEVEL_INCR, then
 *	the header record scores base + R
 *	elevation record i scores BASE + R - 1 - LEVEL_INCR * i
 *	each error records scores one more than corresponding elevation record
 *  except that no score is reduced below the base.
 */
static void setScores(Tile *tile, int base) {
  int i, score;

  score = base + RECORD_WEIGHT;
  HEADREC(tile).tr_interest = score;
  score--;
  for( i = 0; i <= tileOrder; i++ ) {
    DATAREC(tile,i).tr_interest = score;
    ERROREC(tile,2*i).tr_interest = score + 1;
    ERROREC(tile,2*i+1).tr_interest = score + 1;
    score = score - LEVEL_INCR;
    if( score < base )
      score = base;
  }
}



/*
 *  viewChanged() -- check whether the view has changed from the last call.
 */
static int viewChanged(void) {

  if( prevCounter != ovhCounter()
  || exprev != eye[0] || eyprev != eye[1]
  || rxprev != ref[0] || ryprev != ref[1] ) {
    prevCounter = ovhCounter();
    exprev = eye[0];
    eyprev = eye[1];
    rxprev = ref[0];
    ryprev = ref[1];
    return 1;
  } else {
    return 0;
  }
}



/*
 *  scoreTile(tile) -- calculate the score for one tile.
 *
 *  min   1 point to indicate we've calculated a score.
 *  up to 4 * CORNER_WEIGHT points for tile corners within the overhead window.
 *  up to DISTANCE_WEIGHT points depending on distance from reference point.
 *  up to ANGLE_WEIGHT points depending on distance from viewing axis.
 */
static int scoreTile(Tile *tile) {
  int n, score;
  float d, da, dx, dy, m, tx, ty, ta, va;

  score = 1;		/* one point just to show we've done the calculation */

  score += CORNER_WEIGHT * inOvh(tile->xx, tile->yy);	/* count corners */
  score += CORNER_WEIGHT * inOvh(tile->xx, tile->yy + tileSizeM);
  score += CORNER_WEIGHT * inOvh(tile->xx + tileSizeM, tile->yy);
  score += CORNER_WEIGHT * inOvh(tile->xx + tileSizeM, tile->yy + tileSizeM);

  tx = tile->xx + tileSizeM / 2;	/* center coordinates of tile */
  ty = tile->yy + tileSizeM / 2;

  /* distance calculations */
  dx = tx - ref[0];
  dy = ty - ref[1];
  d = sqrt(dx * dx + dy * dy) / tileSizeM;	/* distance in tile units */
  if( d < 1 )
    m = 1.0;
  else
    m = 1 / d;
  n = (DISTANCE_WEIGHT + 0.9999) * m;
  score += n;

  /* angle calculations */
  ta = atan2(ty - eye[1], tx - eye[0]);
  va = atan2(ref[1] - eye[1], ref[0] - eye[0]);
  da = fabs(ta - va);
  if( da > PI )
    da = 2 * PI - da;
  n = (ANGLE_WEIGHT + 0.9999) * (1 - (da / PI));
  score += n;

  return score;
}



/*
 *  interests() -- return an allocated list of all interests.
 */
IntList *interests(void) {
  Tile *tile;
  Wrap *wrap;
  WrapGen *wg;
  IntList *intlist;
  int i, n;

  calcScores();
  intlist = zalloc(sizeof(IntList));

  LOCK(wrapLock);
  wg = genWraps(TYPE_WANTED | TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    addintr(&intlist, tile, &HEADREC(tile), 25); 
    n = MIN(9, tileOrder + 1);
    for( i = 0; i < n; i++ )
      addintr(&intlist, tile, &DATAREC(tile, i), 24 - i);
    n = MIN(15, 2 * tileOrder + 1);
    for( i = 1; i < n; i++ )
      addintr(&intlist, tile, &ERROREC(tile, i), 16 - i);
  }
  UNLOCK(wrapLock);

  qsort(intlist->i_interest, intlist->i_count, sizeof(Interest), intrcmp); 
  return intlist;
}



/*
 *  addintr(&ilp, tile, tilerec, bitnum) -- check interest for one tile record.
 *
 *  May append an Interest record to the IntList pointed to by ilp.
 *  May reallocate the IntList and update *ilp.
 */

#define LIST_INCR 100

static void addintr(IntList **ilp, Tile *tile, TileRec *trec, int bitnum) {
  IntList *intlist = *ilp;
  Interest *newint;

  if( tile->rcvmask & (1 << bitnum) )	/* if already received or denied */
    return;
  if( intlist->i_unused == 0 ) {	/* if list is full */
    intlist = *ilp = realloc(intlist,
      sizeof(IntList) + (intlist->i_count + LIST_INCR - 1) * sizeof(Interest));
    if( !intlist )
      nospace();
    intlist->i_unused = LIST_INCR;
  }
  intlist->i_unused--;
  newint = &intlist->i_interest[intlist->i_count++];
  newint->i_tile = tile;
  newint->i_recn = bitnum;
  newint->i_pri = trec->tr_interest;
}



/*
 *  intrcmp(intr1, intr2) -- compare two Interest structs for sorting.
 */
static int intrcmp(const void *i1, const void *i2) {
  const Interest *intr1 = i1;
  const Interest *intr2 = i2;
  int n = intr2->i_pri - intr1->i_pri;		/* descending priority sort */
  if( n == 0 )
    n = intr2->i_tile->tid - intr2->i_tile->tid;   /* break ties consistently */
  return n;
}
