/*  intvtp.c -- I/O functions for reading from a TopoVista server  */

#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include "topo.h"

typedef struct {
  struct instruct in;			/* standard object header */
  char prev_data_seq;			/* last data record seen */
  char prev_err_seq;			/* last error record seen */
  int sock;				/* UDP socket */
  size_t len;				/* data available */
  unsigned char *buf;			/* data buffer */
  unsigned char *next;			/* next byte available */
} ins_tvtp;

static int loadBuffer(ins_tvtp *tvp);
static int peekTV(inp);
static int charTV(inp);
static void closeTV(inp);



/*  openTVTP(URL) -- create instruct for reading via TVTP.  */

inp openTVTP(char *URL)
{
  int fd, id;
  char request[MAX_URL_LEN];
  struct hostent *hostaddr;
  struct sockaddr_in sockaddr;
  url u;
  ins_tvtp *in;

  if( strlen(URL) > MAX_URL_LEN )
    return NULL;
  if( !parseURL(URL, &u) )
    return NULL;
  if( u.port == 0 )
    u.port = TV_PORT;
  id = tileID(u.path);
  if( !id )
    return NULL;

  /* look up host */
  hostaddr = gethostbyname(u.host);
  if( !hostaddr )
    return NULL;

  /* create socket */
  if( (fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0 )
    return NULL;

  /* bind to port */
  memset(&sockaddr, 0, sizeof(sockaddr));
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_addr.s_addr = INADDR_ANY;
  sockaddr.sin_port = 0;
  if( bind(fd, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) < 0 )
    { close(fd); return NULL; }

  /* connect remote port */
  sockaddr.sin_port = htons(u.port);
  memcpy(&sockaddr.sin_addr, hostaddr->h_addr, hostaddr->h_length);
  if( connect(fd, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) < 0 )
    { close(fd); return NULL; }

  /* send request */
  sprintf(request, "GET %08X %s", id, u.path);
  if( send(fd, request, strlen(request), 0) < 0 )
    { close(fd); return NULL; }

  /* allocate inp struct */
  in = alloc(sizeof(ins_tvtp));
  in->in.peek = peekTV;
  in->in.rchar = charTV;
  in->in.rbytes = readBytes;
  in->in.rline = readLine;
  in->in.close = closeTV;
  in->prev_data_seq = TV_ELEV - 1;
  in->prev_err_seq = TV_ERRS - 1;
  in->sock = fd;
  in->len = 0;
  in->buf = alloc(TV_MAXLEN);

  /* read first response and check for success */
  if( loadBuffer(in) <= 0 )
    { closeTV(&in->in); return NULL; }

  /* if all okay, return struct pointer */
  return &in->in;
}



/*  peekTV(in) -- peek at next character from memory.  */

static int peekTV(inp in)
{
  ins_tvtp *tvp = (ins_tvtp *) in;
  if( tvp->len == 0 && loadBuffer(tvp) == 0 )
    return EOF;
  return *tvp->next;
}



/*  charTV(in) -- read next character from memory.  */

static int charTV(inp in)
{
  ins_tvtp *tvp = (ins_tvtp *) in;
  if( tvp->len == 0 && loadBuffer(tvp) == 0 )
    return EOF;
  tvp->len--;
  return *tvp->next++;
}



/*  closeTV(in) -- close memory instruct.  */

static void closeTV(inp in)
{
  ins_tvtp *tvp = (ins_tvtp *) in;
  close(tvp->sock);			/* close socket */
  free(tvp->buf);			/* free buffer */
  memset(in, 0, sizeof(ins_tvtp));	/* prevent object reuse after free */
  free(in);				/* free object */
}



/*
 *  loadBuffer(in) -- read data from socket, returning length
 *
 *  If TopoVista flag bytes indicate that a data record or error record
 *  has been lost from the sequence, subsequent data or error records
 *  (respectively) are also discarded here because they are invalid
 *  without their predecessors.
 */
static int loadBuffer(ins_tvtp *tvp)
{
  int n, flag;

  if( tvp->buf == NULL )			/* if EOF already handled */
    return 0;
  for( ;; ){
    n = recv(tvp->sock, tvp->buf, TV_MAXLEN, 0);	/* read message */
    if( n < TV_PHLEN || tvp->buf[6] == TV_EOF ) {	/* if error or EOF */
      free(tvp->buf);
      tvp->buf = NULL;					/* prevent retry */
      return 0;
    }
    flag = tvp->buf[TV_PHLEN + 4] & ~TV_GZIP;
    if( flag & TV_ERRS ){
      /* this is an error record */
      if( flag != tvp->prev_err_seq + 1 )
        continue;		/* out of sequence -- ignore */
      else
        tvp->prev_err_seq = flag;
    } else if( flag & TV_ELEV ){
      /* this is a data record */
      if( flag != tvp->prev_data_seq + 1 )
        continue;		/* out of sequence -- ignore */
      else
        tvp->prev_data_seq = flag;
    }
    tvp->next = tvp->buf + TV_PHLEN;		/* success! */
    return tvp->len = n - TV_PHLEN;
  }
}
