/*  lock.c -- multithreading interlocking  */

/*
 *  These functions are called by macros of the same names (but uppercased)
 *  in defns.h.  The macros pass file/line information to assist debugging.
 *  Locks are recursive (as, for example, in Java).
 */

#include <errno.h>
#include <stdio.h>
#include "topo.h"

#define LDEBUG 1    /* 0 = fast, 1 = check, 2 = note waits, 3 = trace all */



static void trace(char *what, Lock *k, char *where, int n) {
  #if LDEBUG >= 3
    fprintf(stderr, "%8s %p => %-2d %s\n", what, k, n, where);
  #endif
}



#ifdef UNITHREAD


  /* implementation for single-threaded use (just dummy stubs) */

  void initlock(Lock *k, char *where)	{ }
  void lock(Lock *k, char *where)	{ }
  void rattle(Lock *k, char *where)	{ }
  void unlock(Lock *k, char *where)	{ }


#else


  /* implementation for multithreaded use  */
  /* (commented in terms of macro usage) */


  /*
   *  INITLOCK(k) -- initialize Lock k.
   */
  void initlock(Lock *k, char *where) { 

    trace("INITLOCK", k, where, 0);
    k->depth = 0;
    k->where = where;
    if( pthread_mutex_init(&k->mxlock,NULL) != 0 ) {
      perror(where);
      exit(1);
    }
  }


  /*
   *  LOCK(k) -- reserve Lock k.
   */
  void lock(Lock *k, char *where) { 
    pthread_t self;

    self = pthread_self();

    /* check for recursive call on a lock already held */
    if( k->depth > 0 && pthread_equal(k->owner, self) ){
      k->depth++;
      trace("LOCK", k, where, k->depth);
      return;
    }

    if( pthread_mutex_trylock(&k->mxlock) != 0 ) {
      #if LDEBUG >= 2
      fprintf(stderr,
        "LOCK at %s waiting on lock grabbed at %s\n", where, k->where);
      #endif
      if( pthread_mutex_lock(&k->mxlock) != 0 ) {
        perror(where);
        exit(1);
      }
      #if LDEBUG >= 2
      fprintf(stderr, "LOCK at %s obtained\n", where);
      #endif
    }
    k->owner = self;
    k->depth = 1;
    k->where = where;
    trace("LOCK", k, where, k->depth);
  }


  /*
   *  RATTLE(k) -- confirm that Lock k is locked.
   */
  void rattle(Lock *k, char *where) {
    #if LDEBUG >= 1
      if( k->depth <= 0 || !pthread_equal(k->owner, pthread_self()) ) {
        fprintf(stderr, "LOCK NOT HELD at %s\n", where);
        exit(1);
      }
    #endif
  }


  /*
   *  UNLOCK(k) -- release Lock k.
   */
  void unlock(Lock *k, char *where) { 

    trace("UNLOCK", k, where, k->depth - 1);
    rattle(k, where);
    if (--k->depth == 0) {
      k->where = where;
      if( pthread_mutex_unlock(&k->mxlock) != 0 ) {
        perror(where);
        exit(1);
      }
    }
  }


#endif
