/*  overhead.c -- creation, drawing, and maintenance of overhead view window  */

#ifdef WIN32
  #include <windows.h>
#endif

#include "topo.h"

static void ovhMouseCB(int button, int state, int x, int y);
static void ovhMotionCB(int x, int y);
static void makeReliefMap(void);
static void showStatus(int x, int y, int xxx, int yyy);
static void statusColor(Tile *tile, TileRec *trec);
static void interestColor(int interest);
static void fillSquare(float x, float y, float r);
static void fillWedge(float x, float y, float r1, float r2, int i, int n);

static int whichBtn;			/* which mouse button is down? */

static int rad1, rad2, rad3;		/* status disc layout */

#define COLORATION 0.7		/* color importance (vs. shading) (0. to 1.) */
#define CONTRAST 0.0470589	/* magic contrast number 120/10/255 */
/* #define CONTRAST   4.0 */
#define CMUL (255 * COLORATION)		/* color multiplier */
#define RMUL (255 * (1 - COLORATION))	/* relief multiplier */



/*  these statics are interlocked using wrapLock  */

static int ovhWidth, ovhHeight;		/* in pixels */
static int reliefxx, reliefyy;		/* SW corner of reliefMap in meters */
static int reliefxxInit, reliefyyInit;
static int nconfigs;			/* updated when ovh config changes */

static unsigned char *reliefMap=NULL;

static double mPerPixel;		/* meters per pixel */
static double border;			/* border width in meters */

static double ogridw, ogridh;		/* size of window onto tile grid */



/*
 *  Requests loading of tiles visible in the overhead window.
 *
 *  Holds the wrapLock for the duration of its loop, so asynchronous
 *  threads see all the tiles being added as batch.
 */
static void reqOvhTiles( int makeRlf )
{
  int x, y;
  int refresh = 0;

  /* load tiles covered by window */
  LOCK(wrapLock);
  for( x = UTM2XXX(reliefxx+ogridw-mPerCell); x >= UTM2XXX(reliefxx); x-- ) {
    for( y = UTM2YYY(reliefyy+ogridh-mPerCell); y >= UTM2YYY(reliefyy); y-- ) {
      makeRlf |= reqTileXY(x,y);
    }
  }
  UNLOCK(wrapLock);

  if( makeRlf ) {
    refresh = RP + NR + NS;	/* forces new tiles to be triangulated */
  }

  postRedisplays(refresh+RO);
}



void ovhReshapeCB(int w, int h)
{
  /* resize to minimum size if necessary */
  if( w < 2 * BORDER ) w = 2 * BORDER;
  if( h < 2 * BORDER ) h = 2 * BORDER;

  ovhWidth = w;
  ovhHeight = h;
  border = BORDER * mPerPixel;
  ogridw = (w - 2 * BORDER) * mPerPixel;
  ogridh = (h - 2 * BORDER) * mPerPixel;

  glViewport(0,0,w,h);

  reqOvhTiles(1);	/* always rebuild relief if window reconfigured */
  nconfigs++;
}



/*  openOvhWin() -- initialize relief map window  */

void openOvhWin(void)
{
  int h = HEIGHT_2D;
  int w = HEIGHT_2D;	/* square */

  mPerPixel = tileSizeM / (double)(w - 2 * BORDER);

  if( mPerPixel < mPerCell ) {
    mPerPixel = mPerCell;
    w = h = tileSizeM / mPerCell + 2 * BORDER;
  }

  rad3 = 20 * mPerPixel;
  rad2 = 0.6 * rad3;
  rad1 = 0.2 * rad3;

  ogridw = ogridh = tileSizeM;

  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
  glutInitWindowSize(w, h);
  ovhWin = glutCreateWindow("overhead");

  reliefxx = reliefxxInit = xxmin;
  reliefyy = reliefyyInit = yymin;

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glPixelZoom(mPerCell / mPerPixel, mPerCell / mPerPixel);

  glClearColor(0.0, 0.0, 0.0, 0.0);

  glutKeyboardFunc(keyCB);
  glutSpecialFunc(specialCB);
  glutMouseFunc(ovhMouseCB);
  glutMotionFunc(ovhMotionCB);
  glutDisplayFunc(ovhDisplayCB);
  glutReshapeFunc(ovhReshapeCB);
}



void ovhDisplayCB(void)
{
  char wname[200];
  Path *pth;
  FlightPath *fpth;
  Wrap *wrap;
  WrapGen *wg;
  int i, x, y, xxx, yyy, xskip, yskip, nx, ny;

  buildSurface();	/* sets FOV as well as surface approximation */

  LOCK(wrapLock);

  if( redoRelief ) {	/* rebuild relief map if necessary */
    redoRelief = 0;
    makeReliefMap();
  }

  glClear(GL_COLOR_BUFFER_BIT);
  glPixelStorei(GL_UNPACK_ROW_LENGTH, (xxmax - xxmin) / mPerCell + 1);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D((GLdouble)(reliefxx - border),
	     (GLdouble)(reliefxx + ogridw + border),
	     (GLdouble)(reliefyy - border),
	     (GLdouble)(reliefyy + ogridh + border));

  if (twoDMesh != FOV2D) {
    /* draw background */
    if( reliefxx < xxmin ) {
      x = xxmin;
      xskip = 0;
    } else {
      x = reliefxx;
      xskip = (reliefxx - xxmin) / mPerCell;
    }
    glPixelStorei(GL_UNPACK_SKIP_PIXELS, xskip);
    if( reliefyy < yymin ) {
      y = yymin;
      yskip = 0;
    } else {
      y = reliefyy;
      yskip = (reliefyy - yymin) / mPerCell;
    }
    glPixelStorei(GL_UNPACK_SKIP_ROWS, yskip);
    glRasterPos2i(x, y);
    nx = (xxmax - reliefxx) / mPerCell + 1;
    ny = (yymax - reliefyy) / mPerCell + 1;
    if( nx > ogridw / mPerCell ) nx = ogridw / mPerCell;
    if( ny > ogridh / mPerCell ) ny = ogridh / mPerCell;
    if( nx > (xxmax - xxmin) / mPerCell ) nx = 1 + (xxmax - xxmin) / mPerCell;
    if( ny > (yymax - yymin) / mPerCell ) ny = 1 + (yymax - yymin) / mPerCell;
    if( nx > 0 && ny > 0 )
      glDrawPixels(nx, ny, GL_RGB, GL_UNSIGNED_BYTE, reliefMap);
  }
	
  if (twoDMesh != NONE2D) {
    /* draw triangles in overhead view */
    glLineWidth(1.0);
    emitSurface(twoDMesh);
    sprintf(wname, "%d triangles", numSurfTri);
  } else {
    synthName(wname, utmZone, UTM2XXX(reliefxx + ogridw / 2.0),
      UTM2YYY(reliefyy + ogridh / 2.0), tileOrder);
  }
  glutSetWindowTitle(wname);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  if (twoDMesh != FOV2D) {

    /* draw paths */
    glLineWidth(2.0);
    wg = genWraps(TYPE_PATH);
    while( (wrap = nextWrap(wg)) != NULL ) {
      for( pth = wrap->data; pth != NULL; pth = pth->next ) {
        glColor3f(pth->color.red, pth->color.green, pth->color.blue);
        glBegin(GL_LINE_STRIP);
        for( i=0; i<pth->n; ++i ) {
	    glVertex2f(pth->p[i][0], pth->p[i][1]);
        }
        glEnd();
      }
    }
    wg = genWraps(TYPE_FLIGHTPATH);
    while( (wrap = nextWrap(wg)) != NULL ) {
      fpth = wrap->data;
      glColor3f(0.0,0.0,1.0);
      glBegin(GL_LINE_STRIP);
      for( i=0; i<fpth->n; ++i ) {
        glVertex2f(fpth->eye[i][0], fpth->eye[i][1]);
      }
      glEnd();
    }

    /* draw reference point */
    glColor3f(0.0,0.0,0.0);
    glPointSize(2);
    glBegin(GL_POINTS);
    glVertex2f(ref[0], ref[1]);
    glEnd();

    /* draw FOV region */
    glBegin(GL_LINE_STRIP);
    if( fovType == WEDGE ) {
      glVertex2f(fovLLpt[0], fovLLpt[1]);
      glVertex2f(eye[0], eye[1]);
      glVertex2f(fovLRpt[0], fovLRpt[1]);
    } else if( fovType == FOUR_GON ) {
      glVertex2f(fovLLpt[0], fovLLpt[1]);
      glVertex2f(fovLRpt[0], fovLRpt[1]);
      glVertex2f(fovURpt[0], fovURpt[1]);
      glVertex2f(fovULpt[0], fovULpt[1]);
      glVertex2f(fovLLpt[0], fovLLpt[1]);
    } else {
      glVertex2f(eye[0], eye[1]);
      glVertex2f(fovLRpt[0], fovLRpt[1]);
      glVertex2f(fovURpt[0], fovURpt[1]);
      glVertex2f(fovULpt[0], fovULpt[1]);
      glVertex2f(fovLLpt[0], fovLLpt[1]);
      glVertex2f(eye[0], eye[1]);
    }
    glEnd();
  }

  /* draw tile status discs */
  if( twoDMesh == FOV2D ) {
    for( xxx=UTM2XXX(reliefxx); (x=XXX2UTM(xxx)) < reliefxx+ogridw; xxx++ ) {
      for( yyy=UTM2YYY(reliefyy); (y=YYY2UTM(yyy)) < reliefyy+ogridh; yyy++ ) {
	showStatus(x + tileSizeM / 2, y + tileSizeM / 2, xxx, yyy);
      }
    }
    glFlush();		/* else some discs may disappear */
    			/* (9/2003 local nVidia graphics) */
  }

  /* draw border regions */
  glColor3f(1.0, 0.8, 0.7);
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  glRecti(reliefxx - border, reliefyy - border,
	  reliefxx, reliefyy + ogridh + border);
  glRecti(reliefxx + ogridw, reliefyy - border,
	  reliefxx + ogridw + border, reliefyy + ogridh + border);
  glRecti(reliefxx, reliefyy - border,
	  reliefxx + ogridw, reliefyy);
  glRecti(reliefxx, reliefyy + ogridh,
	  reliefxx + ogridw, reliefyy + ogridh + border);

  /* draw cross-hairs */
  glLineWidth(1.0);
  glBegin(GL_LINES);
  x = XXX2UTM(UTM2XXX(reliefxx) + 1);
  while( x <= reliefxx + ogridw ) {
    glVertex2f(x, reliefyy);
    glVertex2f(x, reliefyy + ogridh);
    x += tileSizeM;
  }
  y = YYY2UTM(UTM2YYY(reliefyy) + 1);
  while( y <= reliefyy + ogridh ) {
    glVertex2f(reliefxx, y);
    glVertex2f(reliefxx + ogridw, y);
    y += tileSizeM;
  }
  glEnd();

  UNLOCK(wrapLock);

  glutSwapBuffers();
}



/* draw status disc for the tile xxx,yyy at x,y */

static void showStatus(int x, int y, int xxx, int yyy) {
  int i, n;
  char name[MAX_TILE_NAME];
  Wrap *w;
  Tile *t;

  LOCK(wrapLock);

  synthPath(name, utmZone, xxx, yyy, tileOrder);
  w = findWrap(name, TYPE_WANTED | TYPE_TILE | TYPE_UNAV);
  if( w == NULL ) {
    glColor3f(0.0, 0.0, 0.5);  		/* dark blue square if no tile entry */
    fillSquare(x, y, rad2);
    UNLOCK(wrapLock);
    return;
  }
  t = w->data;

  if( w->type == TYPE_UNAV ) {
    glColor3f(0.3, 0.3, 0.3);		/* dark gray square if unavailable */
    fillSquare(x, y, rad2);
    UNLOCK(wrapLock);
    return;
  }

  statusColor(t, &HEADREC(t));		/* status disc: tile header */
  fillSquare(x, y, rad1);

  n = tileOrder + 1;
  for( i = 0; i < n; i++ ) {
    statusColor(t, &DATAREC(t, i));	/* inner ring: elevation records */
    fillWedge(x, y, rad1, rad3, i, n);
  }

  n = 2 * tileOrder + 2;
  for( i = 0; i < n; i++ ) {
    statusColor(t, &ERROREC(t, i));	/* outer ring: error records */
    fillWedge(x, y, rad2, rad3, i, n);
  }

  UNLOCK(wrapLock);
}


/* set color for status display based on wrap type and tile record state */

static void statusColor(Tile *tile, TileRec *tr) {
  Wrap *wrap = tile->wrap;

  if( tr->tr_data != NULL )
    glColor3f(1.0, 0.2, 0.6);		/* hot pink if ready to process */
  else if( tr->tr_done )
    glColor3f(0.8, 0.8, 0.8);		/* light gray if processed */
  else if( tr->tr_unav )
    glColor3f(0.3, 0.3, 0.3);		/* dark gray if record unavailable */
  else if( wrap->type == TYPE_TILE && tile->waiting <= 0 )
    glColor3f(0.0, 0.0, 0.0);		/* black if no more action expected */
  else if( tr->tr_interest > 0 )
    interestColor(tr->tr_interest);		/* record interest */
  else if( HEADREC(tile).tr_interest > 0 )
    interestColor(HEADREC(tile).tr_interest);	/* header interest */
  else if( tile->waiting > 0 )
    glColor3f(0.8, 1.0, 0.8);		/* pale green if wait w/ no interest */
  else if( wrap->type == TYPE_WANTED )
    glColor3f(1.0, 0.5, 0.0);		/* orange if wanted w/ no interest */
  else
    glColor3f(0.0, 1.0, 1.0);		/* cyan for unexpected state */
}


/* set color based on nonzero interest value */

static void interestColor(int interest) {
  float f = 4.0 * interest / (MAX_INTEREST + 1);	/* f in [0,4) */
  float a = f - (int) f;				/* fractional part */
  switch( (int)f ) {
    case 0:  glColor3f(0, .4*a, .8-.4*a);	break;	/* blue to teal */
    case 1:  glColor3f(0, .4+.6*a, .4-.4*a);	break;	/* teal to green */
    case 2:  glColor3f(a, 1, 0);		break;	/* green to yellow */
    case 3:  glColor3f(1, 1, a);		break;	/* yellow to white */
  }
}



/* fill square of side 2*r centered at x,y */

static void fillSquare(float x, float y, float r) {
  glBegin(GL_QUADS);
  glVertex2f(x + r, y + r);
  glVertex2f(x + r, y - r);
  glVertex2f(x - r, y - r);
  glVertex2f(x - r, y + r);
  glEnd();
}


/* fill wedge i of n (clockwise from noon) between radii r1 and r2 */

static void fillWedge(float x, float y, float r1, float r2, int i, int n) {
  float a1 = PI / 2 - i * 2 * PI / n;
  float a2 = PI / 2 - (i + 1) * 2 * PI / n;
  glBegin(GL_QUADS);
  glVertex2f(x + r1 * cos(a1), y + r1 * sin(a1));
  glVertex2f(x + r2 * cos(a1), y + r2 * sin(a1));
  glVertex2f(x + r2 * cos(a2), y + r2 * sin(a2));
  glVertex2f(x + r1 * cos(a2), y + r1 * sin(a2));
  glEnd();
}



static void makeReliefMap( void )
{
  Tile *tile;
  Wrap *wrap;
  WrapGen *wg;
  Rgb *rgb;
  unsigned char *oldMap;
  int w, h;
  int imin, jmin;
  int i, j, k, d0, d1, z, q;
  int dz;
  double m, lum, r, g, b;
  Color *c;

  RATTLE(wrapLock);
  w = (double)(xxmax - xxmin) / mPerCell + 1;
  h = (double)(yymax - yymin) / mPerCell + 1;
  m = CONTRAST / mPerCell;
	
  /* allocate new zeroed (black) buffer, atomically replacing the old one */
  oldMap = reliefMap;
  reliefMap = zalloc(3 * w * h);
  free(oldMap);
	
  wg = genWraps(TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    imin = (tile->xx - xxmin) / mPerCell;
    jmin = (tile->yy - yymin) / mPerCell;

    if (colorLayer >= 0) {
	rgb = tile->clayer[colorLayer];
    } else {
	rgb = NULL;
    }

    /* compute elevation difference for each output pixel */
    for( j = 0; j < rows; ++j ) {
      for( i = 0; i < cols; ++i ) {
	d0 = EMDATA(tile, i, j);
	if( j > 0 )
	  d1 = EMDATA(tile, i, j-1);
	else
	  d1 = d0;
	dz = d1 - d0;
	/* compute color at each pixel. */
	lum = 0.5 + m * dz;		/* luminance */
	if (lum > 1.0)
	  lum = 1.0;			/* clamp to valid range */
	if (lum < 0.0)
	  lum = 0.0;

	if (rgb != NULL) {
	  r = RED(rgb, i, j) / 255.0;
	  g = GREEN(rgb, i, j) / 255.0;
	  b = BLUE(rgb, i, j) / 255.0;

	} else if (colorLayer == T_CONTOURS) {
	  z = EMDATA(tile, i, j);	/* elevation */
	  if (z == 0) {
	    r = 0.4; g = 0.0; b = 0.0;	/* use dark red for non-data */
	  } else if (z <= HIGH_TIDE && z >= LOW_TIDE) {
	    r = 0.0;			/* use dark blue at sea level */
	    g = 0.0;
	    b = 0.4;
	  } else {
	    k = z * z2level;		/* color index */
	    c = &elevMap[k];		/* color entry */
	    r = c->red;
	    g = c->green;
	    b = c->blue;
	  }
	} else {
	  r = g = b = lum;		/* use luminance only */
	}

	q = 3 * (imin + i + w * (jmin + j));
	reliefMap[q] = RMUL * lum + CMUL * r;	/* r */
	reliefMap[q+1] = RMUL * lum + CMUL * g;	/* g */
	reliefMap[q+2] = RMUL * lum + CMUL * b;	/* b */
      }
    }
  }
}



static void ovhMouseCB(int button, int state, int x, int y)
{
  whichBtn = button;			/* remember button for motion events */
  if (state != GLUT_UP)
    mouseto(button, x, y);		/* set position */
}



static void ovhMotionCB(int x, int y)
{
  mouseto(whichBtn, x, y);
}



/*  mouseto(btn, x, y) -- handle mouse click or drag  */

void mouseto(int btn, int xw, int yw)
{
  Tile *tile;
  double x, y, lon, lat;
  int i, j, k, z;
  int movePortal = 0;	/* flags */
  Rgb *rgb;
	
  yw = ovhHeight - yw;	/* Y is reported as distance from top */

  if( !redoSurface &&	/* don't move portal if new surface is pending */
  /* is mouse in border region? */
      xw >= 0 && yw >= 0 && xw <= ovhWidth && yw <= ovhHeight ) {
    if( xw < BORDER ) {
      movePortal = 1;
      reliefxx -= border;
    } else if( xw + BORDER > ovhWidth ) {
      movePortal = 1;
      reliefxx += border;
    }
    if( yw < BORDER ) {
      movePortal = 1;
      reliefyy -= border;
    } else if( yw + BORDER > ovhHeight ) {
      movePortal = 1;
      reliefyy += border;
    }

    if( movePortal ) reqOvhTiles(0);
  }

  x = reliefxx + (xw - BORDER) * mPerPixel;
  y = reliefyy + (yw - BORDER) * mPerPixel;

  LOCK(wrapLock);

  tile = tileCovering(x, y);
  if (btn == GLUT_RIGHT_BUTTON) {
    ref[0] = x;
    ref[1] = y;
    setEyeRefHeight();
    postRedisplays(RO+RP+NS);		/* new surface needed */

  } else if (btn == GLUT_LEFT_BUTTON) {
    eye[0] = x;
    eye[1] = y;
    setEyeRefHeight();
#ifdef TESTING
    if( doingLoop ) postRedisplays(loopRefresh);
    else
#endif
      postRedisplays(RO+RP+NS);		/* new surface needed */
		
  } else {				/* GLUT_MIDDLE_BUTTON */
    if( tile ) {
      i = (x - tile->xx) / mPerCell;
      j = (y - tile->yy) / mPerCell;
      z = zval(tile, i, j);
      utm2lola(utmZone, x, y, &lon, &lat);
      /*    datum2datum(gem->datum, mouseLola, WGS84, &mouseLola); */
      printf("%10.6f %10.6f ", lat / 3600., lon / 3600. );
      if (z==0) {
	printf("no elevation");
      } else {
	printf("elev %6.1fm", (z - ELEVATION_BIAS) / 10.0);
      }
      if (colorLayer >= 0) {
	rgb = tile->clayer[colorLayer];
	k = findLegend(rgb->legend,RED(rgb,i,j),GREEN(rgb,i,j),BLUE(rgb,i,j));
	if( k < rgb->legend.size )
	  printf(" %s",rgb->legend.entry[k].label);
      }
      printf("\n");
    }
  }

  nconfigs++;
  UNLOCK(wrapLock);
}



/*
 * followPathCB() -- follow the flightpath
 */
void followPathCB(int dummy)
{
  FlightPath* p = NULL;
  Wrap* wrap;
  WrapGen *wg;
  int x, y;
  
  assert( flightStep != 0 );	/* should never be called when flightStep=0 */

  if( flightStep < 0 ) {
    flightStep = 0;
    return;
  }

  LOCK(wrapLock);
  wg = genWraps(TYPE_FLIGHTPATH);
  wrap = nextWrap(wg);
  if( !wrap ) {
    UNLOCK(wrapLock);
    return;
  }
  endGen(wg);
  p = wrap->data;
        
  if( p->isRel[flightStep-1] ) {
    eyeIsRelative = 1;
    eyeZRelative = 10 * p->eyez[flightStep-1];  /* convert to decimeters */
  } else {
    eyeIsRelative = 0;
    eyeZAbsolute = 10 * p->eyez[flightStep-1];  /* convert to decimeters */
  }

  x = p->ref[flightStep-1][0];
  y = p->ref[flightStep-1][1];
  ref[0] = x;
  ref[1] = y;
  ovhVis(x, y);

  x = p->eye[flightStep-1][0];
  y = p->eye[flightStep-1][1];
  eye[0] = x;
  eye[1] = y;
  ovhVis(x, y);

  setEyeRefHeight();
  UNLOCK(wrapLock);

  postRedisplays(RC+RO+RP+NS);

  if( flightStep < p->n ) {
    glutTimerFunc(p->delay[flightStep], followPathCB, flightStep);
    ++flightStep;
  } else {
    flightStep = 0;
  }
}



void ovhHome( void )
{
  reliefxx = reliefxxInit;
  reliefyy = reliefyyInit;
  nconfigs++;
}



void ovhEnd( void )
{
  reliefxx = xxmin;
  reliefyy = yymin;
  nconfigs++;
}


/*  ovhVis(x, y) -- ensure that (x, y) is visible within the overhead window  */

void ovhVis(int x, int y)
{
  int xoff, yoff, xmar, ymar, mmax;

  LOCK(wrapLock);

  xmar = ogridw / 4;		/* default the enforced margin to 1/4 window */
  ymar = ogridh / 4;
  mmax = 100 * mPerPixel;	/* limit to a max of 100 pix for lg windows */
  if( xmar > mmax ) xmar = mmax;
  if( ymar > mmax ) ymar = mmax;

  xoff = x - reliefxx;
  yoff = y - reliefyy;
  if( xoff < xmar )
    reliefxx = x - xmar;
  else if( xoff > ogridw - xmar )
    reliefxx = x - ogridw + xmar;
  if( yoff < ymar )
    reliefyy = y - ymar;
  else if( yoff > ogridh - ymar )
    reliefyy = y - ogridh + ymar;

  reqOvhTiles(reqTileXY(UTM2XXX(x), UTM2YYY(y)));
  nconfigs++;
  UNLOCK(wrapLock);
}



/* inOvh(x, y) -- return 1 if (x, y) is within overhead window, else 0.  */

int inOvh(int x, int y)
{
  return (x >= reliefxx) && (y >= reliefyy)
    && (x < reliefxx + ogridw) && (y < reliefyy + ogridh);
}



/* ovhCounter() -- return serial number of this overhead configuration.  */

int ovhCounter(void)
{
  return nconfigs;
}
