/*  path.c -- read and draw paths across terrain  */

/*
 * Created by Xiangyu Zhang, to emit a path on the terrain
 * Rewritten by Will Evans to avoid DFS for each segment of path.
 */
#ifdef WIN32
  #include <windows.h>
#endif

#include "topo.h"



/* Use vertex reflection to get neighbor vertices from current tri? */
/* #define REFLECT */

static Color pathColor[] = {
  {1.0f, 0.0f, 0.0f },	/* red */
  {1.0f, 1.0f, 0.0f },	/* yellow */
  {0.0f, 1.0f, 0.0f },	/* green */
  {0.0f, 0.0f, 1.0f },	/* blue */
};


/* number of colors */
static int pathCount = sizeof(pathColor) / sizeof(pathColor[0]);
static int npaths = 0;

static Triangle tri;

/* global triangle coordinates in meters */
static int gx1, gy1, gx2, gy2, gx3, gy3;
static double z1, z2, z3;
static double bz, b0;



static void emitVertex(Coord2f p)
{
  glVertex3f(p[0], p[1], (float)(bz + 0.1));
}



static void setZ(void)
{
  z1 = EMDATA(tri.tile,
	      (gx1-tri.tile->xx)/mPerCell, (gy1-tri.tile->yy)/mPerCell);
  z2 = EMDATA(tri.tile,
	      (gx2-tri.tile->xx)/mPerCell, (gy2-tri.tile->yy)/mPerCell);
  z3 = EMDATA(tri.tile,
	      (gx3-tri.tile->xx)/mPerCell, (gy3-tri.tile->yy)/mPerCell);

  /* set denominator for barycentric coordinate calculation */
  b0=(gx2-gx1)*(gy3-gy1)-(gx3-gx1)*(gy2-gy1);
  /* b0 shouldn't be 0 */
}



static int setBary(float x, float y)
     /* sets the global bz (the z value of point p) and
      * returns 0 if the point p is outside the current triangle
      * returns 1 if it is inside.
      */
{
  double b1,b2,b3;

  b1=((gx2-x)*(gy3-y)-(gx3-x)*(gy2-y))/b0;
  b2=((gx3-x)*(gy1-y)-(gx1-x)*(gy3-y))/b0;
  b3=((gx1-x)*(gy2-y)-(gx2-x)*(gy1-y))/b0;
  bz = b1*z1 + b2*z2 + b3*z3;

  if( b1 < 0 || b2 < 0 || b3 < 0 ) {
    /*    fprintf(stderr, "b1 %f b2 %f b3 %f\n", b1, b2, b3);*/
    return 0;
  }
  else return 1;
}



static void ptLoc( float x, float y )
/* set "tri" to the triangle that contains the point x,y */
{
  int xm, ym;

  if( ONSURFACE(tri) ) {
    setZ();			/* set the z coords of tri */
    assert(setBary(x,y));		/* set the barycentric coords of x,y */
    return;
  }
  xm = (gx1 + gx2) / 2;
  ym = (gy1 + gy2) / 2;
  if( LEFT_OR_NO_TURN(xm, ym, gx3, gy3, x, y) ) {
    gx2 = gx1;
    gy2 = gy1;
    gx1 = gx3;
    gy1 = gy3;
    tri.i = LCHILD(tri.i);
  } else {
    gx1 = gx2;
    gy1 = gy2;
    gx2 = gx3;
    gy2 = gy3;
    tri.i = RCHILD(tri.i);
  }
  gx3 = xm;
  gy3 = ym;
  ptLoc(x, y);
}



static int pointLocate( Coord2f p )
{
  Tile *tile;
  float x, y;

  tile = tileCovering((double)p[0], (double)p[1]);
  if( tile ) {
    x = (p[0] - tile->xx) / mPerCell;
    y = (p[1] - tile->yy) / mPerCell;
    tri.tile = tile;

    /*
    fprintf(stderr, "(%f,%f) in tile (%d,%d)(%d,%d)\n",
	    p[0], p[1],
	    tile->xx, tile->yy,
	    tile->xx + (cols-1)*mPerCell,
	    tile->yy + (rows-1)*mPerCell);
    */

    if( LEFT_TURN(tile->lx1, tile->ly1, tile->lx2, tile->ly2, x, y) ) {
      tri.i = LCHILD(0);
      gx1 = tile->lx1 * mPerCell + tile->xx;
      gy1 = tile->ly1 * mPerCell + tile->yy;
      gx2 = tile->lx2 * mPerCell + tile->xx;
      gy2 = tile->ly2 * mPerCell + tile->yy;
      gx3 = tile->lx3 * mPerCell + tile->xx;
      gy3 = tile->ly3 * mPerCell + tile->yy;
    } else {
      tri.i = RCHILD(0);
      gx1 = tile->rx1 * mPerCell + tile->xx;
      gy1 = tile->ry1 * mPerCell + tile->yy;
      gx2 = tile->rx2 * mPerCell + tile->xx;
      gy2 = tile->ry2 * mPerCell + tile->yy;
      gx3 = tile->rx3 * mPerCell + tile->xx;
      gy3 = tile->ry3 * mPerCell + tile->yy;
    }
    ptLoc(p[0], p[1]);
    return 1;
  }
  return 0;
}



#define EPSILON 0.000001f

static int clipLeft( int bx1, int by1, int bx2, int by2, Coord2f p, Coord2f q )
     /* return 0 if segment p,q is to right of line b1,b2
      * return 1 if segment p,q is to left of  line b1,b2
      * return 2 if p is left of and q is right of line b1,b2 and change q
      * return 3 if q is left of and p is right of line b1,b2 and change p
      */
{
  int dbx = bx2 - bx1;
  int dby = by2 - by1;
  float num, den;
  float t;

  num = (bx1 - p[0])*dby - (by1 - p[1])*dbx;
  den = (q[0] - p[0])*dby - (q[1] - p[1])*dbx;
  if( den == 0 ) {	/* parallel */
    if( num < 0 ) return 0;
    return 1;
  }
  t = num / den;
  if( den > 0 ) {	/* in to out */
    if( t < 1 ) {
      if( t >= 0 ) {
	t -= EPSILON;	/* make sure new q lies inside halfplane */
	if( t < 0 ) t = 0;
	q[0] = p[0] + t*(q[0]-p[0]);
	q[1] = p[1] + t*(q[1]-p[1]);
	return 2;
      }
      return 0;
    }
    return 1;
  }
  /* out to in */
  if( t > 0 ) {
    if( t <= 1 ) {
      t += EPSILON;	/* make sure new p lies inside halfplane */
      if( t > 1 ) t = 1;
      p[0] = p[0] + t*(q[0]-p[0]);
      p[1] = p[1] + t*(q[1]-p[1]);
      return 3;
    }
    return 0;
  }
  return 1;
}



static int clipToTiles( Coord2f p, Coord2f q )
{
  Coord2f tp, tq;
  Tile *tile;
  Wrap *wrap;
  WrapGen *wg;
  float tin, tout, t, x, y;
  int xmin, ymin, xmax, ymax;

  tin = 1;	/* set so that first clip will change it */
  tout = 0;	/* set so that first clip will change it */

  LOCK(wrapLock);
  wg = genWraps(TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    /* clip to tile boundary */
    tp[0] = p[0];
    tp[1] = p[1];
    tq[0] = q[0];
    tq[1] = q[1];
    xmin = tile->xx;
    ymin = tile->yy;
    xmax = xmin + tileSizeM;
    ymax = ymin + tileSizeM;
    if( !clipLeft(xmin, ymin, xmax, ymin, tp, tq) ) continue;
    if( !clipLeft(xmax, ymin, xmax, ymax, tp, tq) ) continue;
    if( !clipLeft(xmax, ymax, xmin, ymax, tp, tq) ) continue;
    if( !clipLeft(xmin, ymax, xmin, ymin, tp, tq) ) continue;
    if( tp[0] != p[0] ) {
      t = (tp[0] - p[0])/(q[0]-p[0]);
      if( t < tin ) tin = t;
    } else {
      tin = 0;
    }
    if( tq[0] != q[0] ) {
      t = (tq[0] - p[0])/(q[0]-p[0]);
      if( t > tout ) tout = t;
    } else {
      tout = 1;
    }
  }
  UNLOCK(wrapLock);

  if( tin == 1 && tout == 0 ) return 0;
  x = p[0];
  y = p[1];
  if( tin < 1 ) {
    p[0] = p[0] + tin * (q[0]-p[0]);
    p[1] = p[1] + tin * (q[1]-p[1]);
  }
  if( tout > 0 ) {
    q[0] = q[0] + tout * (q[0]-x);
    q[1] = q[1] + tout * (q[1]-y);
  }
  return 1;
}



static int clipExit(Coord2f p, Coord2f q)
{
  Triangle nbr;
  Coord2f pp, qq;

  nbr.tile = NULL;
  nbr.i = 0;

  pp[0] = p[0]; pp[1] = p[1];
  qq[0] = q[0]; qq[1] = q[1];
  /* Since p is in tri, clipLeft will never return 0. */
  if( clipLeft(gx1, gy1, gx2, gy2, pp, qq) == 2 ) {
    nbr = nbr3(tri);
  }
  if( clipLeft(gx2, gy2, gx3, gy3, pp, qq) == 2 ) {
    nbr = nbr1(tri);
  }
  if( clipLeft(gx3, gy3, gx1, gy1, pp, qq) == 2 ) {
    nbr = nbr2(tri);
  }
  setBary(qq[0], qq[1]);	/* set bz.  should always return true. */
  emitVertex(qq);
  p[0] = qq[0]; p[1] = qq[1];	/* set p to the exiting vertex */
  if( nbr.tile == NULL ) return 0;
  /* set the next tri */
  /* Descend from the hierarchy root to get the neighbor vertices.
     This seems to be only slightly less efficient than
     calculating the neighbor vertices from tri using vertex reflection.
  */
  getVerts(nbr, &gx1, &gy1, &gx2, &gy2, &gx3, &gy3);
  gx1 = gx1 * mPerCell + nbr.tile->xx;
  gy1 = gy1 * mPerCell + nbr.tile->yy;
  gx2 = gx2 * mPerCell + nbr.tile->xx;
  gy2 = gy2 * mPerCell + nbr.tile->yy;
  gx3 = gx3 * mPerCell + nbr.tile->xx;
  gy3 = gy3 * mPerCell + nbr.tile->yy;
  tri.i = nbr.i;
  tri.tile = nbr.tile;
  setZ();
  return 1;
}



static void emitPath(Coord2f* a, int n)
     /* This is the main routine.
      * 1) Find the point p and triangle tri where the path first
      *    intersects ROD.
      * 2) Emit vertices in tri until the path leaves tri.
      * 3) Use clipExit to find (and emit) the last vertex on tri and
      *    set p to that vertex.
      * 4) If the path hasn't left tiles, set tri to the just entered
      *    neighbor and goto 2, otherwise goto 1.
      */
{
  Coord2f p, q;
  int i;

  i = 0;
  while( 1 ) {
    while( i+1 < n ) {
      p[0] = a[i][0]; p[1] = a[i][1];
      q[0] = a[i+1][0]; q[1] = a[i+1][1];
      if( clipToTiles(p, q) ) break;
      ++i;
    }
    ++i;
    if( i >= n ) break;
    pointLocate(p);	/* set current triangle and z-coord of p */
    glBegin(GL_LINE_STRIP);
    /*
      glPointSize(5);
      glBegin(GL_POINTS);
    */
    emitVertex(p);
    while( 1 ) {
      while( i < n && setBary(a[i][0], a[i][1]) ) {
	emitVertex(a[i]);
	p[0] = a[i][0]; p[1] = a[i][1];
	++i;
      }
      if( i >= n || !clipExit(p, a[i]) ) break;
    }
    glEnd();
  }
}



/*
 * emitPaths() -- draw all the paths on the Wrap list.
 */
void emitPaths( void )
{
  Path *pth;
  FlightPath *fpth;
  Wrap *wrap;
  WrapGen *wg;

  glLineWidth(4.0);

  LOCK(wrapLock);
  wg = genWraps(TYPE_PATH);
  while( (wrap = nextWrap(wg)) != NULL ) {
    for( pth = wrap->data; pth != NULL; pth = pth->next ) {
      glColor3f(pth->color.red, pth->color.green, pth->color.blue);
      emitPath(pth->p, pth->n);
    }
  }
  wg = genWraps(TYPE_FLIGHTPATH);
  while( (wrap = nextWrap(wg)) != NULL ) {
    fpth = wrap->data;
    glColor3f(1.0,0.0,0.0);
    emitPath(fpth->eye, fpth->n);
  }
  UNLOCK(wrapLock);
}



/* Read file of lon. lat. coords and add to Wrap list */
Path* readPaths(char* filename)
{
  inp f;
  double lon, lat, x, y;
  Wrap *wrap;
  Path* tmp;
  Path* path = NULL;
  char buf[512];
  int i, endpath;

  wrap = fetch(filename, TYPE_PATH);
  f = wrap->file;
  if( f == NULL )
    return destroy(wrap);
  endpath = 1;		/* trigger initial path allocation */
  while( f->rline(f,buf,sizeof(buf)) ) {
    i = strlen(buf);
    if( i == 0 ) { endpath = 1; continue; }	/* blank lines mean new path */
    i--;
    while( i >= 0 && isspace(buf[i]) ) --i;
    while( i >= 0 && !isspace(buf[i]) ) --i;
    if( sscanf(buf+i+1, "%lf", &lon) < 1 ) { endpath = 1; continue; }
    while( i >= 0 && isspace(buf[i]) ) --i;
    while( i >= 0 && !isspace(buf[i]) ) --i;
    if( sscanf(buf+i+1, "%lf", &lat) < 1 ) { endpath = 1; continue; }

    if( endpath ) {
      tmp = (Path *)malloc(sizeof(Path));
      if( tmp == NULL ) nospace();
      tmp->p = NULL;
      tmp->next = path;
      tmp->n = tmp->allocation = 0;
      tmp->color.red = pathColor[npaths].red;
      tmp->color.green = pathColor[npaths].green;
      tmp->color.blue = pathColor[npaths].blue;
      path = tmp;
      endpath = 0;
    }
    lon *= 3600.0;
    lat *= 3600.0;
    /*   datum2datum(WGS84, lola, gem->datum, &lola); */
    lola2utm(lon, lat, &x, &y);

    if( path->n >= path->allocation ) {
      path->allocation += path->allocation + 1;
      path->p = (Coord2f *)realloc(path->p, path->allocation*sizeof(Coord2f));
      if( path->p == NULL ) nospace();
    }
    path->p[path->n][0] = (float)x;
    path->p[path->n][1] = (float)y;
    path->n++;
  }
  npaths = (npaths + 1) % pathCount;
  f->close(f);
  wrap->file = NULL;
  wrap->type = TYPE_PATH;
  wrap->data = path;
  addWrap(wrap);
  return path;
}



/* Read flightpath file. */
FlightPath* readFlightPath(char* filename)
{
  inp f;
  int delay, eyez;
  char c;
  char isRel;
  double eyelon, eyelat, reflon, reflat, x, y;
  Wrap *wrap;
  FlightPath* path = NULL;
  char buf[512];

  wrap = fetch(filename, TYPE_FLIGHTPATH);
  f = wrap->file;
  if( f == NULL )
    return destroy(wrap);
  f->rline(f,buf,sizeof(buf));
  if( sscanf(buf," # flight pat%c",&c) != 1 )  /* parse #flightpath liberally */
    return destroy(wrap);

  path = malloc(sizeof(FlightPath));
  if( path == NULL ) nospace();
  path->n = path->allocation = 0;
  path->delay = NULL;
  path->isRel = NULL;
  path->eyez = NULL;
  path->eye = NULL;
  path->ref = NULL;

  while( f->rline(f,buf,sizeof(buf)) ) {
    if( strlen(buf) == 0 ) continue;	/* skip blank lines */
    if( buf[0] == '#' ) continue;	/* skip comments */
    if( sscanf(buf,"%d +%d %lf %lf %lf %lf",
	       &delay, &eyez, &eyelat, &eyelon, &reflat, &reflon) == 6 ) {
      isRel = 1;
    } else if( sscanf(buf,"%d %d %lf %lf %lf %lf",
		      &delay, &eyez, &eyelat, &eyelon,
		      &reflat, &reflon) == 6 ) {
      isRel = 0;
    } else {
      continue;
    }

    if( path->n >= path->allocation ) {
      path->allocation += path->allocation + 1;
      path->delay = realloc(path->delay, path->allocation*sizeof(int));
      path->isRel = realloc(path->isRel, path->allocation*sizeof(char));
      path->eyez = realloc(path->eyez, path->allocation*sizeof(int));
      path->eye = realloc(path->eye, path->allocation*sizeof(Coord2f));
      path->ref = realloc(path->ref, path->allocation*sizeof(Coord2f));
      if( path->delay == NULL ||
	  path->isRel == NULL ||
	  path->eyez == NULL ||
	  path->eye == NULL ||
	  path->ref == NULL ) nospace();
    }
    path->delay[path->n] = delay;
    path->isRel[path->n] = isRel;
    path->eyez[path->n] = eyez;

    eyelon *= 3600.0;
    eyelat *= 3600.0;
    lola2utm(eyelon, eyelat, &x, &y);
    path->eye[path->n][0] = (float)x;
    path->eye[path->n][1] = (float)y;

    reflon *= 3600.0;
    reflat *= 3600.0;
    lola2utm(reflon, reflat, &x, &y);
    path->ref[path->n][0] = (float)x;
    path->ref[path->n][1] = (float)y;

    path->n++;
  }
  f->close(f);
  wrap->file = NULL;
  wrap->type = TYPE_FLIGHTPATH;
  wrap->data = path;
  addWrap(wrap);
  return path;
}
