/*  tile.c -- overall control of tile loading and linking  */

#include "topo.h"

static void removeLR (Tile *deletedTile);



/**
 *  Allocates and initializes a tile structure for containment in the
 *  specified Wrap structure.  The tile ID is set from the wrap name.
 *  The size of a tile depends on the global variable "tileOrder".
 */
Tile *alcTile(Wrap *wrap) {
  Tile *tile = zalloc(sizeof(Tile) + (RECCOUNT - 1) * sizeof(TileRec));
  setID(tile, wrap->name);
  tile->wrap = wrap;
  return tile;
}



/**
 *  Sets the tid, xx, yy, and tname fields of a tile structure, given a URL.
 *  Returns the tile ID, or zero if the URL is of bad form.
 */
int setID(Tile *tile, char *url) {
  int tid = tileID(url);
  tile->tid = tid;
  tile->xx = XXX2UTM(TIDXXX(tid));
  tile->yy = YYY2UTM(TIDYYY(tid));
  synthName(tile->tname, TIDZONE(tid),TIDXXX(tid),TIDYYY(tid),TIDORDER(tid));
  return tile->tid;
}



/**
 *  Strips the header from a tile record and copies the data into the
 *  appropriate TileRec of a Tile struct.
 *
 *  Various sanity checks are performed, but replacing an existing
 *  record is specifically allowed.  Returns 1 if successful, 0 if not.
 */
int addRec(Tile *tile, void *buf, int len) {
  int dlen, flag, uflag;
  unsigned char *ubuf;
  TileRec *tr;

  LOCK(wrapLock);
  if( len < TV_RHLEN )
    return 0;
  if( strncmp(buf, "TV4F", 4) != 0 )
    return 0;

  ubuf = buf;
  dlen = (ubuf[5] << 16) | (ubuf[6] << 8) | (ubuf[7]);
  if( dlen > len )
    return 0;

  flag = ubuf[4];
  uflag = flag & ~TV_GZIP;
  if( uflag == TV_HEAD ) {
    tr = &HEADREC(tile);
  } else if( uflag >= TV_ELEV && uflag <= TV_ELEV + tileOrder ) {
    tr = &DATAREC(tile, uflag - TV_ELEV);
  } else if( uflag >= TV_ERRS && uflag <= TV_ERRS + 2 * tileOrder + 1) {
    tr = &ERROREC(tile, uflag - TV_ERRS);
  } else {
    return 0;
  }

  free(tr->tr_data);			/* in case we have previous data */
  tr->tr_flags = flag;
  tr->tr_length = dlen - TV_RHLEN;
  tr->tr_data = alloc(tr->tr_length);
  memcpy(tr->tr_data, (char *)buf + TV_RHLEN, tr->tr_length);

  UNLOCK(wrapLock);
  return 1;
}



/**
 *  Frees a Tile and all the associated data.
 */
void freeTile(Tile *tile) {
  int i;
  TileRec *r;

  if( tile == NULL )			/* allow and ignore if arg is null */
    return;

  LOCK(wrapLock);
  removeLR(tile);			/* unlink from neighbors */
  UNLOCK(wrapLock);

  free(tile->err);			/* free allocated data */
  free(tile->info);
  free(tile->data);
  tile->err = NULL;			/* zero pointers against illegal reuse*/
  tile->info = NULL;
  tile->data = NULL;
  tile->wrap = NULL;
  for( i = 0; i < RECCOUNT; i++ ) {
    r = &tile->trec[i];
    free(r->tr_data);			/* free allocated record data */
    r->tr_data = NULL;
    r->tr_length = 0;
  }
  free(tile);				/* free the tile struct itself */
}



/* removeLR
       This is a quick and dirty way to remove a tile from the L1, L2,
       R1, R2 pointer network. Should think of a better way as performance is
       not so good.
   The caller must hold the wrapLock mutex.
*/
static void removeLR (Tile *deletedTile)
{
  WrapGen *wg;
  Wrap *wrap;
  Tile *tile;

  RATTLE(wrapLock);
  wg = genWraps(TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    if (tile->L1 == deletedTile)
      tile->L1 = NULL;
    if (tile->L2 == deletedTile)
      tile->L2 = NULL;
    if (tile->R1 == deletedTile)
      tile->R1 = NULL;
    if (tile->R2 == deletedTile)
      tile->R2 = NULL;
  }
}



/**
 *  Reads a TopoVista tile and adds it to the Wrap list.
 *
 *  If the tile is not yet loaded (or is invalid),
 *  then load it and return 1 if successful.
 *  Otherwise, return 0.
 */
int getTile(char *name)
{
  int i, len, t;
  unsigned char header[TV_RHLEN];
  unsigned char *buf;
  Wrap *wrap;
  Tile *tile;
  inp f;

  /* check for an existing tile */
  LOCK(wrapLock);
  if( (wrap = findWrap(name, TYPE_ALL)) != NULL ) {
    t = wrap->type;
    if( (t == TYPE_TILE && !wrap->stale) || t == TYPE_UNAV ) {
      /* if good, existing tile, or if tile is known unavailable, do nothing */
      UNLOCK(wrapLock);
      return 0;
    } else {
      /* otherwise, delete old version */
      if( t <= TYPE_TILE )
        freeTile(wrap->data);
      delWrap(wrap);
    }
  }
  UNLOCK(wrapLock);

  /* open a new tile */
  wrap = fetch(name, TYPE_TILE);	/* locate source and create Wrap */
  if( wrap->file == NULL ) {		/* if no data, bail out */
    wrap->type = TYPE_UNAV;		/* mark as unavailable */
    addWrap(wrap);
    return 0;
  }
  f = unzip(wrap->file);		/* handle gzipped input */
  tile = wrap->data = alcTile(wrap);

  /* read and store tile records */
  while( f->rbytes(f, header, TV_RHLEN) == TV_RHLEN ) {
    if( memcmp(header, "TV4F", 4) != 0 )
      fatal("getTile: no TV4F header");
    i = header[4] & ~TV_GZIP;
    if( i >= TV_NFLAGS )
      fatal("getTile: bad sequence number");
    len = (header[5] << 16) + (header[6] << 8) + header[7];
    buf = alloc(len);
    memcpy(buf, header, TV_RHLEN);
    len -= TV_RHLEN;
    if( f->rbytes(f, buf + TV_RHLEN, len) < len )
      fatal("getTile: short read");
    addRec(tile, buf, len + TV_RHLEN);
    free(buf);
  }

  /* unpack and register the data */
  addWrap(wrap);
  if( ! buildTile(tile) )
    fatal("getTile: bad data");
  return 1;
}



/**
 *  Requests loading of a tile given its corner location.
 *  Loading is asynchronous with some protocols, and will
 *  not necessarily be complete when this function returns.
 */
int reqTileXY(int xxx, int yyy)
{
  char name[MAX_TILE_NAME];

  synthPath(name, utmZone, xxx, yyy, tileOrder);

  #ifndef UNITHREAD
    if( preskip(name, "tvtp:") ) {
      tvfetch(name);			/* load asynchronously */
      return 0;				/* return 0: not loaded yet */
    } else if( preskip(name, "tvrp:") ) {
      rpfetch(name);			/* load asynchronously */
      return 0;				/* return 0: not loaded yet */
    }
  #endif

  return getTile(name);
}



/**
 *  Adds a TYPE_WANTED entry to the wrap list for a tile that is
 *  to be loaded asynchronously.  Sets wrap->tile->tid, xx, and yy.
 */
Wrap *wantTile(char *name) {
  Tile *tile;
  Wrap *wrap;
  
  wrap = newWrap(name, TYPE_WANTED);
  tile = alcTile(wrap);
  wrap->data = tile;
  addWrap(wrap);
  return wrap;
}




/*
 *  Color a tile.
 */
void colorTile( Tile *tile )
{
  Transform *pnmll2xy;
  int i, x, y, px, py;
  float xmul, ymul;
  UV s, ll;
  Rgb *rgb;
  unsigned char *red, *green, *blue;

  /* fill in color layers that are available */
  tile->clayer = alloc(nLayers * sizeof(Rgb *));
  for( i=0; i<nLayers; ++i ) {
    rgb = initRGB(rows, cols);		/* init an rgb layer */
    tile->clayer[i] = rgb;
    if (rgb == NULL) {
      continue;
    }
    rgb->legend.size = pnmArray[i]->legend.size;
    rgb->legend.entry = pnmArray[i]->legend.entry;

    red = rgb->red;
    green = rgb->green;
    blue = rgb->blue;

    if (pnmArray[i]->ll2xy) {
      /* PNM file contained geographic coordinates */
      pnmll2xy = pnmArray[i]->ll2xy;
      for(y = 0; y < rows; ++y) {
	for(x = 0; x < cols; ++x) {
	  utm2lola(utmZone, x * mPerCell + tile->xx,
		   y * mPerCell + tile->yy, &ll.u, &ll.v);
	  transform(pnmll2xy, ll, &s);

	  if (s.u < pnmArray[i]->width && s.u >= 0 &&
	      s.v < pnmArray[i]->height && s.v >= 0) {
	    *red = PNMRED(pnmArray[i], s.u, s.v);
	    *green = PNMGREEN(pnmArray[i], s.u, s.v);
	    *blue = PNMBLUE(pnmArray[i], s.u, s.v);
	  }
	  ++red;
	  ++green;
	  ++blue;
	}
      }
    } else {
      /* PNM file lacked geographic coordinates */
      xmul = (pnmArray[i]->width + 1.0) / (cols + 1.0);
      ymul = (pnmArray[i]->height + 1.0) / (cols + 1.0);   
      for(y = 0; y < rows; ++y) {
	py = ymul * y;
	for(x = 0; x < cols; ++x) {
	  px = xmul * x;
	  *red++ = PNMRED(pnmArray[i], px, py);
	  *green++ = PNMGREEN(pnmArray[i], px, py);
	  *blue++ = PNMBLUE(pnmArray[i], px, py);
	}
      }
    }

  }
}



/*
 *  Find the tile corresponding to (xx, yy).
 */
Tile *tileCovering( double xx, double yy )
{
  Tile *tile;
  Wrap *wrap;
  WrapGen *wg;
  float x, y;

  LOCK(wrapLock);
  wg = genWraps(TYPE_TILE);
  while( (wrap = nextWrap(wg)) != NULL ) {
    tile = wrap->data;
    x = (xx - tile->xx) / mPerCell;
    y = (yy - tile->yy) / mPerCell;
    if( 0 <= x && x <= cols-1 && 0 <= y && y <= rows-1 ) {
      endGen(wg);
      UNLOCK(wrapLock);
      return tile;
    }
  }
  UNLOCK(wrapLock);
  return NULL;
}
