/*  tvrp.c -- asynchronous interface to TV Record Protocol  */

#ifndef UNITHREAD

#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include "topo.h"

static void issue(void);
static void init(char *name);
static void unav(Tile *tile, unsigned char *buf);
static void *request(void *dummy);
static void *listener(void *msgQueue);
static void *recorder(void *msgQueue);
static void *heartbeat(void *dummy);
static void *temporizer(void *dummy);
static void *newTile(void *tile);

typedef struct {			/* UDP data buffer struct */
  int nbytes;				/* number of bytes stored */
  unsigned char data[RP_MAXLEN];	/* message buffer */
} udpBuf;



static struct sockaddr_in servaddr;	/* server address */
static int udp = -1;			/* file descriptor for UDP port */

static int req_count;		/* number of requests in last message sent */
static int req_maxed;		/* set if last request limited by RP_MAXASK */
static time_t last_send;	/* timestamp of last message sent */


static int send_pending = 0;	/* is send pending? (locked by wrapLock) */

static int rcv_count;		/* number of msgs received since last send */



/*
 *  rpfetch(name) -- fetch file asynchronously using TVRP protocol.
 */
void rpfetch(char *name) {
  Wrap *wrap;

  if( udp < 0 )			/* if first call, start listener thread */
    init(name);

  LOCK(wrapLock);

  /* check for existing or in-progress tile */
  wrap = findWrap(name, TYPE_WANTED | TYPE_TILE | TYPE_UNAV);
  if( wrap != NULL ) {
    if( wrap->stale ) {
      freeTile(wrap->data);	/* delete old, stale tile */
      delWrap(wrap);
    } else {
      UNLOCK(wrapLock);		/* otherwise return if already have tile */
      return;
    }
  }

  /* add to list and send request */
  wrap = wantTile(name);
  issue();

  UNLOCK(wrapLock);

}



/*
 *  init(name) -- initialize port and threads.
 */
static void init(char *name) {
  url u;
  MsgQueue *mq;
  struct hostent *hostaddr;
  struct sockaddr_in sockaddr;
  pthread_t lthread;

  /* parse name */
  if( !parseURL(name, &u) )
    fatal("tvrp: unparseable");
  if( u.port == 0 )
    u.port = RP_PORT;

  /* look up host */
  hostaddr = gethostbyname(u.host);
  if( !hostaddr )
    fatal("tvrp: no hostaddr");

  /* create socket */
  if( (udp = socket(AF_INET, SOCK_DGRAM, 0)) < 0 )
    pfatal("tvrp socket");

  /* bind to port */
  memset(&sockaddr, 0, sizeof(sockaddr));
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_addr.s_addr = INADDR_ANY;
  sockaddr.sin_port = 0;
  if( bind(udp, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) < 0 )
    pfatal("tvrp bind");

  /* prepare address label */
  servaddr = sockaddr;
  servaddr.sin_port = htons(u.port);
  memcpy(&servaddr.sin_addr, hostaddr->h_addr, hostaddr->h_length);
 
  /* spawn heartbeat thread */
  pthread_create(&lthread, NULL, heartbeat, NULL);

  /* create message queue; spawn listener and recorder threads */
  mq = messageQueue();
  pthread_create(&lthread, NULL, listener, mq);
  pthread_create(&lthread, NULL, recorder, mq);

  /* spawn temporizer thread */
  pthread_create(&lthread, NULL, temporizer, NULL);
}



/*
 *  issue() -- send request to server.
 *
 *  Spawns a sender thread, if one is not already pending.
 */
static void issue(void) {
  pthread_t pth;

  if( !send_pending) {
    LOCK(wrapLock);
    if( !send_pending ){
      send_pending = 1;
      rcv_count = 0;
      pthread_create(&pth, NULL, request, NULL);
    }
    UNLOCK(wrapLock);
  }
}


/*
 *  request() -- transmit interest list to server.
 *
 *  Scans the wrap list and sends a TVRP REQUEST message to the server.
 *
 *  Parameters are compatible with pthread_create.  Normally run
 *  as a separate thread so that it can wait for the wrapLock.
 *  This in turn allows a batch of tiles to be added atomically
 *  so that the sender will then prioritize them all together.
 */
static void *request(void *dummy) {
  int i, w, pri;
  IntList *intlist;
  Interest *intr;
  Tile *tile;
  char buf[8 + 8 * RP_MAXASK], *p, *q;
  int nsent = 0;

  LOCK(wrapLock);
  intlist = interests();
  send_pending = 0;
  UNLOCK(wrapLock);

  memcpy(buf, "TVRP\0\0\0\0", 8);
  p = buf + 8;
  q = buf + sizeof(buf);

  for( i = 0; p < q && i < intlist->i_count; i++ ) {
    intr = &intlist->i_interest[i];
    tile = intr->i_tile;
    pri = intr->i_pri;
    w = (pri << 26) | (1 << intr->i_recn);
    nsent++;
    intr++;
    while( i+1 < intlist->i_count && intr->i_tile==tile && intr->i_pri==pri ) {
      i++;
      w |= (1 << intr->i_recn);
      nsent++;
      intr++;
    }
    pack4(p, tile->tid);	p += 4;
    pack4(p, w);		p += 4;
  }
  req_maxed = i < intlist->i_count;	/* true if stopped at RP_MAXASK */

  if( sendto(udp,buf,p-buf,0,(struct sockaddr *)&servaddr,sizeof(servaddr))<0 )
    pfatal("tvrp issue");

  if( nsent > 0 ) {
    intr = &intlist->i_interest[0];
    tile = intr->i_tile;
    fprintf(stderr,
      "  req %d/%d recs %s+, %ld bytes %08X %08X %08X %08X +\n",
      nsent, intlist->i_count, tile->tname, (long)(p - buf),
      unpack4(buf+8), unpack4(buf+12), unpack4(buf+16), unpack4(buf+20));
  } else
    fprintf(stderr, "  req nothing\n");

  req_count = nsent;
  last_send = time(NULL);
  free(intlist);
  return NULL;
}



/*
 *  heartbeat(dummy) -- asynchronous heartbeat sender.
 *
 *  This function runs asynchronously in its own thread to send a 
 *  heartbeat message, which is just a reiteration of the current
 *  request, if we have not sent anything recently.
 *
 *  "Recently" is RP_WAIT_HBT seconds if we are awaiting data or
 *  RP_IDLE_HBT seconds if we're just telling the server that we
 *  are still here.
 */
static void *heartbeat(void *dummy) {
  time_t now, idle, interval;

  for( ;; ) {
    sleep(1);
    now = time(NULL);
    idle = now - last_send;
    if( req_count == 0 )
      interval = RP_IDLE_HBT;
    else
      interval = RP_WAIT_HBT;
    if( idle >= interval )
      issue();
  }
}



/*
 *  listener(messageQueue) -- asynchronous UDP port listener.
 *
 *  This function runs asynchronously in its own high-priority thread and
 *  reads from the UDP port.  Each message is added to the message queue
 *  for the recorder to handle.
 */

static void *listener(void *messageQueue) {
  MsgQueue *mq = messageQueue;
  udpBuf *b = alloc(sizeof(*b));
  struct sched_param sp;
  int policy;

  pthread_getschedparam(pthread_self(), &policy, &sp);
  sp.sched_priority++;
  pthread_setschedparam(pthread_self(), policy, &sp);

  for( ;; ) {
    b->nbytes = recv(udp, b->data, sizeof(b->data), 0);
    if( b->nbytes < 0 )
      pfatal("tvrp recv");
    if( b->nbytes < RP_PHLEN || memcmp(b->data, "TVIN", 4) != 0 )
      fatal("tvrp: corrupt message");
    sendMessage(mq, b);
    b = alloc(sizeof(*b));
  }
}


/*
 *  recorder(messageQueue) -- asynchronous data filer.
 *
 *  This function runs asynchronously in its own thread and processes
 *  datagrams placed in the message queue by the listener. 
 *  When data accumulation is complete for a tile, 
 *  it spawns a transient thread to install the tile.
 */

static void *recorder(void *messageQueue) {
  int id, k, n, xxx, yyy, zone, rcv;
  Tile *tile;
  Wrap *wrap;
  pthread_t pth;
  char name[MAX_TILE_NAME];
  MsgQueue *mq = messageQueue;
  udpBuf *b;
  unsigned char *buf;

  for( ;; ) {

    /* get next datagram */
    b = rcvMessage(mq);
    n = b->nbytes;
    buf = b->data;

    if( ++rcv_count >= RP_REFRESH && req_maxed )
      issue();

    /* identify the tile */
    id = unpack4(buf + 4);
    k = TIDORDER(id);
    zone = TIDZONE(id);
    xxx = TIDXXX(id);
    yyy = TIDYYY(id);
    synthPath(name, zone, xxx, yyy, k);

    /* find wrap entry, or make one if never before seen */
    LOCK(wrapLock);
    wrap = findWrap(name, TYPE_WANTED | TYPE_TILE | TYPE_UNAV);
    if( wrap == NULL || wrap->type == TYPE_UNAV )
      wrap = wantTile(name);		/* make new wrap list entry */
    tile = wrap->data;

    /* handle message */
    if( (buf[8] & 0x80) == 0 )		/* if message contains data */
      addRec(tile, buf + RP_PHLEN, n - RP_PHLEN);
    else
      unav(tile, buf);

    /* update record status and spawn thread if data is now complete */
    rcv = unpack4(buf + 8) & 0x03FFFFFF;
    tile->rcvmask |= rcv;
    if( tile->rcvmask == 0x03FFFFFF) {	/* if have received all */
      tile->waiting = -1;		/* mark as complete */
      pthread_create(&pth, NULL, newTile, tile);
      pthread_detach(pth);
    } else if( rcv & 0x0FF0000 ) {	/* if received data (not hdr/err rec) */
      tile->waiting = 1;		/* start/reset stall timer */
    }

    UNLOCK(wrapLock);

    free(b);				/* free allocated message */
    postRedisplays(RO);			/* update status discs */
  }
}



/*
 * Marks tile records as unavailable.
 */
static void unav(Tile *tile, unsigned char *buf) {
    int i, n, mask;

    mask = unpack4(buf + 8);
    if( mask & 0x02000000 )
      HEADREC(tile).tr_unav = 1;
    n = MIN(8, tileOrder);
    for( i = 0; i <= n; i++ )
      if( mask & (1 << (24 - i)) )
        DATAREC(tile, i).tr_unav = 1;
    n = MIN(15, 2 * tileOrder + 1);
    for( i = 0; i <= n; i++ )
      if( mask & (1 << (15 - i)) )
        ERROREC(tile, i).tr_unav = 1;
}



/*
 *  Adds a new tile and rebuilds the relief map.
 *  Parameters are compatible with pthread_create.
 */
void *newTile(void *arg) {

  Tile *tile = arg;
  Wrap *wrap = tile->wrap;
  LOCK(wrapLock);
  if( buildTile(tile) ) {
    postRedisplays(NR + NS + RO + RP);
  } else {
    if( tile->rcvmask == 0x03FFFFFF && tile->dlevels == 0 )
      wrap->type = TYPE_UNAV;
    postRedisplays(RO);
  }
  tile->waiting = 0;
  UNLOCK(wrapLock);
  return NULL;
}



/*
 *  temporizer(dummy) -- renderer of partial tiles.
 *
 *  This function runs asynchronously in its own thread and renders
 *  partially received tiles that seem to have stalled.
 *  A tile is rendered if we have received actual data (not just
 *  error records) since the last rendering, but not in the last
 *  RP_STALLED seconds.
 */
static void *temporizer(void *dummy) {
  WrapGen *wg;
  Wrap *wrap;
  Tile *tile;

  for( ;; ) {
    sleep(1);
    LOCK(wrapLock);
    wg = genWraps(TYPE_WANTED | TYPE_TILE);
    while( (wrap = nextWrap(wg)) != NULL ) {
      tile = wrap->data;
      if( tile->waiting > 0 && tile->waiting++ > RP_STALLED ) {
        if( buildTile(tile) ) {
          postRedisplays(NR + NS + RO + RP);
	}
	tile->waiting = 1;
      }
    }
    UNLOCK(wrapLock);
  }
}



#endif  /* UNITHREAD */
