/*  utils.c -- general-purpose utility functions  */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "basic.h"



/*  fatal(msg) -- abort with error message.  */

void fatal(char *msg) {
  fprintf(stderr, "  fatal error: %s\n", msg);
  exit(1);
}



/*  pfatal(msg) -- call perror(msg) and abort.  */

void pfatal(char *msg) {
  perror(msg);
  exit(1);
}



/*  nospace() -- abort with "out of memory" message.  */

void nospace(void) {
  fatal("out of memory");
}



/*  alloc(nbytes) -- allocate memory, aborting if unavailable.  */

void *alloc(size_t nbytes) {
  void *p = malloc(nbytes);
  if (!p)
    nospace();
  return p;
}



/*  zalloc(nbytes) -- allocate and zero memory, aborting if unavailable.  */

void *zalloc(size_t nbytes) {
  void *p = alloc(nbytes);
  memset(p, 0, nbytes);
  return p;
}



/*  salloc(s) -- allocate memory and initialize with string s.  */

char *salloc(char *s) {
  size_t n = strlen(s) + 1;
  char *t = alloc(n);
  memcpy(t, s, n);
  return t;
}



/*  pack4(buf, n) -- pack integer into 4-byte buffer, network order  */

char *pack4(void *buf, int n) {
  unsigned char *ubuf = buf;
  ubuf[3] = n;  n >>= 8;
  ubuf[2] = n;  n >>= 8;
  ubuf[1] = n;  n >>= 8;
  ubuf[0] = n;
  return buf;
}



/*  unpack4(buf) -- unpack integer from 4-byte buffer, network order  */

int unpack4(void *buf) {
  unsigned char *ubuf = buf;
  return (ubuf[0] << 24) | (ubuf[1] << 16) | (ubuf[2] << 8) | ubuf[3];
}



/*
 *  preskip(s, prefix) -- skip prefix of s, case insensitive.
 *
 *  Returns pointer s advanced beyond prefix.
 *  Returns NULL if prefix does not match.
 */
char *preskip(char *s, char *prefix) {
  while (*prefix != '\0') {
    if (tolower(*prefix++) != tolower(*s++))
      return NULL;
  }
  return s;
}



/*
 *  tileID(s) -- return the ID of a tile, given its name, path, or URL.
 *
 *  Returns zero if it can't parse the string.
 */

unsigned int tileID(char *s) {
  char z, c, *p;
  int zone, xxx, yyy, k;

  if( (p = strrchr(s, '/')) != NULL )
    s = p + 1;
  if( sscanf(s, "%d%1c%d%1c", &xxx, &z, &yyy, &c) != 4 )
    return 0;
  zone = z - 'A' + 1;
  k = (c - 'a' + 1) - 6;
  if( k < 0 || k > 7 || zone < 1 || zone > 26 )
    return 0;
  return (k << 29) | (zone << 24) | (xxx << 14) | yyy;
}



/* Under Windows this gets the clock time using _ftime(). */
/* Under UNIX this gets the resource usage time using getrusage(). */
double getTime(void)
{
#ifdef WIN32
  struct _timeb tstruct;

  _ftime( &tstruct );
  return( (double)tstruct.time + (double)tstruct.millitm/1000.);
#else
  struct rusage r;

  getrusage(RUSAGE_SELF,&r);
  return((double)r.ru_utime.tv_sec + (double)r.ru_utime.tv_usec/1000000.);
#endif
}
