/*  view.c -- display perspective view  */

#ifdef WIN32
  #include <windows.h>
#endif

#include "topo.h"

static void pspDisplayCB(void);
static void idleWorker(void);

static double angle = 45.0;

static double aspect;
static GLdouble modelMatrix[16] = {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};
static GLdouble projMatrix[16];
static GLint viewport[4];

static int rdpFlags = -1;	/* redisplay flags */
static Lock rdpLock;		/* redisplay mutex */



static int proj2plane(GLdouble x, GLdouble y, GLdouble z, Coord2f xy)
{
  double t = -1.0;

  if( z != eye[2] )
    t = eye[2] / (eye[2] - z);
  if( t < 0 ) {
    xy[0] = x; xy[1] = y;
    return 0;
  }
  xy[0] = eye[0] + t * (x - eye[0]);
  xy[1] = eye[1] + t * (y - eye[1]);
  return 1;
}



/* set the projection matrix for the perspective view */

static void setPspProjMatrix( void )
{
  double s, n, e, w, afar;

  s = yymin - eye[1]; s = s * s;
  n = yymax - eye[1]; n = n * n;
  e = xxmax - eye[0]; e = e * e;
  w = xxmin - eye[0]; w = w * w;
  afar = MAX(s+w, s+e);
  afar = MAX(afar, n+e);
  afar = MAX(afar, n+w);
  afar = sqrt(afar);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  gluPerspective((GLdouble)(angle/aspect),(GLdouble)aspect,
		 /* fore clipping plane - needs help */
		 20.0,	/* empirical -- mar2002/gmt */
		 (GLdouble)afar);

  gluLookAt((GLdouble)eye[0], (GLdouble)eye[1], (GLdouble)eye[2],
	    (GLdouble)ref[0], (GLdouble)ref[1], (GLdouble)ref[2],
	    (GLdouble)0, (GLdouble)0, (GLdouble)1);
}



void setEyeRefHeight(void)
{
  int x, y, z;
  Tile *eyeTile, *refTile;

  LOCK(wrapLock);
  eyeTile = tileCovering(eye[0],eye[1]);
  refTile = tileCovering(ref[0],ref[1]);

  /* compute eye height depending on location and eyeIsRelative flag */
  if( eyeIsRelative ) {
    if( eyeTile ) {
      /* eye down: just above max height of cell corners */
      x = (eye[0] - eyeTile->xx) / mPerCell;
      y = (eye[1] - eyeTile->yy) / mPerCell;
      z = zval(eyeTile, x, y);
      z = MAX(z, zval(eyeTile, x + 1, y));
      z = MAX(z, zval(eyeTile, x + 1, y + 1));
      z = MAX(z, zval(eyeTile, x, y + 1));
    } else {
      z = zzmin;
    }
    eye[2] = vexagger / 10.0 * (z + eyeZRelative);
  } else {
    /* eye is absolute */
    eye[2] = vexagger / 10.0 * eyeZAbsolute;
  }

  /* compute reference height */
  if( refTile ) {
    x = (ref[0] - refTile->xx) / mPerCell;
    y = (ref[1] - refTile->yy) / mPerCell;
    z = zval(refTile, x, y);
  } else {
    z = zzmin;
  }
  ref[2] = vexagger / 10.0 * (z + REF_HEIGHT);
  UNLOCK(wrapLock);
}



/*  buildSurface() -- rebuild the surface */

void buildSurface(void)
{
  GLdouble x, y, z;
  GLdouble origProjMatrix[16];

  if( !redoSurface ) return;

  /* Calculate projection matrix for FOV determination */
  /* save original proj matrix */
  glGetDoublev(GL_PROJECTION_MATRIX, origProjMatrix);
  /* set proj matrix to perspective */
  setPspProjMatrix();
  glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);
  /* restore saved matrix */
  glMatrixMode(GL_PROJECTION);
  glLoadMatrixd(origProjMatrix);

  gluUnProject(0.0, viewport[3]-1, 1.0,
	       modelMatrix, projMatrix, viewport, &x, &y, &z);
  if( proj2plane(x, y, z, fovULpt) ) {
    /* Ray from eye through upper left corner of view window hits z=0 plane. */
    /* This means FOV is finite, either 4gon or 5gon. */
    gluUnProject(viewport[2]-1, viewport[3]-1, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovURpt);
    gluUnProject(0.0, 0.0, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovLLpt);
    gluUnProject(viewport[2]-1, 0.0, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovLRpt);

    fovTnorm[0] = fovULpt[1] - fovURpt[1];
    fovTnorm[1] = fovURpt[0] - fovULpt[0];
    fovLnorm[0] = fovLLpt[1] - fovULpt[1];
    fovLnorm[1] = fovULpt[0] - fovLLpt[0];
    fovRnorm[0] = fovURpt[1] - fovLRpt[1];
    fovRnorm[1] = fovLRpt[0] - fovURpt[0];
    if(LEFT_TURN(eye[0],eye[1],fovLRpt[0],fovLRpt[1],fovLLpt[0],fovLLpt[1])) {
      fovType = FIVE_GON;
      fovBLnorm[0] = eye[1] - fovLLpt[1];
      fovBLnorm[1] = fovLLpt[0] - eye[0];
      fovBRnorm[0] = fovLRpt[1] - eye[1];
      fovBRnorm[1] = eye[0] - fovLRpt[0];
    } else {
      fovType = FOUR_GON;
      fovBLnorm[0] = fovLRpt[1] - fovLLpt[1];
      fovBLnorm[1] = fovLLpt[0] - fovLRpt[0];
    }
  } else {
    fovType = WEDGE;
    gluUnProject(0.0, 0.0, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovLLpt);
    fovLnorm[0] = eye[1] - fovLLpt[1];
    fovLnorm[1] = fovLLpt[0] - eye[0];
    gluUnProject(viewport[2]-1, 0.0, 1.0,
		 modelMatrix, projMatrix, viewport, &x, &y, &z);
    proj2plane(x, y, z, fovLRpt);
    fovRnorm[0] = fovLRpt[1] - eye[1];
    fovRnorm[1] = eye[0] - fovLRpt[0];
  }

  /* build the surface */
  errSurface();

  redoSurface = 0;
}



/*
 *  Set flags for rebuilding surface and refreshing displays.
 *  This function can be called from any thread, but the actual
 *  work is deferred to idleWorker, below, which is called from
 *  the GLUT main loop.  This avoids race conditions involving
 *  global state such as the "current window".
 *
 *  If called in the timing loop, however, (doingLoop != 0),
 *  then the work is done right here.
 */
void postRedisplays( int refresh )
{
  if( rdpFlags < 0 ) {		/* first-time initialization only */
    INITLOCK(rdpLock);
    rdpFlags = 0;
    glutIdleFunc(idleWorker);
  }

  #ifdef TESTING
    if( doingLoop ) {		/* if running a timing loop */
      redoSurface = 1;
      buildSurface();
      if( refresh & RO ) { glutSetWindow(ovhWin); ovhDisplayCB(); }
      if( refresh & RP ) { glutSetWindow(pspWin); pspDisplayCB(); }
      triPerLoop += numSurfTri;
      return;
    }
  #endif

  if( refresh & NS ) {		/* NS flag sets redoSurface flag */
    redoSurface = 1;
  }
  if( refresh & NR ) {		/* NR flag sets redoRelief flag */
    redoRelief = 1;
  }

  LOCK(rdpLock);
  rdpFlags |= refresh;		/* others are saved for idleWorker */
  UNLOCK(rdpLock);
}



/*
 *  This function is called from the GLUT idle loop.
 *  It's the one that actually posts GLUT redisplays.
 */
static void idleWorker(void) {
  int refresh;
  struct timespec ts;

  LOCK(rdpLock);
  refresh = rdpFlags;		/* remember flag values */
  rdpFlags = 0;			/* and clear them */
  UNLOCK(rdpLock);

  /* post GLUT redisplays according to flags set by postRedisplays() */
  if( refresh & RC ) { glutSetWindow(ctlWin); glutPostRedisplay(); }
  if( refresh & RO ) { glutSetWindow(ovhWin); glutPostRedisplay(); }
  if( refresh & RP ) { glutSetWindow(pspWin); glutPostRedisplay(); }

  ts.tv_sec = 0;
  ts.tv_nsec = 1;
  nanosleep(&ts, NULL);		/* don't spin the CPU in this thread */
}



static void pspReshapeCB(int w, int h)
{
  glViewport(0,0,w,h);
  glGetIntegerv(GL_VIEWPORT, viewport);
  aspect = (double)w / (double)h;
}



/*  pspDisplayCB -- redraw perspective (3-D) view  */

static void pspDisplayCB(void)
{
  double eyelon, eyelat, reflon, reflat;

  /* Record flight here so as to capture only those eye/ref positions
   * corresponding to displayed views.
   */
  if( flightRecord ) {
    /* store eye height as meters */
    if( eyeIsRelative ) {
      fprintf(flightRecord, "%d +%d ", 100, eyeZRelative / 10);
    } else {
      fprintf(flightRecord, "%d %d ", 100, eyeZAbsolute / 10);
    }
    utm2lola(utmZone, eye[0], eye[1], &eyelon, &eyelat);
    utm2lola(utmZone, ref[0], ref[1], &reflon, &reflat);
    
    fprintf(flightRecord, "%f %f %f %f\n",
	    eyelat/3600.0, eyelon/3600.0, reflat/3600.0, reflon/3600.0);
  }

  LOCK(wrapLock);
  buildSurface();
  setPspProjMatrix();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  renderPsp();
  UNLOCK(wrapLock);

  glutSwapBuffers();
}



/* renderPsp - used by pspDisplayCB and outputEPS */

void renderPsp(void)
{
  if (threeDMesh) {
    glLineWidth(1.0);
    emitSurface(WIRE3D);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0,1.0);
  }

  emitSurface(FILL3D);

  /*modified by Xiangyu Zhang, for path rendering*/
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glScalef(1.0, 1.0, vexagger / 10.0);
  emitPaths();

  if (threeDMesh) {
    glDisable(GL_POLYGON_OFFSET_FILL);
  }
}



/*  zval(tile, x, y) -- return z value from tile, or 0 if out of bounds  */

int zval(Tile *tile, int x, int y)
{
  if (x < 0 || y < 0 || x > cols || y > rows)
    return 0;
  else
    return EMDATA(tile, x, y);
}



/*  openPspWin() -- open perspective view window  */

void openPspWin(void)
{
  GLfloat light_specular[] = { 0.2F, 0.2F, 0.2F, 1.0F };
  GLfloat light_diffuse[] = { 1.0F, 1.0F, 1.0F, 1.0F };
  GLfloat light_ambient[] = { 0.0F, 0.0F, 0.0F, 1.0F };
  GLfloat light_position[] = {0.0F, 1.0F, 2.0F, 0.0F};/*high in northern sky*/

  aspect = (double)WIDTH_3D / (double)HEIGHT_3D;
  viewport[0] = viewport[1] = 0;
  viewport[2] = WIDTH_3D;
  viewport[3] = HEIGHT_3D;

  glutInitWindowSize(WIDTH_3D, HEIGHT_3D);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
  pspWin = glutCreateWindow("perspective");

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  /*  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);*/
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);

  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  glCullFace(GL_BACK);		/* enabled/disabled in control.c */

  glEnable(GL_NORMALIZE);	/* so OpenGL normalizes normal vectors */

  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);
  glClearColor(0.0,0.0,1.0,0.0);
  /*  glClearColor(0.9,0.9,0.9,0.0);*/
  /*  glClearDepth((GLclampd)afar);*/

  glutDisplayFunc(pspDisplayCB);
  glutReshapeFunc(pspReshapeCB);
  glutKeyboardFunc(keyCB);
  glutSpecialFunc(specialCB);
}
