/*  wrap.c -- Wrap list management  */

/*
 *  All access to the Wrap list is via the functions in here.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "topo.h"



struct WrapGen {		/* opaque structure outside this file */
  int mask;
  Wrap *nextWrap;
  WrapGen *nextGen;
};

static Wrap *wlist = NULL;	/* master wrap list */
static WrapGen *glist = NULL;	/* active generator list */



/*
 *  newWrap(name, type) -- allocate and initialize a Wrap.
 *
 *  Returns a newly initialized Wrap struct of the given type.
 *  The specified name is copied to an allocated string.
 *
 *  The returned Wrap has NOT been put on the wrap list.
 */
Wrap *newWrap(char *name, int type) {
  Wrap *w;

  w = zalloc(sizeof(Wrap));
  w->type = type;
  if( name != NULL )
    w->name = salloc(name);
  return w;
}



/*
 *  addWrap(w) -- add a Wrap to the list.
 *
 *  Adds a Wrap, which is assumed valid, to the Wrap list.
 *
 *  Returns w.
 */
Wrap *addWrap(Wrap *w) {

  LOCK(wrapLock);
  w->wnext = wlist;
  wlist = w;
  UNLOCK(wrapLock);
  return w;
}



/*
 *  findWrap(name, type) -- find a Wrap by name and mask.
 *
 *  The caller must hold the wrapLock mutex.
 *
 *  Returns NULL if not found.
 */
Wrap *findWrap(char *name, int mask) {
  Wrap *w;
  WrapGen *wg;

  RATTLE(wrapLock);
  wg = genWraps(mask);
  while( (w = nextWrap(wg)) != NULL ) {
    if( w->name != NULL && strcmp(w->name, name) == 0 ) {
      endGen(wg);
      return w;
    }
  }
  return NULL;
}



/*
 *  delWrap(w) -- delete a Wrap.
 *
 *  Removes a Wrap from the Wrap list and from any active generators.
 *  Does not free any storage (see destroy() for that).
 *
 *  The caller must hold the wrapLock mutex.
 *
 *  Returns NULL.
 */
void *delWrap(Wrap *w) {
  Wrap **p;
  WrapGen *wg;

  RATTLE(wrapLock);

  /* remove from all generators */
  for( wg = glist; wg != NULL; wg = wg->nextGen )
    if( wg->nextWrap == w )
      wg->nextWrap = w->wnext;

  /* remove from wrap list */
  for( p = &wlist; *p != NULL; p = &(*p)->wnext )
    if( *p == w ) {
      *p = w->wnext;
      break;
    }

  #ifdef TSLOOKUP
    if( w->token )
      Tok_Release(w->token);
  #endif

  return NULL;
}



/*
 *  genWraps(mask) -- create Wrap generator.
 *
 *  Returns a WrapGen pointer through which successive members of the
 *  wrap list having the specified type(s) can be accessed.  Usage is:
 *
 *    WrapGen *wg = genWraps(mask);
 *    while( (Wrap w = nextWrap(wg)) != NULL ) {
 *      // process Wrap w //
 *    }
 *
 *  The caller must hold the wrapLock mutex.
 *
 *  If mask is TYPE_ALL, all list members match.
 *
 *  If it is necessary to break out of the loop before exhausting
 *  the list (before NULL is returned), call endGen(wg) to free
 *  the WrapGen struct and avoid a memory leak.
 */
WrapGen *genWraps(int mask) {
  WrapGen *wg;

  RATTLE(wrapLock);
  wg = alloc(sizeof(WrapGen));
  wg->mask = mask;
  wg->nextWrap = wlist;
  wg->nextGen = glist;
  glist = wg;
  return wg;
}



/*
 *  nextWrap(wg) -- return next Wrap list member.
 *
 *  The caller must hold the wrapLock mutex.
 *
 *  Returns null (and calls endGen()) when the list is exhausted.
 */
Wrap *nextWrap(WrapGen *wg) {
  Wrap *w;

  RATTLE(wrapLock);
  while( (w = wg->nextWrap) != NULL ) {
    wg->nextWrap = w->wnext;
    if( w->type & wg->mask )
      break;
  }
  if( w == NULL )
    endGen(wg);
  return w;
}



/*
 *  endGen(wg) -- terminate generator.
 *
 *  Frees the memory associated with a Wrap generator.
 *  Should only be called once, and only before nextWrap returns NULL.
 *
 *  The caller must hold the wrapLock mutex.
 */
void endGen(WrapGen *wg) {
  WrapGen **p;

  RATTLE(wrapLock);
  for( p = &glist; *p != NULL; p = &(*p)->nextGen )
    if( *p == wg ) {
      *p = wg->nextGen;
      break;
    }
  wg->nextWrap = NULL;
  wg->nextGen = NULL;
  free(wg);
}
