package topovista.explorer;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;



/**
 *  A Plat object draws a map of known grid cells.
 */

class Plat extends JPanel {

    World world;	// world being platted
    Cell locn;		// current location

    int xmid, ymid;	// grid coordinates of center
    int xc, yc;		// screen coordinates of center

    int csz;		// cell size
    int csp;		// cell spacing

    boolean needReconfig;	// reconfiguration needed?


final static int MAX_SPACING = 10;
final static float GAP_RATIO = 0.25f;
final static float GROWTH_FACTOR = 1.5f;

public final static Color HASDATA = new Color(96, 240, 160);
public final static Color EXPLORED = new Color(0, 0, 128);
public final static Color VISITED = new Color(192, 192, 224);
public final static Color SCOUTED = new Color(255, 192, 192);

final static Color CURRENT = new Color(255, 0, 0);
final static Stroke wideStroke = new BasicStroke(3f);



/**
 *  Constructs a Plat, which must later be configured before use.
 */
public Plat() {

    addComponentListener(new ComponentAdapter() {
	public void componentResized(ComponentEvent e) {
	    configure(world);
	}
    });
}



/**
 *  Configures the Plat according to the current window size
 *  and World membership.
 */
public void configure(World wd) {

    if (wd == null) {			// if not yet configured
	return;
    }

    world = wd;

    Cell[] list = world.all();
    Cell ce = list[0];
    int xmin = ce.x;
    int xmax = ce.x;
    int ymin = ce.y;
    int ymax = ce.y;

    for (int i = 1; i < list.length; i++) {
	ce = list[i];
	if (xmin > ce.x)  { xmin = ce.x; }
	if (xmax < ce.x)  { xmax = ce.x; }
	if (ymin > ce.y)  { ymin = ce.y; }
	if (ymax < ce.y)  { ymax = ce.y; }
    }

    xmid = (xmin + xmax) / 2;
    ymid = (ymin + ymax) / 2;

    Insets insets = getInsets();
    int w = getWidth() - insets.left - insets.right;
    int h = getHeight() - insets.top - insets.bottom;

    int xspacing = (int) (w / (xmax - xmin + 1) / GROWTH_FACTOR);
    int yspacing = (int) (h / (ymax - ymin + 1) / GROWTH_FACTOR);

    csp = MAX_SPACING;
    if (csp > xspacing)	{ csp = xspacing; }
    if (csp > yspacing)	{ csp = yspacing; }
    if (csp < 1)	{ csp = 1; }

    csz = csp - (int) (GAP_RATIO * csp);

    needReconfig = false;
}



/**
 *  Records a Cell as the current location, for special highlighting.
 */
public void setCurrent(Cell ce) {
    locn = ce;
}



/**
 *  Completely redraws the entire Plat.
 */
public void paintComponent(Graphics g) {

    super.paintComponent(g);			// repaint border

    if (world == null) {			// if not yet configured
	return;
    }
    if (needReconfig) {				// if need to reconfigure
	configure(world);
    }

    Insets insets = getInsets();
    int x0 = insets.left;
    int y0 = insets.top;
    int xlim = getWidth() - insets.right;
    int ylim = getHeight() - insets.bottom;
    int w = xlim - x0;
    int h = ylim - y0;
    xc = x0 + w / 2 - csp / 2;
    yc = y0 + h / 2 + csp / 2;

    Cell[] list = world.all();
    for (int i = 0; i < list.length; i++) {
	Cell ce = list[i];
	if (ce.hasdata) {
	    if (ce.visited) {
		g.setColor(EXPLORED);
	    } else {
		g.setColor(HASDATA);
	    }
	} else {
	    if (ce.visited) {
		g.setColor(VISITED);
	    } else {
		g.setColor(SCOUTED);
	    }
	}
	int x = xc + csp * (ce.x - xmid);
	int y = yc - csp * (ce.y - ymid);

	if (x >= x0 && y >= y0 && x + csz < xlim && y + csz < ylim) {
	    g.fillRect(x, y, csz, csz);
	} else if (csp > 1) {
	    needReconfig = true;
	}
    }

    Graphics2D g2 = (Graphics2D) g;

    if (locn != null) {
	int x = csp * (locn.x - xmid);
	int y = csp * (locn.y - ymid);
	g2.setColor(CURRENT);
	g2.setStroke(wideStroke);
	g2.setRenderingHint(
	    RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	g2.drawArc(xc + x - csz, yc - y - csz, 3 * csz, 3 * csz, 0, 360);
    }

}



} // class Plat
