package topovista.grid;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import topovista.util.*;



/**
 *  The TileGrid class reads and writes elevation grids in
 *  TopoVista format (Version 4F).  Error trees are synthesized
 *  on output and ignored on input.
 */

public class TileGrid extends Grid {



public final static String MAGIC = "TV4F";

public final static int TVR_HEADER = 0x10;	// header flag
public final static int TVR_ELEVATION = 0x20;	// elevation flag
public final static int TVR_ERROR = 0x40;	// error flag
public final static int TVR_GZIP = 0x80;	// gzipped data flag

public final static int X_ORIGIN = UTM_FALSE_EASTING + 10;
public final static int X_OFFSET = 500;



private static class TVRecord {		// one record from the input file:
    int flag;				// flag byte value, excluding gzip flag
    int gzflag;				// gzip flag bit
    int length;				// record length as encoded in file
    TypedInputStream tin;		// data stream
};



/**
 *  Constructs the "xxxUyyyc" form of a tile name.
 */
public static String makeName(int zone, int xxx, int yyy, int k) {
    return "" + xxx + (char)('A' + zone - 1) + yyy + (char)('a' + k - 1);
}



/**
 *  Constructs the "xxxUyyyc" form of a tile name from a tile ID.
 */
public static String makeName(int tid) {
    int[] a = parseID(tid);
    return makeName(a[0], a[1], a[2], a[3]);
}



/**
 *  Parses the "xxxUyyyc" form of a tile name,
 *  returning an int array of [zone, xxx, yyy, k].
 *  A leading protocol and path, and a trailing extension,
 *  are allowed and ignored.
 */
public static int[] parseName(String s) {
    try {
	s = s.substring(s.indexOf(':') + 1);
	s = s.substring(s.lastIndexOf('/') + 1);
	int i = s.lastIndexOf('.');
	if (i >= 0) {
	    s = s.substring(0, i);
	}
	StringTokenizer tk = new StringTokenizer(
	    s, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ", true);
	int xxx = Integer.parseInt(tk.nextToken());
	int zone = tk.nextToken().charAt(0) - 'A' + 1;
	int yyy = Integer.parseInt(tk.nextToken());
	int k = tk.nextToken().charAt(0) - 'a' + 1;
	if (xxx > 0 && yyy >= 0 && zone > 0 && zone < 27 && k > 0 && k < 27
	    && !tk.hasMoreTokens()) {
		return new int[] { zone, xxx, yyy, k };
	}
    } catch (Throwable t) {
    }
    throw new RuntimeException("can't parse tile name " + s);
}



/**
 *  Constructs a 32-bit tile ID as used by the TVTP protocol.
 */
public static int makeID(int zone, int xxx, int yyy, int k) {
    return ((k - 6) << 29) + ((zone % 32) << 24) + (xxx << 14) + (yyy << 0);
}



/**
 *  Constructs a 32-bit tile ID from a name in "xxxUyyyc" form.
 */
public static int makeID(String s) {
    int vals[] = parseName(s);
    return makeID(vals[0], vals[1], vals[2], vals[3]);
}



/**
 *  Converts a 32-bit tile ID to an 8-character uppercase hexadecimal string.
 */
public static String hexID(int id) {
    long n = ((long) id) & 0xFFFFFFFFL;
    return Long.toHexString(0x100000000L + n).substring(1).toUpperCase();
}



/**
 *  Parses a tile ID, returning an int array of [zone, xxx, yyy, k].
 */
public static int[] parseID(int id) {
    int k = ((id >> 29) & 7) + 6;
    int zone = (id >> 24) & 0x1F;
    int xxx = (id >> 14) & 0x03FF;
    int yyy = id & 0x3FFF;
    return new int[] { zone, xxx, yyy, k };
}



/**
 *  Reads elevation information from a TopoVista grid.
 */
TileGrid(InputStream ins, boolean partial) throws IOException {

    DataInputStream din = new DataInputStream(ins);

    TVRecord tvr = readRecord(din);
    TypedInputStream tin = tvr.tin;
    String t = tin.readString();	// 1. title
    String v = tin.readString();	// 2. version
    String c = tin.readString();	// 3. comments
    int r = tin.readInt();		// 4. resolution
    int z = tin.readInt();		// 5. zone
    int d = tin.readInt();		// 6. datum
    int k = tin.readInt();		// 7. order
    int e = tin.readInt();		// 8. error tree levels
    int xxx = tin.readInt();		// 9. SW x
    int yyy = tin.readInt();		// 10. SW y
    int g = tin.readInt();		// 11. elevation granularity
    int l = tin.readInt();		// 12. minimum elevation
    int h = tin.readInt();		// 13. maximum elevation

    int edge = 1 << k;
    rows = cols = edge + 1;
    int gridw = r * (rows - 1);

    refsys = UTM_SYSTEM;
    datum = d;
    zone = z;
    xyUnit = 2;
    treelevels = e;

    if (g == 305) {
	zUnit = FEET;
	zRes = 1;
	zMin = l;
	zMax = h;
    } else {
	zUnit = METERS;
	zRes = g / 1000.0;
	zMin = l * zRes;
	zMax = h * zRes;
    }

    xMin = xLL = X_ORIGIN + gridw * (xxx - X_OFFSET);
    yMin = yLL = gridw * yyy;
    xMax = xLL + gridw;
    yMax = yLL + gridw;
    xRes = yRes = r;

    header = "TopoVista " + r + "m grid " + makeName(z, xxx, yyy, k) +
	" (ID=" + hexID(makeID(z, xxx, yyy, k)) + "), " +
	((cols - 1) * r) / 1000.0 + " km square";
    freeform = "";

    if (partial) {
	return;
    }

    alloc();

    // loop over data records to load elevation data

    int dx = 1 << k;
    while ((tvr = readRecord(din)) != null) {
	if (tvr.flag < TVR_ELEVATION || tvr.flag >= TVR_ERROR) {
	    continue;			// ignore all but elevation data
	}
	tin = tvr.tin;
	int i = tvr.flag - TVR_ELEVATION;

	if (i == 0) {
	    elev[0][0] = (char) (tin.readInt() + ELEVATION_BIAS);
	    elev[0][edge] = (char) (tin.readInt() + ELEVATION_BIAS);
	    elev[edge][0] = (char) (tin.readInt() + ELEVATION_BIAS);
	    elev[edge][edge] = (char) (tin.readInt() + ELEVATION_BIAS);
	} else {
	    dx = 1 << (k - i);
	    readInterp(tin, dx, 0, 2 * dx, 2 * dx, edge);
	    readInterp(tin, 0, dx, dx, 2 * dx, edge);
	}
    }

    // interpolate to compensate for any missing trailing elevation records
    while (dx > 1) {
	dx >>= 1;
	readInterp(null, dx, 0, 2 * dx, 2 * dx, edge);
	readInterp(null, 0, dx, dx, 2 * dx, edge);
    }
}



/**
 *  Reads a regular subgrid using differences from interpolated values.
 *  (If dx < dy, interpolation is vertical, else horizontal.)
 *  If tin is null, pure interpolation is performed and nothing is read.
 */
private void readInterp(TypedInputStream tin,
int x0, int y0, int dx, int dy, int edge) throws IOException {

    int xoff, yoff;
    if (dx < dy) {
	xoff = 0;
	yoff = dy / 2;		// interpolate vertically
    } else {
	xoff = dx / 2;		// interpolate horizontally
	yoff = 0;
    }
    int xlim = x0 + edge - xoff;
    int ylim = y0 + edge - yoff;
    int delta = 0;

    for (int y = y0; y <= ylim; y += dy) {
	for (int x = x0; x <= xlim; x += dx) {

	    int e1 = elev[y-yoff][x-xoff];
	    int e2 = elev[y+yoff][x+xoff];
	    int e = (e1 + e2) / 2;
	    if (tin != null) {
		e += tin.readInt();
	    } else if (e1 == MISSING_ELEV) {
		e = e2;
	    } else if (e2 == MISSING_ELEV) {
		e = e1;
	    }
	    elev[y][x] = (char) e;
	}
    }
}



/**
 *  Reads one TopoVista record from an input stream, returning a TVRecord.
 *  Null is returned if EOF is reached before the record is complete.
 */
private static TVRecord readRecord(DataInputStream din) throws IOException {

    try {
	byte[] magic = new byte[4];
	din.readFully(magic);
	if (! (new String(magic)).equals(MAGIC)) {
	    throw new Error("found unmagical header");
	}
	int v = din.readInt();
	int flag = v >>> 24;
	int len = v & 0x00FFFFFF;
	byte data[] = new byte[len - 8];
	din.readFully(data);

	InputStream ins = new ByteArrayInputStream(data);
	if ((flag & TVR_GZIP) != 0) {
	    ins = new GZInputStream(ins);
	}
	TVRecord tvr = new TVRecord();
	tvr.flag = flag & ~TVR_GZIP;
	tvr.gzflag = flag & TVR_GZIP;
	tvr.length = len;
	tvr.tin = new TypedInputStream(ins);
	return tvr;

    } catch (EOFException e) {
	return null;
    }
}



/**
 *  Reads a TopoVista grid file and prints it in readable form.
 *
 *  <P> Usage:  java TileGrid [-v | -s] file.tvg...
 *
 *  <P>  If -v is given, prints full dump of all values;
 *  <BR> If -s is given, just prints a single summary line.
 *  <BR> If -c is given, prints a one-line concise list of record sizes.
 *  <BR> Otherwise, prints one line per record, plus summary.
 */
public static void main(String[] args) throws IOException {

    boolean concise = args.length > 0 && args[0].equals("-c");
    boolean verbose = args.length > 0 && args[0].equals("-v");
    boolean summary = args.length > 0 && args[0].equals("-s");
    int fni = (concise || verbose || summary) ? 1 : 0;

    while (fni < args.length) {

	String fname = args[fni++];
	DataInputStream din = new DataInputStream(new FileInputStream(fname));
	int n1 = 0, n2 = 0, n3 = 0, b1 = 0, b2 = 0, b3 = 0, t = 0;

	if (concise) {
	    System.out.print(fname + ":");
	} else if (!summary) {
	    System.out.println();
	}

	TVRecord tvr;
	while ((tvr = readRecord(din)) != null) {

	    TypedInputStream tin = tvr.tin;
	    Vector v = new Vector(tvr.length);
	    if (!concise && !summary) {
		for (;;) try {
		    v.addElement(tin.readValue());
		} catch (EOFException e) {
		    break;
		}
	    }
	    int n = v.size();

	    switch ((tvr.flag >> 4) & 6) {
		case 2:
		    n2++;
		    t += tvr.length;
		    b2 = Math.max(b2, tvr.length);
		    break;
		case 4:
		case 6:
		    n3++;
		    t += tvr.length;
		    b3 = Math.max(b3, tvr.length);
		    break;
		default:
		    n1++;
		    t += tvr.length;
		    b1 = Math.max(b1, tvr.length);
		    break;
	    }


	    if (concise) {
		System.out.print(" " + tvr.length);
	    } else if (!summary) {
		System.out.print((verbose ? "\n" : "") + fname + ":" +
		    " flag=0x" + Integer.toString(tvr.flag, 16) +
		    " count=" + n +
		    " length=" + tvr.length +
		    ((tvr.gzflag!=0) ? " (compressed)" : " (not compressed)"));
	    }

	    if (verbose) {
		for (int i = 0; i < n; i++) {
		    Object o = v.elementAt(i);
		    System.out.print(((i % 16) == 0) ? "\n" : " ");
		    System.out.print((o instanceof String) ? ("\""+o+"\"") : o);
		}
	    }
	    if (!concise && !summary) {
		System.out.println();
	    }
	}

	if (concise) {
	    System.out.println();
	} else {
	    System.out.println((verbose ? "\n" : "") + fname + ": " +
		n1 + " + " + n2 + " + " + n3 + " records, max " +
		b1 + " / " + b2 + " / " + b3 + " bytes, total " + t);
	    }
    }
}



/**
 *  Writes files of TopoVista grids named "xxxUyyyc.tvg".
 *  Grids that contain no data in the interior (excluding edges)
 *  are suppressed.
 *
 *  @param g       elevation grid
 *  @param k       grid order
 *  @param nprune  number of error levels to prune (normally 3)
 */
static void write(Grid g, int k, int nprune, boolean unzip) throws IOException {

    if (g.refsys != UTM_SYSTEM) {
	throw new Error("not a UTM grid");
    }

    int edge = 1 << k;			// number of cells along an edge
    int hgt = 2 * k + 2;		// height of full tree
    if (nprune > hgt) {
	nprune = hgt;			// limit pruning by max height
    }

    int res = (int) g.xRes;		// grid cell size
    int gridw = edge * res;		// width of the grid, in meters

    int xrem = X_ORIGIN % gridw;
    int xllo = (int)g.xLL - xrem;
    int x0 = xrem + ((xllo + gridw - 1) / gridw) * gridw;
    int x1 = xrem + ((xllo + (g.cols - 1) * res) / gridw) * gridw;
    int nx = (x1 - x0) / gridw;

    int yllo = (int) g.yLL;
    int y0 = ((yllo + gridw - 1) / gridw) * gridw;
    int y1 = ((yllo + (g.rows - 1) * res) / gridw) * gridw;
    int ny = (y1 - y0) / gridw;

    if (nx <= 0 || ny <= 0) {
	System.out.println("   No whole grids are contained in this data.");
	return;
    }

    System.err.println("   " + nx + " x " + ny + " = " + (nx * ny) +
	((nx * ny == 1) ? " tile, " : " tiles, ") +
	(edge + 1) + " x " + (edge + 1) + " (k=" + k + "), " +
	(hgt - nprune) + "/" + hgt + " error levels");
    System.err.print("   loading  ");

    for (int y = y1 - gridw; y >= y0; y -= gridw) {
	int yy = (int) ((y - g.yLL) / g.yRes);
	for (int i = edge; i >= 0; i--) {
	    g.loadRow(yy + i);
	}
	System.err.print("\n   ");
	for (int x = x0; x < x1; x += gridw) {
	    writeGrid(g, k, hgt, nprune, x, y, unzip);
	}
	System.err.print("  ");
	for (int i = edge; i >= 1; i--) {
	    g.discardRow(yy + i);
	}
    }
}



/**
 *  Writes one grid file.
 */
private static void writeGrid(Grid g, int k, int hgt, int nprune,
int xutm, int yutm, boolean unzip) throws IOException {

    int errlev = hgt - nprune;

    int edge = 1 << k;			// number of cells along an edge

    int res = (int) g.xRes;		// grid cell size
    int gridw = edge * res;		// width of the grid, in meters

    int yyy = yutm / gridw;
    int xxx = (xutm - X_ORIGIN) / gridw + X_OFFSET;

    int x0 = (int) ((xutm - g.xLL) / g.xRes);
    int y0 = (int) ((yutm - g.yLL) / g.yRes);
    int x1 = x0 + edge;
    int y1 = y0 + edge;

    int e;
    if (g.zUnit == FEET) {
	e = (int) (304.8 * g.zRes + 0.5);
    } else {
	e = (int) (1000 * g.zRes + 0.5);
    }

    // calculate minimum and maximum elevation, checking for valid data

    boolean empty = true;
    int l = Integer.MAX_VALUE;
    int h = Integer.MIN_VALUE;
    for (int y = y0; y <= y1; y++) {
	for (int x = x0; x <= x1; x++) {
	    int v = g.elev[y][x];
	    if (v != MISSING_ELEV) {
		if (l > v) {
		    l = v;
		}
		if (h < v) {
		    h = v;
		}
		if (y > y0 && y < y1 && x > x0 && x < x1) {
		    empty = false;
		}
	    }
	}
    }

    if (empty) {
	System.err.print(". ");
	System.err.flush();
	return;				// suppress tile with no data
    }

    // open output file
    String filename = makeName(g.zone, xxx, yyy, k) + ".tvg";
    OutputStream out = new FileOutputStream(filename);

    // write the header
    ByteArrayOutputStream byo = new ByteArrayOutputStream(100);
    TypedOutputStream tyo = new TypedOutputStream(byo);
    tyo.write(filename);		// filename as title
    tyo.write("");			// empty version
    tyo.write("");			// empty comment
    tyo.write(res);
    tyo.write(g.zone);
    tyo.write(g.datum);
    tyo.write(k);
    tyo.write(errlev);
    tyo.write(xxx);
    tyo.write(yyy);
    tyo.write(e);
    tyo.write(l - ELEVATION_BIAS);
    tyo.write(h - ELEVATION_BIAS);
    writeRecord(out, tyo, byo, TVR_HEADER, unzip);

    // write elevation records 0 through k

    for (int i = 0; i <= k; i++) {
	int side = (1 << i) + 1;
	byo = new ByteArrayOutputStream(side * side);	// generous estimate
	tyo = new TypedOutputStream(byo);
	if (i == 0) {
	    tyo.write(g.elev[y0][x0] - ELEVATION_BIAS);
	    tyo.write(g.elev[y0][x0+edge] - ELEVATION_BIAS);
	    tyo.write(g.elev[y0+edge][x0] - ELEVATION_BIAS);
	    tyo.write(g.elev[y0+edge][x0+edge] - ELEVATION_BIAS);
	} else {
	    int dx = 1 << (k - i);
	    writeInterp(tyo, g, x0 + dx, y0, 2 * dx, 2 * dx, edge);
	    writeInterp(tyo, g, x0, y0 + dx, dx, 2 * dx, edge);
	}
	writeRecord(out, tyo, byo, TVR_ELEVATION + i, unzip);
    }

    // build array (flattened binary tree) of error values
    // erray[0] is not used
    // erray[1] is root node (arbitrarily assigned error=0)

    int asize = 1 << errlev;
    short[] earray = new short[asize];

    if (asize > 2)  switch (2 * (xxx & 1) + (yyy & 1)) {

	case 0:				/* xxx even, yyy even */
	    tri(g, earray, hgt - 1, 2, 6, x0, y0, x1, y1, x0, y1);
	    tri(g, earray, hgt - 1, 3, 2, x1, y1, x0, y0, x1, y0);
	    break;
	case 1:				/* xxx even, yyy odd */
	    tri(g, earray, hgt - 1, 2, 0, x1, y0, x0, y1, x0, y0);
	    tri(g, earray, hgt - 1, 3, 4, x0, y1, x1, y0, x1, y1);
	    break;
	case 2:				/* xxx odd, yyy even */
	    tri(g, earray, hgt - 1, 2, 4, x0, y1, x1, y0, x1, y1);
	    tri(g, earray, hgt - 1, 3, 0, x1, y0, x0, y1, x0, y0);
	    break;
	case 3:				/* xxx odd, yyy odd */
	    tri(g, earray, hgt - 1, 2, 2, x1, y1, x0, y0, x1, y0);
	    tri(g, earray, hgt - 1, 3, 6, x0, y0, x1, y1, x0, y1);
	    break;
    }

    // write error records 0 through n
    for (int i = 0; i < errlev; i++) {

	int b = 1 << i;			// base index, equals number of values
	byo = new ByteArrayOutputStream(5 * b / 4);	// generous estimate
	tyo = new TypedOutputStream(byo);
	for (int j = 0; j < b; j++) {
	    int p = (b + j) / 2;		// parent error index
	    tyo.write(earray[b+j] - earray[p]);	// write difference
	}
	writeRecord(out, tyo, byo, TVR_ERROR + i, unzip);
    }

    out.close();
    System.err.print("o ");
    System.err.flush();
}



/**
 *  Writes a regular subgrid using differences from interpolated values.
 *  (If dx < dy, interpolation is vertical, else horizontal.)
 */
private static void writeInterp(TypedOutputStream out, Grid g,
int x0, int y0, int dx, int dy, int edge) throws IOException {

    int xoff, yoff;
    if (dx < dy) {
	xoff = 0;
	yoff = dy / 2;		// interpolate vertically
    } else {
	xoff = dx / 2;		// interpolate horizontally
	yoff = 0;
    }
    int xlim = x0 + edge - xoff;
    int ylim = y0 + edge - yoff;

    for (int y = y0; y <= ylim; y += dy) {
	for (int x = x0; x <= xlim; x += dx) {
	    int e1 = g.elev[y-yoff][x-xoff];
	    int e2 = g.elev[y+yoff][x+xoff];
	    int e = (e1 + e2) / 2;
	    out.write(g.elev[y][x] - e);
	}
    }
}



/**
 *  Calculates, stores, and returns the error for a triangle.
 *
 *  @param g		Grid object
 *  @param earray	result array for as many levels as desired
 *  @param hgt		height of this subtree (including levels to be pruned)
 *  @param n		array index for this node
 *  @param o		triangle orientation (units of 45deg from 1st quadrant)
 *  @param x1,y1	first counterclockwise vertex
 *  @param x2,y2	second counterclockwise vertex
 *  @param x3,y3	right-angled vertex; origin for orientation measurement
 */
private static int tri(Grid g, short[] earray, int hgt, int n, int o,
int x1, int y1, int x2, int y2, int x3, int y3) {

    int nn = 2 * n;			// index of left child
    int xm = (x1 + x2) / 2;		// midpoint of hypotenuse
    int ym = (y1 + y2) / 2;

    if (nn < earray.length) {		// if not at deepest level, recurse
	tri(g, earray, hgt - 1, nn,     o + 5, x3, y3, x1, y1, xm, ym);
	tri(g, earray, hgt - 1, nn + 1, o + 3, x2, y2, x3, y3, xm, ym);
    }

    if (hgt <= 3) {			// if easy special case near bottom
	if (hgt <= 1) {			// if leaf node
	    return earray[n] = 0;	// error is zero
	}

	float err =			// error at midpoint of hypotenuse
	    Math.abs(g.elev[ym][xm] - (g.elev[y1][x1] + g.elev[y2][x2]) / 2f);

	if (hgt == 3) {			// if two more points to consider
	    int x4 = (x1 + x3) / 2;
	    int y4 = (y1 + y3) / 2;
	    float e4 = g.elev[y4][x4] - (g.elev[y1][x1] + g.elev[y3][x3]) / 2f;
	    err = Math.max(err, Math.abs(e4));
	    int x5 = (x2 + x3) / 2;
	    int y5 = (y2 + y3) / 2;
	    float e5 = g.elev[y5][x5] - (g.elev[y2][x2] + g.elev[y3][x3]) / 2f;
	    err = Math.max(err, Math.abs(e5));
	}

	return earray[n] = (short) (err + 0.5);
    }

    // general case: find the plane containing the triangle,
    // and check its fit at every point within.

    Plane p = new Plane(g.elev, x1, y1, x2, y2, x3, y3);
    int w = Math.max(Math.abs(x3 - x1), Math.abs(x3 - x2));
    switch (o & 7) {		// scanning depends on triangle orientation
	case 0:
	    for (int i = w; i >= 0; i--)
		for (int j = w - i; j >= 0; j--)
		    p.eval(x3 + j, y3 + i);
	    break;
	case 1:
	    for (int i = 0; i <= w; i++)
		for (int j = -i; j <= i; j++)
		    p.eval(x3 + j, y3 + i);
	    break;
	case 2:
	    for (int i = w; i >= 0; i--)
		for (int j = w - i; j >= 0; j--)
		    p.eval(x3 - j, y3 + i);
	    break;
	case 3:
	    for (int j = 0; j <= w; j++)
		for (int i = -j; i <= j; i++)
		    p.eval(x3 - j, y3 + i);
	    break;
	case 4:
	    for (int i = w; i >= 0; i--)
		for (int j = w - i; j >= 0; j--)
		    p.eval(x3 - j, y3 - i);
	    break;
	case 5:
	    for (int i = 0; i <= w; i++)
		for (int j = -i; j <= i; j++)
		    p.eval(x3 + j, y3 - i);
	    break;
	case 6:
	    for (int i = w; i >= 0; i--)
		for (int j = w - i; j >= 0; j--)
		    p.eval(x3 + j, y3 - i);
	    break;
	case 7:
	    for (int j = 0; j <= w; j++)
		for (int i = -j; i <= j; i++)
		    p.eval(x3 + j, y3 + i);
	    break;
    }
    return earray[n] = (short) (p.error + 0.5);
}



/**
 *  Writes a TopoVista record.
 *  The data field is gzipped if that saves space.
 */
private static void writeRecord(OutputStream out, TypedOutputStream tyo,
ByteArrayOutputStream byo, int flag, boolean unzip) throws IOException {

    tyo.flush();			// flush stream
    byte[] b1 = byo.toByteArray();	// get uncompressed array
    byte[] b2 = b1;

    if (!unzip) {			// if zipped version allowed
	ByteArrayOutputStream bo = new ByteArrayOutputStream(b1.length);
	GZIPOutputStream zo = new GZIPOutputStream(bo);
	DataOutputStream o = new DataOutputStream(zo);
	o.write(b1, 0, b1.length);
	o.flush();
	zo.finish();
	zo.close();			// to avoid memleak (JDK 1.4.1_02)
	b2 = bo.toByteArray();		// get uncompressed array
    }

    DataOutputStream o = new DataOutputStream(out);
    o.writeBytes(MAGIC);
    if (b2.length < b1.length) {	// write compressed only if shorter
	o.writeInt(((flag | TVR_GZIP) << 24) | (8 + b2.length));
	o.write(b2, 0, b2.length);
    } else {
	o.writeInt((flag << 24) | (8 + b1.length));
	o.write(b1, 0, b1.length);
    }
    o.flush();
}



} // class TileGrid
