//  Certifier.java -- keeps track of token authority


package topovista.toksvc;

import java.util.*;



/**
 *  A Certifier maintains information about the
 *  authority of a Server to serve particular tokens.
 */
class Certifier {

    private int nservers;		// number of servers
    private int myid;			// index of this server

    private boolean transferred[];	// list of servers we've inited from
    private byte[] sstates;		// current roster of server states

    private Hashtable waitlist;		// clients from which awaiting reports
	// initially, these are the clients needed to declare this server ready
	// later, they're the clients needed to serve visiting tokens safely



private static SessionList emptySessionList = new SessionList();



/**
 *  Designates the state of a token for which the local server has
 *  responsibility and is prepared to serve requests.
 */
final static int READY = 1;

/**
 *  Designates the state of a token for which the local server has
 *  responsibility but cannot serve pending receipt of information.
 */
final static int WAITING = 2;

/**
 *  Designates the state of a token for which the local server has
 *  lost responsibility but is retaining for transfer elsewhere.
 */
final static int HOLDING = 3;

/**
 *  Designates the state of a token for which the local server has
 *  no responsibility and retains no information.
 */
final static int FOREIGN = 4;



/**
 *  Constructs a Certifier for use by Server i.
 */
Certifier(Council c, int i) {
    nservers = c.getMembers().length;		// save server count
    myid = i;					// save local id
    transferred = new boolean[nservers];	// init transfer checklist
    transferred[i] = true;			// don't need to hear from self
    waitlist = new Hashtable();
}



/**
 *  Updates the Certifier's knowledge of server and client states.
 */
void newRoster(RosterMessage m) {
    byte[] prevstates = sstates;
    sstates = m.sstates;		// save server states for reference
    SessionList slist = m.slist;
    boolean addc = false;

    // check the clients we're awaiting, and remove any that have died
    if (waitlist.size() > 0) {
	Vector v = new Vector();
	for (Enumeration e = waitlist.elements(); e.hasMoreElements(); ) {
	    Session s = (Session) e.nextElement();
	    if (slist.get(s) == null) {		// if no longer current
		v.addElement(s);		// add to hitlist
	    }
	}
	for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
	    waitlist.remove((Session) e.nextElement());
	}
    }

    // check to see if we need a new set of client updates
    if (prevstates == null) {		// if first roster
	addc = true;			// need to hear from all clients
    } else {
	for (int i = 0; i < nservers; i++) {
	    if (prevstates[i] != Server.DOWN && sstates[i] == Server.DOWN) {
		addc = true;		// a server just crashed; need to query
		break;
	    }
	}
    }

    // if we need to hear from these clients, put them on the waitlist
    if (addc) {
	for (Enumeration e = slist.elements(); e.hasMoreElements(); ) {
	    Session s = (Session) e.nextElement();
	    waitlist.put(s, s);
	}
    }
}



/**
 *  Tells the Certifier that the local server has received a transfer
 *  from the specified other server.
 */
void heardFrom(int id) {
    transferred[id] = true;
}



/**
 *  Tells the Certifier that the server has received a catalog
 *  from the specified client session.
 */
void heardFrom(Session s) {
    waitlist.remove(s);
}



/**
 *  Returns true if the server can declare itself READY, or
 *  false if still awaiting transfers from some other servers
 *  or clients.
 */
boolean ready() {

    if (sstates == null) {		// if don't have a roster yet
	return false;
    }
    if (waitlist.size() != 0) {		// if still waiting for some clients
	return false;
    }
    for (int i = 0; i < nservers; i++) {
	if (sstates[i] != Server.DOWN && !transferred[i]) {
	    return false;		// if still waiting for server transfer
	}
    }
    return true;
}



/**
 *  Returns a list of client sessions from which we await catalogs.
 */
SessionList awaiting() {
    if (waitlist.size() == 0) {
	return emptySessionList;
    }
    SessionList sl = new SessionList();
    for (Enumeration e = waitlist.elements(); e.hasMoreElements(); ) {
	sl.add((Session) e.nextElement());
    }
    return sl;
}



/**
 *  Returns the state of the token server's authority
 *  with regard to the token specified:
 *  <UL>
 *	<LI> READY: local responsibility, ready to serve
 *	<LI> WAITING: local responsibility, but awaiting information first
 *	<LI> HOLDING: no responsibility, but hold data for transfer elsewhere
 *	<LI> FOREIGN: no responsibility and no need to remember
 *  </UL>
 */
int status(ServerToken t) {
    if (sstates == null) {			// if not we're not ready yet
	return WAITING;
    }
    int n = t.server(sstates);
    if (n != myid) {				// if not our token
	if (sstates[n] == Server.READY) {
	    return FOREIGN;
	} else {
	    return HOLDING;
	}
    }
    if (sstates[myid] != Server.READY) {
	return WAITING;		// if our token, but not ready yet
    }
    if (t.primary(nservers) != myid && waitlist.size() != 0) {
	return WAITING;		// if visiting token, may need data first
    }
    return READY;		// it's our token and we're ready to go
}



}  // class Certifier
