//  Council.java -- implements a group of cooperating network processes


package topovista.toksvc;

import java.io.*;
import java.net.*;


/**
 *  Council implements membership in a group of
 *  cooperating network processes.  Facilities for sending,
 *  broadcasting, and receiving messages are provided.
 *  All members must supply exactly the same membership list.
 */

class Council extends Thread {

    private String[] members;		// host:port strings of paxos group
    private InetAddress[] addrs;	// corresponding internet addresses
    private int ports[];		// corresponding port numbers

    private int lsig;			// signature of member list
    private int self;			// index of this member's entry

    private DatagramSocket rcvport;	// UDP socket for receiving messages
    private DatagramSocket[] sendport;	// UDP sockets for sending messages

    private Receiver[] rcvrs;		// receiver callbacks
    private boolean closing;		// set when shutting down

    private String strrep;		// toString() value
    private String dmpname;		// dump file name if aborting
    private Logger log = new Logger();	// debugging log

    private double reliability;		// probability of processing message



/**
 *  Joins a council.  Exactly one member index must match the current host.
 *
 *  @param members	group membership, as strings in "host:portnum" form.
 */

Council(String[] members)
throws SocketException, UnknownHostException {
    this(members, -1);
}



/**
 *  Constructor.
 *
 *  @param members	group membership, as strings in "host:portnum" form
 *  @param index	index of this member, or -1 for automatic detection
 */

Council(String[] members, int index)
throws SocketException, UnknownHostException {

    // save private copy of member list
    this.members = new String[members.length];
    System.arraycopy(members, 0, this.members, 0, members.length);

    // translate host:port strings
    ports = new int[members.length];
    addrs = new InetAddress[members.length];
    for (int i = 0; i < members.length; i++) {
	ports[i] = Util.portNumber(members[i]);
	addrs[i] = InetAddress.getByName(Util.hostPart(members[i]));
    }

    // identify and/or check the index of the present host
    if (index < 0) {
	self = findLocal(addrs);
    } else if (isLocal(addrs[index])) {
	self = index;
    } else {
	throw new SocketException(members[index] + " is not local");
    }

    // calculate member list signature
    lsig = Util.hash(members) & 0x1FFF;

    // create representation for toString()
    // ==> when modifying format, update toString() documentation below
    strrep = Util.classname(this) + "(" + self + "/" + members.length + "," +
	lsig + "," + members[self] + ")";

    // define dump file name and remove old copy, if any
    dmpname = "mbr" + self + ".dmp";
    try {
	new File(dmpname).delete();
    } catch (Exception e) {
    }

    // start off the debugging log
    log.write('i', "new " + this);

    // open UDP ports for sending and receiving packets
    rcvport = new DatagramSocket(Util.portNumber(members[self]));
    sendport = new DatagramSocket[members.length];
    for (int i = 0; i < members.length; i++) {
	sendport[i] = new DatagramSocket();
    }

    // initialize empty receiver list
    rcvrs = new Receiver[0];

    // start listener thread
    reliability = 1.0;
    this.setName("listener");
    this.setDaemon(true);
    this.start();
}



/**
 *  Returns a string representation of this Council member.
 *  The string includes the member index, member count, member
 *  list signature, and host:port string.
 */
public String toString() {
    return strrep;
}



/**
 *  Returns the membership list of this Council group.
 */
String[] getMembers() {
    return members;
}



/**
 *  Returns the index of this participant in the member list.
 */
int getIndex() {
    return self;
}



/**
 *  Finds the index of the unique local address.
 */
private static int findLocal(InetAddress[] a)
throws SocketException {

    int h = 0;
    int n = 0;
    for (int i = 0; i < a.length; i++) {
	if (isLocal(a[i])) {
	    h = i;
	    n++;
	}
    }
    switch (n) {
	case 0:  throw new SocketException("no local socket in list");
	case 1:  return h;
	default: throw new SocketException(n + " local sockets in list");
    }
}



/**
 *  Checks whether a given InetAddress represents the local host.
 */
private static boolean isLocal(InetAddress a) {
    if (localhost == null || localaddrs == null) {
	try {
	    localhost = InetAddress.getByName("localhost");
	    localaddrs = InetAddress.getAllByName(
		InetAddress.getLocalHost().getHostName());
	} catch (UnknownHostException e) {
	    return false;
	}
    }
    return localhost.equals(a) || Util.matchAny(localaddrs, a);
}

private static InetAddress localhost;		// "localhost" addr (127.0.0.1)
private static InetAddress[] localaddrs;	// net address(es) of local host



/**
 *  Sets the probability that an incoming message will be processed.
 *  A probability of less than 1.0 allow simulation of an unreliable
 *  network.
 */
public void setReliability(double p) {
    reliability = p;
}



/**
 *  Adds a receiver callback to be invoked for every
 *  incoming message.
 */
synchronized void listen(Receiver r) {
    Receiver[] newlist = new Receiver[rcvrs.length + 1];
    System.arraycopy(rcvrs, 0, newlist, 0, rcvrs.length);
    rcvrs = newlist;
    rcvrs[rcvrs.length - 1] = r;
}



/**
 *  Listens for incoming datagrams, and responds.
 *  Started automatically by the constructor; do not start explicitly.
 */
public void run() {
    log.write('i', "starting listener");
    int n = Message.MAX_LENGTH;
    for (;;) {
	try {
	    DatagramPacket d = new DatagramPacket(new byte[n], n);
	    rcvport.receive(d);
	    if (reliability < Math.random()) {
		continue;			// simulate network loss
	    }
	    try {
		Message m = Message.unpickle(d);
		log.write('r', m.toString());
		if (m.msig == lsig) {		// if valid for this council
		    rcv(m);
		}
	    } catch (java.util.NoSuchElementException e) {
		// ignore malformed (probably truncated) message
		log.write('r', e.getMessage());
	    }
	} catch (IOException e) {
	    if (closing) {
		rcv(null);			// signal closing to receivers
		for (int i = 0; i < members.length; i++) {
		    sendport[i].close();
		}
		return;				// normal exit
	    } else {
		abort("receiver", e);
	    }
	} catch (Throwable t) {
	    abort("receiver", t);
	}
    }
}



/**
 *  Distributes an incoming message to all receivers.
 */
private synchronized void rcv(Message m) {
    for (int i = 0; i < rcvrs.length; i++) {
	rcvrs[i].incoming(m);
    }
}



/**
 *  Sends a message to all council members.
 *  I/O Exceptions are ignored.
 */
void broadcast(Message m) {
    send(m, -1);
}



/**
 *  Sends a message, setting the from, to, and msig fields of the packet.
 *  A to value of -1 indicates a broadcast to all members, including self.
 *  I/O Exceptions are ignored.
 */
void send(Message m, int to) {
    m.msig = lsig;
    m.from = self;
    m.to   = to;
    log.write('s', m.toString());
    byte[] buf = m.pickle(new Pickle()).toByteArray();

    if (to == self) {
	// copy data, don't share it!
	rcv(Message.unpickle(new Pickle(buf)));
    } else if (to >= 0) {
	transmit(buf, to);
    } else {
	// send a copy to every member (could alternatively use IP multicast)
	for (int i = 0; i < addrs.length; i++) {
	    if (i == self) {
		rcv(Message.unpickle(new Pickle(buf)));
	    } else {
		transmit(buf, i);
	    }
	}
    }
}



/**
 *  Sends a message to a client session.
 */
void send(Message m, Session s) {
    m.msig = lsig;
    m.from = self;
    m.to = s.getID();
    log.write('s', m.toString());
    s.send(m);
}



/**
 *  Actually sends a message, ignoring I/O exceptions.
 */
private void transmit(byte[] buf, int to) {
    InetAddress dest = addrs[to];
    int port = ports[to];
    DatagramPacket p = new DatagramPacket(buf, buf.length, dest, port);
    try {
	sendport[to].send(p);
    } catch (IOException e) {
    }
}



/**
 *  Terminates the protocol, sending a null message to each receiver.
 */
void close() {
    log.write('i', "Council.close()");
    closing = true;			// signal shutdown
    rcvport.close();			// interrupt the listener thread
}



/**
 *  Terminates the protocol abnormally, tracing the given exception
 *  and writing the last log messages to a file.
 */
void abort(String msg, Throwable t) {
    try {
	log.write('i', msg);
	log.setFile(dmpname);
	log.close();
    } catch (Exception e) {
    }
    if (msg == null && t != null) {
	msg = t.toString();
    }
    System.err.println("Council abort: " + msg);
    if (t != null) {
	t.printStackTrace();
    }
    System.exit(1);
}



}  // class Council
