//  Director.java -- director of token service operations


package topovista.toksvc;

import java.io.*;
import java.util.*;



/**
 *  A Director, when chosen to lead,
 *  maintains and distributes client and server membership lists.
 */

class Director extends Leader {

    private Council council;		// the council we belong to
    private int myid;			// our member index
    private int nservers;		// number of members

    private boolean incharge;		// set true when we're the leader
    private int hbcount;		// heartbeat count

    private RosterMessage lastroster;	// last RosterMessage received
    private int[] lastmsg;		// last message time (heartbeat count)

    private int nextsid;		// next session ID to be assigned
    private SessionList slist;		// list of client sessions
    private Hashtable lastalive;	// last message time (heartbeat count)

    private byte[] sstates;		// server state array



private static final int SERVER_MISSES = 5;	// missed heartbts to call DOWN
private static final int SERVER_STARTUP = 5;	// additional margin over switch

private static final int CLIENT_MISSES = 10;	// missed heartbts to call DOWN
private static final int CLIENT_STARTUP = 10;	// additional margin over switch



/**
 *  Constructor.
 */
Director(Council c) {

    super(c);
    council = c;			// save parent council
    myid = c.getIndex();		// save our index
    nservers = c.getMembers().length;	// save count of members

    slist = new SessionList();		// init empty session list
    lastalive = new Hashtable();	// init empty client-alive table

    sstates = new byte[nservers];	// init server states (all DOWN)
    for (int i = 0; i < nservers; i++) {
	sstates[i] = Server.DOWN;
    }
    lastmsg = new int[nservers];	// last-time-heard-from array
    lastroster = new RosterMessage(sstates, slist);  // init w/ empty msg

    c.listen(new Receiver() {
	public void incoming(Message m) {
	    receive(m);
	}
    });
}



/**
 *  Called when this instance becomes the token leader.
 */
void accede() {

    System.arraycopy(lastroster.sstates, 0, sstates, 0, nservers);
    slist = lastroster.slist;

    // initialize "lastmsg" array to give servers a chance to switch leaders
    for (int i = 0; i < nservers; i++) {
	lastmsg[i] = hbcount + SERVER_STARTUP;
    }

    // initialize similar table for clients
    lastalive = new Hashtable();
    for (Enumeration e = slist.elements(); e.hasMoreElements(); ) {
	lastalive.put(e.nextElement(), new Integer(hbcount + CLIENT_STARTUP));
    }

    incharge = true;

    // initialize session ID assignments, leaving a deliberate gap
    nextsid = Session.getMaxID();	// get maximum ID used so far
    if (nextsid < nservers) {		// ensure distinct session & server IDs
	nextsid = nservers;
    }
    nextsid = nextsid + 100;		// allocate, with arbitrary gap
}



/**
 *  Called when this instance loses leadership from failure to communicate.
 */
void abdicate() {
    incharge = false;
}



/**
 *  Send a heartbeat RosterMessage when asked by the superclass.
 */
void heartbeat() {

    // mark as DOWN any server that hasn't been heard from recently
    for (int i = 0; i < nservers; i++) {
	if (lastmsg[i] + SERVER_MISSES < hbcount) {
	    sstates[i] = Server.DOWN;
	}
    }

    // remove any clients that haven't been heard from lately
    for (Enumeration e = slist.elements(); e.hasMoreElements(); ) {
	Session s = (Session) e.nextElement();
	Integer i = (Integer) lastalive.get(s);
	if (i != null && i.intValue() + CLIENT_MISSES < hbcount) {
	    slist.remove(s);
	    lastalive.remove(s);
	}
    }

    // send latest state data
    broadcast(new RosterMessage(sstates, slist));

    hbcount++;		// update heartbeat count
}



/**
 *  Called for every incoming message.
 */
private void receive(Message m) {

    switch (m.type) {

	case AliveMessage.MTYPE: {
	    // This is a client confirming that it's still functioning
	    Session s = slist.get(m.from);
	    if (s != null) {
		lastalive.put(s, new Integer(hbcount));
	    }
	    break;
	}

	case ServerMessage.MTYPE: {

	    // This is a server reporting its state.
	    // Believe it, except that a DOWN host cannot report READY
	    // (it must go through a BOOTING stage first).

	    int i = m.from;
	    byte b = (byte) ((ServerMessage)m).state;
	    if (b != Server.READY || sstates[i] != Server.DOWN) {
		sstates[i] = b;
		lastmsg[i] = hbcount;
	    }
	    break;
	}

	case RosterMessage.MTYPE: {

	    // This message reports the roster of server states.
	    // If it's the newest roster we've seen, save it for future ref

	    RosterMessage r = (RosterMessage) m;
	    if (r.msgnum > lastroster.msgnum) {
		lastroster = r;
	    }
	    break;
	}

	case LoginMessage.MTYPE: {

	    if (lastroster.sstates[myid] != Server.READY) {
		break;			// if we're still initializing
	    }

	    // This is a new client trying to get an ID assigned.
	    // If we're leading, give it one.  If not, don't.
	    // Either way send it a ConfigMessage telling it who's in charge.

	    LoginMessage lm = (LoginMessage) m;
	    String host = lm.pkt.getAddress().getHostAddress();
	    int port = Util.portNumber(lm.address);
	    int id = incharge ? nextsid++ : 0;
	    Session s = new Session(id, host + ":" + port);
	    if (incharge) {
		slist.add(s);
		lastalive.put(s, new Integer(hbcount));
		// send extra roster to notify other servers of new client
		broadcast(new RosterMessage(sstates, slist));
	    }
	    ConfigMessage cm =
		new ConfigMessage(lastroster.from, lastroster.sstates);
	    council.send(cm, s);
	    break;
	}

	case LogoutMessage.MTYPE: {

	    // This client is leaving.  Remove it from the list.
	    Session s = slist.get(m.from);
	    if (s != null) {
		slist.remove(s);
		lastalive.remove(s);
	    }
	    break;
	}
    }
}


}  // class Director
