//  Elector.java -- cooperates to vote for a council leader


package topovista.toksvc;

import java.io.*;



/**
 *  A Elector cooperates with other electors of a council
 *  to choose a leader.
 */

class Elector {

    private Council council;		// the council we belong to
    private int myid;			// our member index

    private Queue evq;			// event queue
    private Timer tm;			// timeout alarm

    private ElectMessage[] emsgs;	// most recent election messages seen
    private LeadMessage smsg;		// most recent state message seen



/**
 *  Constructor.
 */
Elector(Council c) {

    council = c;			// save parent council
    myid = c.getIndex();		// save our index

    evq = new Queue();			// establish event queue

    smsg = new LeadMessage();		// initialize "last state message seen"
    smsg.from = -1;			// with nonexistent sender

    tm = new Timer() {			// create timer
	public void alarm(Timer t, Object val) {
	    evq.put(val);		// on timeout, send event to queue
	}
    };

    //  start state machine to handle events
    Thread t = new Thread() {
	public void run() {
	    elector();			// run elector in its own thread
	}
    };
    t.setDaemon(true);
    t.start();

    //  send incoming messages to the state machine via its event queue
    c.listen(new Receiver() {
	public void incoming(Message m) {
	    if (m instanceof CouncilMessage) {
		evq.put(m);
	    }
	}
    });
}



//  The elector just oscillates between two states.

private void elector() {
    try {
	for (;;) {
	    watch();
	    vote();
	}
    } catch (Throwable t) {
	council.abort("elector", t);
    }
}


//  The watching state persists until it sees that the leader is dead
//  or that another watcher has called an election.

private void watch() throws IOException {

    emsgs = new ElectMessage[council.getMembers().length];
    tm.set(5 * Leader.HBT, "noLeader");

    for (;;) {
	Object o = evq.get();

	if (o instanceof LeadMessage) {
	    // saw a Lead message from leader; reset timeout
	    LeadMessage m = (LeadMessage) o;
	    if (m.term > smsg.term ||
		    (m.term == smsg.term && m.msgnum > smsg.msgnum)) {
		smsg = m;
		tm.set(3 * Leader.HBT, "noLeader");
	    }

	} else if (o instanceof ElectMessage) {
	    // saw an election message; go vote
	    ElectMessage m = (ElectMessage) o;
	    emsgs[m.from] = m;
	    return;		// exit WATCHING state

	} else if (o == "noLeader") {
	    // got a timeout: leader is dead
	    return;		// enter VOTING state
	}
    }
}



//  The voting state lasts for one RTT timeout.

private void vote() throws IOException {

    tm.set(Leader.RTT, "voteDone");
    sendElect();

    for (;;) {
	Object o = evq.get();

	if (o instanceof ElectMessage) {
	    ElectMessage m = (ElectMessage) o;
	    emsgs[m.from] = m;

	} else if (o == "voteDone") {
	    sendVote();
	    return;
	}
    }
}



//  broadcast an Elect message telling everyone the last state we received

private void sendElect() throws IOException {
    ElectMessage m = new ElectMessage(smsg.term, smsg.msgnum);
    council.broadcast(m);	// broadcast data about last state msg seen
    emsgs[myid] = m;		// save copy of what we sent
}



//  choose best new (or returning) leader, and send it a vote

private void sendVote() throws IOException {

    int ldr = smsg.from;
    if (ldr >= 0 && emsgs[ldr] != null && emsgs[ldr].msgnum >= smsg.msgnum) {

	// current leader is still alive; choose it
	council.send(new VoteMessage(smsg.term, smsg.msgnum + 1), smsg.from);

    } else {

	// choose new best leader
	int best = myid;
	for (int i = 0; i < emsgs.length; i++) {
	    if (emsgs[i] == null) {
		continue;
	    }
	    if (emsgs[i].term > emsgs[best].term) {
		best = i;
		continue;
	    } else if (emsgs[i].term < emsgs[best].term) {
		continue;
	    }
	    if (emsgs[i].msgnum > emsgs[best].msgnum) {
		best = i;
		continue;
	    } else if (emsgs[i].msgnum < emsgs[best].msgnum) {
		continue;
	    }
	    if (i < best) {
		best = i;
	    }
	}
	int newterm = emsgs[best].term + 1;
	long newmsg = emsgs[best].msgnum + 1;
	council.send(new VoteMessage(newterm, newmsg), best);
    }
}



}  // class Elector
