//  Leader.java -- competes to become the council leader


package topovista.toksvc;

import java.io.*;



/**
 *  A Leader counts votes and assumes leadership
 *  if elected.  It runs the Candidate portion of the election
 *  protocol and also spawns a separate Elector process.
 *  This class is typically extended by a class
 *  that actually *does* something when acting as leader.
 */

class Leader {

    private Council council;		// the council we belong to
    private int myid;			// our member index

    private Queue evq;			// event queue
    private Timer tm;			// timeout alarm

    private VoteMessage[] votes;	// most recent election messages seen

    private int term;			// current administrative term
    private long msgnum;		// next message number to send



/**
 *  Defines a (generous) upper bound for round-trip message latency, in msec.
 */
static final int RTT = 250;

/**
 *  Defines the heartbeat interval, in msec.
 */
static final int HBT = 1000;



/**
 *  Constructor.
 */
Leader(Council c) {

    council = c;			// save parent council
    myid = c.getIndex();		// save our index

    evq = new Queue();			// establish event queue

    tm = new Timer() {			// create timer
	public void alarm(Timer t, Object val) {
	    evq.put(val);		// on timeout, send event to queue
	}
    };

    //  start state machine to handle events
    Thread t = new Thread() {
	public void run() {
	    candidate();		// run candidate in its own thread
	}
    };
    t.setDaemon(true);
    t.start();

    //  send incoming messages to the state machine via its event queue
    c.listen(new Receiver() {
	public void incoming(Message m) {
	    if (m instanceof CouncilMessage) {
		evq.put(m);
	    }
	}
    });

    //  start separate Elector machine
    new Elector(c);
}



/**
 * The candidate just oscillates between two states.
 */
private void candidate() {
    try {
	for (;;) {
	    follow();
	    lead();
	}
    } catch (Throwable t) {
	council.abort("candidate", t);
    }
}


/**
 *  The follower watches for incoming votes, and notes
 *  whether this candidate has been elected.
 */
private void follow() throws IOException {

    votes = new VoteMessage[council.getMembers().length];

    for (;;) {
	Object o = evq.get();

	if (o instanceof LeadMessage) {

	    // saw a Lead message, indicating that leadership is functional;
	    // clear all votes
	    for (int i = 0; i < votes.length; i++) {
		votes[i] = null;
	    }

	} else if (o instanceof VoteMessage) {

	    // received Vote message.  Count votes identical to this one.
	    VoteMessage m = (VoteMessage) o;
	    votes[m.from] = m;
	    int n = 0;
	    for (int i = 0; i < votes.length; i++) {
		VoteMessage v = votes[i];
		if (v != null && v.term == m.term && v.msgnum == m.msgnum) {
		    n++;
		}
	    }
	    if (n > votes.length / 2) {		// if this makes a majority,
		term = m.term;
		msgnum = m.msgnum;
		return;				// we're elected leader
	    }
	}
    }
}


/**
 *  The leader rules forever, sending state messages periodically -- unless
 *  something goes wrong and another candidate claims the leadership.
 */
private void lead() throws IOException {

    accede();				// assume control
    tm.set(0, "heartbeat");		// start with immediate heartbeat

    for (;;) {
	Object o = evq.get();		// get next event from queue

	if (o == "heartbeat") {		// if time to send a heartbeat
	    heartbeat();			// send message
	} else if (o instanceof LeadMessage) {
	    LeadMessage m = (LeadMessage) o;
	    if (m.from != myid && m.term >= term) {  // if saw a new leader
		abdicate();				// step down
		return;					// reenter follower
	    }
	}
    }
}



/**
 *  Assumes leadership after becoming elected.
 *  Does nothing, but can be overridden by subclasses
 *  to perform initialization.
 */
void accede() {
}



/**
 *  Relinquishes leadership, because a new leader has been elected.
 *  This normally indicates a communications breakdown of some sort.
 *  Does nothing, but can be overridden by subclasses if they need
 *  to be informed of this occurrence.
 */
void abdicate() {
}



/**
 *  Send a heartbeat LeadMessage by calling this.broadcast().
 *  Can be overridden by subclasses, but they must be sure to call
 *  broadcast(), else the heartbeat timer will stop.
 */
void heartbeat() {
    broadcast(new LeadMessage());
}



/**
 *  Broadcast a LeadMessage, adding the current term and a new message number,
 *  and resetting the heartbeat timer.
 */
void broadcast(LeadMessage m) {
    m.term = term;
    m.msgnum = msgnum++;
    council.broadcast(m);
    tm.set(Leader.HBT, "heartbeat");	// reset timer for next one
}


}  // class Leader
