//  Message.java -- a structured form of a datagram.


//  IMPORTANT: add a case to the Message(p) constructor
//  for every new subclass created.


package topovista.toksvc;

import java.net.*;



/**
 *  A Message represents a datagram with an eight-byte
 *  header giving type, msig, from, and to values.  Each subclass
 *  of Message knows how to encode and decode its values to/from
 *  a byte array.  Subclasses and superclasses cooperate to encode
 *  a complete message.
 */
abstract class Message {

    /**
     *  An uninterpreted value for distinguishing classes of messages.
     */
    int type;

    /**
     *  A signature of the member list, for ensuring consistency
     *  between sender and receiver.
     */
    int msig;

    /**
     *  The sender, as an index into the member list.
     */
    int from;

    /**
     *  The receiver, or -1 for a broadcast (actually multicast).
     */
    int to;

    /**
     *  The original datagram, if any, from which this message came.
     */
    DatagramPacket pkt;



/**
 *  Defines the maximum expected message length.
 */
final static int MAX_LENGTH = 8192;



/**
 *  Constructs a new message.
 *  The specified type should be that defined for the particular subclass.
 *  Other fields are filled in later, when the message is sent.
 */
Message(int type) {
    this.type = type;
}



/**
 *  Returns a string representation of this message.
 */
public String toString() {
    return Util.classname(this) +
	"(" + type + "," + from + "," + to + "," + msig + ")";
}



/**
 *  Adds the encoding of this object to a Pickle.
 *  Overriding subclasses should call this method before
 *  adding their own data.
 */
Pickle pickle(Pickle p) {
    p.put(type);
    p.put(from);
    p.put(to);
    p.put(msig);
    return p;
}



/**
 *  Constructs a Message by reading a Pickle.
 *  Subclass constructors that read a Pickle should call this first.
 */
Message(Pickle p) {
    type = p.getInt();
    from = p.getInt();
    to = p.getInt();
    msig = p.getInt();
}



/**
 *  Creates a Message of the appropriate class by reading a datagram,
 *  to which a reference is saved in the resulting message.
 */
static Message unpickle(DatagramPacket d) {
    Message m = unpickle(new Pickle(d));
    m.pkt = d;
    return m;
}



/**
 *  Creates a Message of the appropriate class by reading from a Pickle.
 */
static Message unpickle(Pickle p) {
    int type = p.getInt();
    p.reset();
    switch (type) {
	case AliveMessage.MTYPE:	return new AliveMessage(p);
	case CatalogMessage.MTYPE:	return new CatalogMessage(p);
	case ConfigMessage.MTYPE:	return new ConfigMessage(p);
	case ConfirmMessage.MTYPE:	return new ConfirmMessage(p);
	case DataMessage.MTYPE:		return new DataMessage(p);
	case ElectMessage.MTYPE:	return new ElectMessage(p);
	case GrantMessage.MTYPE:	return new GrantMessage(p);
	case LeadMessage.MTYPE:		return new LeadMessage(p);
	case LoginMessage.MTYPE:	return new LoginMessage(p);
	case LogoutMessage.MTYPE:	return new LogoutMessage(p);
	case RequestMessage.MTYPE:	return new RequestMessage(p);
	case ReturnMessage.MTYPE:	return new ReturnMessage(p);
	case RevokeMessage.MTYPE:	return new RevokeMessage(p);
	case RosterMessage.MTYPE:	return new RosterMessage(p);
	case ServerMessage.MTYPE:	return new ServerMessage(p);
	case TransferMessage.MTYPE:	return new TransferMessage(p);
	case VoteMessage.MTYPE:		return new VoteMessage(p);
	default:			return new UnknownMessage(type);
    }
}



}  // class Message
