//  Observer.java -- monitors passing messages


package topovista.toksvc;

import java.io.*;
import java.util.*;



/**
 *  A Observer monitors incoming messages,
 *  writing a log but not participating actively in any way.
 */

class Observer {

    private PrintWriter out;			// log file
    private Council council;			// council being monitored

    private BitSet ignored = new BitSet();	// message types to suppress

    private long tbase = System.currentTimeMillis();
    private long tprev = tbase;

    private String prevroster = "";		// previous roster printed



/**
 *  Constructor.
 */
Observer(PrintWriter logfile, Council c) {

    out = logfile;
    council = c;

    int n = c.getIndex();
    out.println("Observer(" + c + "):");

    String[] members = c.getMembers();
    for (int i = 0; i < members.length; i++) {
	out.print((i == n) ? " >> " : "    ");
	out.println(i + ". " + members[i]);
    }
    out.println();
    out.flush();

    c.listen(new Receiver() {
	public void incoming(Message m) {
	    report(m);
	}
    });
}



/**
 *  Suppresses logging of one type of message.
 *  Initially, all types are logged.
 */
void suppress(int msgtype) {
    ignored.set(msgtype);
}



/**
 *  Allows logging of one type of message,
 *  reversing the effects of a suppress call.
 */
void allow(int msgtype) {
    ignored.clear(msgtype);
}



/**
 *  Reports a passage message, if its type is enabled.
 */
private synchronized void report(Message m) {
    if (wanted(m)) {
	long t = System.currentTimeMillis();
	out.println((t - tbase) + "\t" + (t - tprev) + "\t" + m);
	out.flush();
	tprev = t;
    }
}



/**
 *  Checks whether a particular message is to be traced.
 *  A RosterMessage that differs from its predecessor is always traced.
 */
private boolean wanted(Message m) {
    if (m.type < 0) {
	return true;
    }
    if (m.type == RosterMessage.MTYPE) {
	String s = m.toString();
	s = s.substring(s.lastIndexOf("(") + 1);
	if (! s.equals(prevroster)) {
	    prevroster = s;
	    return true;
	}
    }
    return !ignored.get(m.type);
}



}  // class Observer
