//  Pickle.java -- encode/decode values for serial transmission


package topovista.toksvc;

import java.io.*;
import java.net.*;



/**
 *  The Pickle class supports encoding and decoding of values
 *  for serial transmission.  We do our own pickling for interoperability
 *  with C.
 */

public class Pickle {

    private ByteArrayInputStream istream;
    private ByteArrayOutputStream ostream;

    int nread = 0;
    int nwritten = 0;



/**
 *  Returns a string representation of this pickle.
 */
public String toString() {
    if (istream != null) {
	return "Pickle(" + nread + "/" + nwritten + ")";
    } else {
	return "Pickle(" + nwritten + ")";
    }
}



/**
 *  Constructs an empty Pickle to which values can be added.
 */
public Pickle() {
    ostream = new ByteArrayOutputStream();
}



/**
 *  Gets the byte string representing a completed pickle.
 */
public byte[] toByteArray() {
    return ostream.toByteArray();
}



/**
 *  Adds one byte to the output stream.
 */
private void addByte(int b) {
    ostream.write((byte) b);
    nwritten++;
}



/**
 *  Adds an integral value (char, long, etc.) to this Pickle.
 *  The encoding is:
 *  <BR> 0snn nnnn			7-bit signed value (in one byte)
 *  <BR> 1bbb snnn  nnnn nnnn  ...	12 + 8 * bbb bits (in (bbb + 2) bytes)
 */
public Pickle put(long v) {

    if (-0x40 <= v && v < 0x40) {
	addByte((byte) v & 0x7F);
    } else {
	int b = 0;
	for (long x = v >> 11; x != 0 && x != -1; x >>= 8) {
	    b++;
	}
	if (b == 7) {
	    addByte((v < 0) ? 0xFF : 0xF0);
	} else {
	    addByte(0x80 + (b << 4) + (byte) ((v >> (8 * (b + 1))) & 0xF));
	}
	while (b >= 0) {
	    addByte((byte) (v >> (8 * b--)));
	}
    }
    return this;
}



/**
 *  Adds a String value to this Pickle.
 */
public Pickle put(String s) {
    put(s.getBytes());
    return this;
}



/**
 *  Adds an array of bytes to this Pickle.
 */
public Pickle put(byte[] b) {
    put(b.length);
    ostream.write(b, 0, b.length);
    nwritten += b.length;
    return this;
}



/**
 *  Constructs a Pickle from which values can be obtained.
 */
public Pickle(byte[] b) {
    istream = new ByteArrayInputStream(b);
    nwritten = b.length;
}



/**
 *  Constructs a Pickle that reads values from a portion of a buffer.
 */
public Pickle(byte[] b, int offset, int len) {
    istream = new ByteArrayInputStream(b, offset, len);
    nwritten = len;
}



/**
 *  Constructs a Pickle that reads values from a datagram.
 */
public Pickle(DatagramPacket d) {
    this(d.getData(), 0, d.getLength());
}



/**
 *  Rewinds the buffer allowing this Pickle to be reread from the beginning.
 */
public Pickle reset() {
    istream.reset();
    nread = 0;
    return this;
}



/**
 *  Reads the next byte from this pickle.
 *  Throws java.util.NoSuchElementException if data is exhausted.
 */
private int nextByte() {
    int b = istream.read();
    if (b < 0) {
	throw new java.util.NoSuchElementException(this.toString());
    }
    nread++;
    return b;
}



/**
 *  Reads an integer from this Pickle.
 *  Throws java.util.NoSuchElementException if data is exhausted.
 */
public int getInt() {
    return (int) getLong();
}



/**
 *  Reads a long value from this Pickle.
 *  Throws java.util.NoSuchElementException if data is exhausted.
 */
public long getLong() {
    long v;
    int s;

    int b = nextByte();		// get first byte
    if (b < 0) {
	throw new java.util.NoSuchElementException(this.toString());
    }
    if (b < 0x40) {
	return b;			// 6-bit positive value
    }
    if (b < 0x80) {
	return b - 0x80;		// 6-bit negative value
    }

    int n = (b >> 4) & 0x7;		// get bytecount-1
    v = (- (b & 8)) | (b & 7);		// get extended sign and 3 data bits

    while (n-- >= 0) {			// while more bits
	v = (v << 8) | nextByte();	// squeeze them in
    }

    return v;
}



/**
 *  Reads a String from this Pickle.
 *  Throws java.util.NoSuchElementException if data is exhausted.
 */
public String getString() {
    return new String(getByteArray());
}



/**
 *  Reads a byte array from this Pickle.
 *  Throws java.util.NoSuchElementException if data is exhausted.
 */
public byte[] getByteArray() {
    int n = getInt();
    byte[] b = new byte[n];
    int k;
    if (n == 0) {
	return b;
    } else if ((k = istream.read(b, 0, n)) == n) {
	nread += n;
	return b;
    } else {
	throw new java.util.NoSuchElementException(this.toString());
    }
}



}  // class Pickle
