//  Server.java -- all the non-leader token server functions.



package topovista.toksvc;

import java.io.*;
import java.util.*;



/**
 *  Server contains the logic of a token server,
 *  with the exception of code used in the role of a Leader.
 */

public class Server {



    private Council council;		// the council we belong to
    private int myid;			// our member index
    private int nservers;		// number or members

    private RosterMessage roster;	// current host roster

    private int mystate;		// state of this server
    private Vault vault;		// database of all tokens we hold
    private Certifier cert;		// info about token authority

    private Timer hbtimer;		// heartbeat timer



// token server states

final static byte DOWN = 0;	// not functioning
final static byte BOOTING = 1;	// initializing
final static byte READY = 2;	// fully functional



/**
 *  Constructor.
 */
Server(Council c) {
    council = c;			// save parent council
    myid = c.getIndex();		// save our index
    nservers = c.getMembers().length;	// save count of members
    restart();				// initialize data structures

    hbtimer = new Timer() {		// create heartbeat timer
	public void alarm(Timer tm, Object val) {
	    heartbeat();
	}
    };
    hbtimer.set(Leader.HBT, null);

    c.listen(new Receiver() {		// register message handler
	public void incoming(Message m) {
	    receive(m);
	}
    });
}



/**
 *  Initialize or reinitialize the server data structures.
 */
private void restart() {
    mystate = BOOTING;
    vault = new Vault();
    cert = new Certifier(council, myid);
    roster = new RosterMessage(new byte[nservers], new SessionList());
}



/**
 *  Called for every incoming message.
 */
private void receive(Message m) {
    switch (m.type) {
	case CatalogMessage.MTYPE:   rcvCatalog((CatalogMessage) m);    break;
	case TransferMessage.MTYPE:  rcvTransfer((TransferMessage) m);  break;
	case RosterMessage.MTYPE:    rcvRoster((RosterMessage) m);      break;
	case RequestMessage.MTYPE:   rcvRequest((RequestMessage) m);	break;
	case ReturnMessage.MTYPE:    rcvReturn((ReturnMessage) m);	break;
    }
}



/**
 *  Process an incoming RosterMessage, which reports the states
 *  of all servers and the current client list.
 */
private void rcvRoster(RosterMessage m) {

    if (m.term < roster.term || m.msgnum <= roster.msgnum) {
	return;			// if old message, just ignore it
    }

    if (m.sstates[myid] == DOWN && roster.sstates[myid] != DOWN) {
	restart();		// if server has declared us down, start over;
	return;			// do no further processing of this message
    }

    RosterMessage prev = roster;
    roster = m;				// save new roster
    cert.newRoster(m);			// register with Certifier

    for (int i = 0; i < nservers; i++) { // check state of every server
	if (i == myid) {		// except ourself
	    continue;
	}
	if (roster.sstates[i] == BOOTING) {
	    council.send(vault.transferTo(i, roster.sstates), i);
	} else if (roster.sstates[i] == READY && prev.sstates[i] == BOOTING) {
	    vault.forget(cert);		// time to clean out old tokens
	}
    }

    // once the roster shows this server as BOOTING or READY,
    // request token reports from any clients we need to hear from
    if (roster.sstates[myid] != DOWN) {
	SessionList sl = cert.awaiting();
	for (Enumeration e = sl.elements(); e.hasMoreElements(); ) {
	    Session s = (Session) e.nextElement();
	    council.send(new ConfigMessage(m.from, m.sstates), s);
	}
    }
}



/**
 *  Sends a state message to the current Leader.
 */
private void heartbeat() {
    checkReady();		// check to see if we've just come ready
    council.send(new ServerMessage(mystate), roster.from);
    hbtimer.set(Leader.HBT, null);
}



/**
 *  Process an incoming CatalogMessage, which tells the server about
 *  the state of one client's holdings.
 */
private void rcvCatalog(CatalogMessage m) {
    Session s = roster.slist.get(m.from);
    if (s != null) {
	vault.register(m, s);
	cert.heardFrom(s);			// notify guard
    }
}



/**
 *  Process an incoming TransferMessage, which passes a set of tokens
 *  from a backup server to a (more) primary server.
 */
private void rcvTransfer(TransferMessage m) {
    if (roster.sstates[m.from] != Server.DOWN) { // if sender is still ok
	vault.transferIn(m);			// accept tokens
	cert.heardFrom(m.from);			// notify guard
	checkReady();				// check if this makes us ready
    }
}



/**
 *  Check to see if it's time to declare this server READY,
 *  and if so, updates mystate.
 */
private void checkReady() {
    if (roster.sstates[myid] == BOOTING && cert.ready()) {
	mystate = READY;		// ready to go!
    }
}



/**
 *  Called for an incoming RequestMessage.
 */
private void rcvRequest(RequestMessage m) {
    ServerToken t = vault.lookup(m.token.name);
    Session s = getSender(m.from, t);
    if (s == null) {
	return;			// client died, not our token, or not ready
    }

    // attempt the assignment
    t.cksess(roster.slist);
    if (t.assign(s, m.access)) {
	t.pending = null;
	council.send(new GrantMessage(m.msgnum, t), s);	// report success
	return;
    }

    // save pending request in token, displacing older pending request
    // (fancier algorithm here could produce other policies)
    t.pending = m;

    // send Revoke messages to current holders
    Vector v = t.getAssignments();
    synchronized (t) {
	for (int i = 0; i < v.size(); i++) {
	    Session h = (Session) v.elementAt(i);
	    council.send(new RevokeMessage(m.token.name), h);
	}
    }
}



/**
 *  Called for an incoming ReturnMessage.
 */
private void rcvReturn(ReturnMessage m) {
    ServerToken t = vault.lookup(m.token.name);
    Session s = getSender(m.from, t);
    if (s == null) {
	return;			// client died, not our token, or not ready
    }

    if ((m.flags & ReturnMessage.UPDATE) != 0) {
	t.setData(m.token.getData());		// update token data
    }
    if ((m.flags & ReturnMessage.DEASSIGN) != 0) {
	t.deassign(s);				// deassign token
    }
    council.send(new ConfirmMessage(m.msgnum, m.token.name), s);
						// ack update or deassignment

    if (t.getState() == 0 && t.pending != null) {
	rcvRequest(t.pending);
    }
}



/**
 *  Gets the sender's Session entry for a message regarding a token.
 *  Returns null if client has died or if the message is misdirected.
 *  (And if misdirected, sends a ConfigMessage to correct the client.)
 */
private Session getSender(int from, ServerToken t) {
    Session s = roster.slist.get(from);		// look up session
    if (s == null) {				// if no longer registered
	return null;
    }
    switch (cert.status(t)) {
	case Certifier.READY:
	    return s;
	case Certifier.FOREIGN:
	    council.send(new ConfigMessage(roster.from, roster.sstates), s);
	    return null;
	default:
	    return null;
    }
}



/**
 *  Runs a Server process.  The process runs until killed.
 *
 *  <P> Usage:  java Server [-flags] config-file server-index
 *
 *  <P> flags:  <UL>
 *	<LI> -c: trace client heartbeats
 *	<LI> -s: trace server heartbeats
 *	<LI> -l: trace leader heartbeats
 *	<LI> -m: trace all other incoming messages
 *	<LI> -a: trace all incoming messages
 *	<LI> -1: drop 10% of incoming messages (and so on through...)
 *	<LI> -9: drop 90% of incoming messages
 *  </UL>
 */
public static void main(String args[]) {

    String flags = "";
    int i = 0;
    if (args.length > 0 && args[i].charAt(0) == '-') {
	flags = args[i++];
    }

    if (args.length - i != 2) {
	usage();
    }

    Configuration cf = null;
    try {
	cf = new Configuration(args[i]);
    } catch (Exception e) {
	System.err.println("Cannot configure: " + e);
	System.exit(1);
    }

    String[] servers = cf.getArray("server");
    int n = -1;
    try {
	n = Integer.parseInt(args[i+1]);	// parse server index
	String s = servers[n];			// validate it
    } catch (Throwable e) {
	System.err.println("Invalid server index: " + args[i+1]);
	System.exit(1);
    }

    Council c = null;
    try {
	c = new Council(servers, n);
	Director tl = new Director(c);
	Server ts = new Server(c);
	if (flags.length() > 1) {
	    Observer o = new Observer(new PrintWriter(System.out), c);
	    if (! Util.upto("ca", flags)) { o.suppress(AliveMessage.MTYPE); }
	    if (! Util.upto("sa", flags)) { o.suppress(ServerMessage.MTYPE); }
	    if (! Util.upto("la", flags)) { o.suppress(RosterMessage.MTYPE); }
	}
    } catch (Throwable e) {
	if (c != null) {
	    c.abort("Server exception", e);
	} else {
	    System.err.println("Council initialization error: " + e);
	    System.exit(1);
	}
    }

    for (i = 0; i < flags.length(); i++) {
	char ch = flags.charAt(i);
	if (ch >= '1' && ch <= '9') {
	    c.setReliability(1.0 - 0.1 * (ch - '0'));
	}
    }

    for (;;) {			// sleep forever while Server runs
	try {
	    Thread.sleep(60 * 1000);
	} catch (InterruptedException e) {
	}
    }
}



/**
 *  Diagnoses an error in invoking the server as a program.
 */
private static void usage() {
    System.err.println(
	"usage: java Server [-flags] config-file server-index");
    System.exit(1);
}



}  // class Server
