//  Session.java -- representing an active client.


package topovista.toksvc;

import java.io.*;
import java.net.*;
import java.util.*;



/**
 *  A Session in a server represents an active client handle.
 *  Sessions are created by the token leader.
 */
class Session {

    private int sessID;		// unique ID
    private String address;	// input address string ("nn.nn.nn.nn:port")
    private InetAddress ipaddr;	// corresponding InetAddress
    private int portnum;	// and integer port number

private static DatagramSocket sendport;	// socket for outgoing messages
private static int maxsid = 0;		// maximum session ID ever seen



/**
 *  Constructs a Session with specified data.
 */
Session(int sessID, String address) {
    this.sessID = sessID;
    this.address = address;
    if (sessID > maxsid) {
	maxsid = sessID;
    }
    try {
	ipaddr = InetAddress.getByName(Util.hostPart(address));
    } catch (UnknownHostException e) {
	Util.abort("unknown login", e);
    }
    portnum = Util.portNumber(address);
}



/**
 *  Constructs a Session by reading a Pickle.
 */
Session(Pickle p) {
    this(p.getInt(), p.getString());
}



/**
 *  Adds the encoding of this Session to a Pickle.
 */
Pickle pickle(Pickle p) {
    p.put(sessID);
    p.put(address);
    return p;
}



/**
 *  Returns the integer ID of this Session.
 */
int getID() {
    return sessID;
}



/**
 *  Returns the largest ID ever seen in any Session.
 */
static int getMaxID() {
    return maxsid;
}


/**
 *  Sends a message to the client represented by this Session.
 *  I/O exceptions are ignored.
 *  It is assumed that the "from", "to", and "msig" fields have been set;
 *  this is usually done by calling council.send(Message), which
 *  fills in these fields and then calls this method.
 */
void send(Message m) {
    byte[] buf = m.pickle(new Pickle()).toByteArray();
    DatagramPacket p = new DatagramPacket(buf, buf.length, ipaddr, portnum);
    try {
	if (sendport == null) {
	    sendport = new DatagramSocket();
	}
	sendport.send(p);
    } catch (IOException e) {
    }
}



/**
 *  Returns a hashcode for this Session.
 */
public int hashCode() {
    return address.hashCode() ^ (37 * sessID);
}



/**
 *  Compares this Session to the specified object.
 *  Two Sessions are equal if they have the same ID and address.
 */
public boolean equals(Object o) {
    if (o == null || ! (o instanceof Session)) {
	return false;
    }
    Session s = (Session) o;
    return s.sessID == this.sessID && s.address.equals(this.address);
}



/**
 *  Returns a string representation of this message.
 */
public String toString() {
    return "Session(" + sessID + "," + address + ")";
}



}  // class Session
