//  Stress.java -- a stress tester


package topovista.toksvc;

import topovista.util.*;



/**
 *  Implements a command-line program for stress-testing the token system.
 */
public class Stress {

    Thread thread;	// process thread
    String tlabel;	// thread label

    Client client;	// client handle
    Revoker revoker;	// general-purpose unconditional revoker

    int nloops;		// loop counter

    Token shared;	// handle for token named "shared"
    Token excl;		// handle for token named "excl"



static Stress[] testers;		// array of testers (Stress instances)

static String options = "";		// string of option characters
static String enabled = "";		// enabled tests
static int nthreads = 5;		// number of threads



/**
 *  Stress-tests the token system.
 *  It runs multiple clients, each of which
 *  runs multiple testing processes.
 *
 *  <P> Usage: java Stress [-options] configfile
 *
 *  <P> The <VAR>options</VAR> argument is a string of single-character flags.
 *
 *  <P> Options that are digits (1 through 9) specify the number of
 *  concurrent testing threads.  The default is 5.
 *
 *  <P> The option character <STRONG>t</STRONG> enables tracing output.
 *  Other alphabetic characters enable particular tests, which
 *  repeat until the program is halted:
 *  <UL>
 *  <LI> <STRONG>b</STRONG>:
 *	request/release token <STRONG>busy</STRONG> as fast as possible
 *  <LI> <STRONG>i</STRONG>:
 *	request/increment/release token <STRONG>incr</STRONG>
 *  <LI> <STRONG>s</STRONG>:
 *	update token <STRONG>shared</STRONG> with process label and loop counter
 *  <LI> <STRONG>u</STRONG>:
 *	request/release new, unique tokens
 *  <LI> <STRONG>w</STRONG>:
 *	request/release token <STRONG>who</STRONG>, updating with process label
 *  <LI> <STRONG>x</STRONG>:
 *	request token <STRONG>excl</STRONG> and hold until revoked
 *  </UL>
 *  If no tests are specified, all are enabled.
 */
public static void main(String args[]) throws Exception {

    if (args.length < 1 || args.length > 2 ||
	    (args.length == 2 && args[0].charAt(0) != '-')) {
	Util.abort("usage: java Stress [-options] configfile");
    }

    Configuration cfg;
    if (args.length == 2) {
	setoptions(args[0]);
	cfg = new Configuration(args[1]);
    } else {
	cfg = new Configuration(args[0]);
    }

    testers = new Stress[nthreads];
    for (int i = 0; i < nthreads; i++) {
	testers[i] = new Stress(cfg, "ABCDEFGHI".substring(i, i + 1));
    }
    for (int i = 0; i < nthreads; i++) {
	testers[i].thread.start();
    }
}



//  handle option string

static void setoptions(String s) {
    options = s;
    enabled = s;
    for (int i = 0; i < s.length(); i++) {
	char c = s.charAt(i);
	if (c >= '1' && c <= '9') {
	    nthreads = c - '0';
	}
    }
}



//   construct one of several parallel testers, spawning a test process

Stress(Configuration cfg, final String lbl) throws Exception {

    tlabel = lbl;
    client = new Client(cfg);
    if (Util.upto("t", options)) {
	client.setTrace(true);
    }

    revoker = new Revoker() {
	public void revoke(Token t, Object o) {
	    t.release();
	}
    };

    shared = client.request("shared", Token.SHARED, null, null);
    excl = client.request("excl", Token.EXCLUSIVE, revoker, null);
    excl.release();

    thread = new Thread() {
	public void run() {
	    try {
		for (;;) {
		    nloops++;
		    something();
		    Thread.yield();
		}
	    } catch (Throwable t) {
		Util.abort(t);
	    }
	}
    };
}



//  do something, choosing one of several actions at random

void something() {

    String e = enabled;		// private copy avoids concurrency problems
    int n = e.length();
    if (n == 0) {
	e = enabled = "abcdefghijklmnopqrstuvwxyz";
	n = e.length();
    }
    int i = (int) (n * Math.random());
    char c = e.charAt(i);

    String s;
    Token t;
    switch (c) {

	case 'b':
	    // request and release the token named "busy"
	    grab("busy").release();
	    break;

	case 'i':
	    // request and increment a token named "incr"
	    t = grab("incr");
	    s = t.getString();
	    if (s.length() == 0) {
		s = "0";
	    }
	    t.release(Integer.toString(Integer.parseInt(s) + 1));
	    break;

	case 's':
	    // store our label and loop counter in the shared token "shared"
	    shared.update(tlabel + nloops);
	    break;

	case 'u':
	    // request and release a new, unique token
	    grab("u" + nloops + tlabel).release();
	    break;

	case 'w':
	    // request and release the token named "who",
	    // but first update with our process label if not already so marked
	    t = grab("who");
	    if (! t.getString().equals(tlabel)) {
		t.update(tlabel);
		Thread.yield();
	    }
	    t.release();
	    break;

	case 'x':
	    // request the token "excl" and hold until revoked
	    synchronized (excl) {
		if (excl.isAssigned()) {
		    excl.release();		// already held by self
		}
		excl = client.request("excl", Token.EXCLUSIVE, revoker, null);
	    }
	    break;

	default:
	    // selected a nonexistent test; remove from choices
	    i = e.indexOf(c);
	    if (i >= 0) {
		enabled = e.substring(0, i) + e.substring(i + 1);
	    }
    }
}



//  grab a token unconditionally and exclusively, with no revoker code

Token grab(String name) {
    return client.request(name, Token.EXCLUSIVE, null, null);
}



} // class Stress
