//  Token.java -- a token used by client programs



package topovista.toksvc;

import java.util.*;



/**
 *  A Token represents an assigned token obtained by calling
 *  Client.request().  There is no public constructor.
 */

public class Token extends Tk {

    private Client client;	// the associated client

    private Revoker revoker;	// client callback on revocation message
    private Object arg;		// client callback argument

    private int state;		// current state of this token
    private boolean dirty;	// is data newer than server knows?

    private long waitnum;	// message number awaiting acknowledgment

    boolean requested;		// client in-use flag



/**
 *  Designates shared access mode for a token.
 */
public final static int SHARED = +1;



/**
 *  Designates exclusive access mode for a token.
 */
public final static int EXCLUSIVE = -1;



/**
 *  Token states stored in "state".
 *  States advance in sequence, and also a few other ways, e.g.:
 *	from HELD to LEAVING on receipt of RevokeMessage
 *	from HELD to RELEASED by release() call
 *	from REVISED to HELD on receiving confirmation of update
 */
private static final int FREE = 0;	// initial state
private static final int REQUESTED = 1;	// requested from server
private static final int RECEIVED = 2;	// GrantMessage received
private static final int HELD = 3;	// requesting thread awakened
private static final int REVISED = 4;	// updated, not yet confirmed
private static final int LEAVING = 5;	// RevokeMessage received
private static final int RELEASED = 6;	// released, not yet confirmed



/**
 *  Constructs a client with specified data.
 */
Token(Client c, String name, Revoker r, Object arg) {
    super(name, null);
    client = c;
    revoker = r;
    this.arg = arg;
}



/**
 *  Constructs a clientless Token by reading a Pickle.
 */
Token(Pickle p) {
    super(p);
}



/**
 *  Constructs a clientless Token for testing purposes.
 */
Token(String name, byte[] data) {
    super(name, data);
}



/**
 *  Requests this Token from the appropriate server.
 */
public synchronized void request(int access) {

    while (state == RELEASED) {
	try {
	    this.wait();
	} catch (InterruptedException e) {
	}
    }
    if (state != FREE) {
	throw new Error("duplicate request for token: " + this.name);
    }

    this.data = null;
    seqnum();
    client.asend(new RequestMessage(waitnum, this, access));
    state = REQUESTED;
    while (state != RECEIVED) {
	try {
	    this.wait();
	} catch (InterruptedException e) {
	}
    }
    state = HELD;
}



/**
 *  Returns true if this Token is currently assigned to the holder.
 */
public boolean isAssigned() {
    switch (state) {
	case RECEIVED:
	case HELD:
	case REVISED:
	case LEAVING:
	    return true;
	default:
	    return false;
    }
}



/**
 *  Processes a GrantMessage that concerns this token.
 */
synchronized void grant(GrantMessage m) {

    if (state == REQUESTED && m.msgnum == waitnum) {
	this.setData(m.token.getData());
	dirty = false;
	state = RECEIVED;
	this.notifyAll();
    }
}



/**
 *  Updates the Token data on the server without releasing the Token.
 *  (If a revoke is pending, though, the update awaits the expected release.)
 */
public synchronized void update(byte[] data) {

    this.setData(data);
    dirty = true;

    switch (state) {

	case REVISED:
	    // an earlier update has not yet been acknowledged, so wait for it;
	    // this keeps changes in order and prevents flooding the server
	    while (state == REVISED) {
		try {
		    this.wait();
		} catch (InterruptedException e) {
		}
	    }
	    if (state != HELD) {	// if revoked or released, do no more
		return;
	    }
	    dirty = true;		// may have been reset by CONFIRM
	    break;

	case HELD:
	    break;

	case LEAVING:
	    // have already received REVOKE for this token, so do nothing --
	    // just hold data pending expected release() call
	    return;

	default:
	    throw new Error("attempt to update released token: " + this.name);
    }

    // send RETURN message to update the data on the server
    seqnum();
    client.asend(new ReturnMessage(waitnum, this, ReturnMessage.UPDATE));
    state = REVISED;
}



/**
 *  Updates the Token data on the server without releasing the Token.
 */
public void update(String s) {
    update(s.getBytes());
}



/**
 *  Releases this Token, updating its data attribute.
 */
public synchronized void release(byte[] data) {

    this.setData(data);
    dirty = true;
    release();
}



/**
 *  Releases this Token, updating its data attribute.
 */
public void release(String s) {
    release(s.getBytes());
}



/**
 *  Releases this Token without updating its data.
 */
public synchronized void release() {

    switch(state) {
	case HELD:
	case REVISED:
	case LEAVING:
	    int f = ReturnMessage.DEASSIGN;
	    if (dirty) {
		f += ReturnMessage.UPDATE;
	    } else {
		this.setData(empty);
	    }
	    seqnum();
	    client.asend(new ReturnMessage(waitnum, this, f));
	    state = RELEASED;
	    this.notifyAll();
	    break;

	default:
	    throw new Error("attempt to release token not held: " + this.name);
    }
}

private static byte[] empty = {};



/**
 *  Processes a RevokeMessage that concerns this token.
 *  The first such message invokes the user callback, if not null.
 */
void revoke(RevokeMessage m) {
    synchronized (this) {
	if (state != HELD && state != REVISED) {
	    return;			// nothing to do
	}
	state = LEAVING;
	this.notifyAll();
    }
    // now call revoker without holding Token monitor
    try {
	if (revoker != null) {
	    revoker.revoke(this, arg);
	}
    } catch (Throwable t) {
	System.err.println("Token callback error: " + name);
	t.printStackTrace();
	System.exit(1);
    }
}



/**
 *  Processes a ConfirmMessage that concerns this token.
 */
synchronized void confirm(ConfirmMessage m) {
    if (m.msgnum != waitnum) {
	return;
    }
    switch (state) {
	case REVISED:
	    state = HELD;
	    dirty = false;
	    this.notifyAll();
	    break;
	case LEAVING:
	    // no change in status
	    // release remains pending
	    break;
	case RELEASED:
	    state = FREE;
	    this.notifyAll();
	    client.forget(this);
	    break;
	default:
	    // ignore unexpected message
	    break;
    }
}



/**
 *  Allocates, stores, and returns the next sequence number from a global pool.
 */
private synchronized long seqnum() {
    synchronized (this.getClass()) {
	waitnum = nextsn++;
    }
    return waitnum;
}

private static long nextsn = 1;	// next sequence number to be assigned



}  // class Token
