//  Util.java -- miscellaneous utility methods


package topovista.toksvc;


/**
 *  Util contains a collection of small static utility methods.
 */

public class Util {



/**
 *  Returns the host portion of a host:port string.
 */
public static String hostPart(String s) {
    int i = s.lastIndexOf(':');
    if (i >= 0) {
	return s.substring(0, i);
    } else {
	return s;
    }
}



/**
 *  Returns the port number from a host:port string, or -1 if
 *  invalid or missing.
 */
public static int portNumber(String s) {
    int i = s.lastIndexOf(':');
    if ((i < 0) || (i == s.length() - 1)) {
	return -1;
    } else try {
	return Integer.parseInt(s.substring(i + 1));
    } catch (Exception e) {
	return -1;
    }
}



/**
 *  Returns the simplified name of the class of an object,
 *  with everything before the last "." stripped off.
 */
public static String classname(Object o) {
    if (o == null) {
	return "null";
    } else {
	String cname = o.getClass().getName();
	return cname.substring(1 + cname.lastIndexOf('.'));
    }
}



/**
 *  Returns a printable image of a string value.
 *  Unprintable characters are replaced by '~' characters.
 *  The maximum returned length is maxlen; ellipses are used if necessary.
 */
public static String image(String s, int maxlen) {
    StringBuffer b = new StringBuffer(maxlen);
    int len = s.length();
    int n = (len > maxlen) ? (maxlen - 3) : len;
    for (int i = 0; i < n; i++) {
	char c = s.charAt(i);
	b.append((c < ' ' || c > '~') ? '~' : c);
    }
    if (len > maxlen) {
	while (b.length() < maxlen) {
	    b.append('.');
	}
    }
    return b.toString();
}



/**
 *  Returns true if any character of c is contained in s.
 *  (Inspired by the "upto" function of the Icon programming language.)
 */
public static boolean upto(String c, String s) {
    for (int i = 0; i < c.length(); i++) {
	if (s.indexOf(c.charAt(i)) >= 0) {
	    return true;
	}
    }
    return false;
}



/**
 *  Returns a non-negative hash code for a string, avoiding
 *  java.lang.String.hashcode() which differs across Java versions.
 */
public static int hash(String s) {
    int h = 0;
    for (int i = 0; i < s.length(); i++) {
	h = 37 * h + s.charAt(i);
    }
    return h & 0x7FFFFFFF;
}



/**
 *  Returns a non-negative hash code for a list of strings.
 */
public static int hash(String[] s) {
    int h = 0;
    for (int i = 0; i < s.length; i++) {
	h = 39 * h + hash(s[i]);
    }
    return h & 0x7FFFFFFF;
}



/**
 *  Returns a new non-negative hash code based on an existing hash code.
 *  Successive calls produce a pseudorandom sequence.
 */
public static int rehash(int n) {
    // Constants are per Knuth vol 2 [1969], pp. 155-157.
    return (314159261 * n + 453816707) & 0x7FFFFFFF;
}



/**
 *  Returns true if an object matches (via equals()) any entry in a list.
 */
public static boolean matchAny(Object a[], Object b) {
    if (a != null && b != null) {
	for (int i = 0; i < a.length; i++) {
	    if (b.equals(a[i])) {
		return true;
	    }
	}
    }
    return false;
}



/**
 *  Swaps two items in an array of shorts.
 */
public static void swap(short[] a, int i, int j) {
    short t = a[i];
    a[i] = a[j];
    a[j] = t;
}



/**
 *  Terminates the program abnormally, issuing an error message.
 */
public static void abort(String msg) {
    abort(msg, null);
}



/**
 *  Terminates the program abnormally, tracing the given exception.
 */
public static void abort(Throwable t) {
    abort(null, t);
}



/**
 *  Terminates the program abnormally, writing an error message
 *  and tracing the given exception.
 */
public static void abort(String msg, Throwable t) {
    if (msg == null && t != null) {
	msg = t.toString();
    }
    System.err.println("abort: " + msg);
    if (t != null) {
	t.printStackTrace();
    }
    System.exit(1);
}



}  // class Util
