package topovista.tvsucker;

import java.io.*;
import java.net.*;
import java.util.*;

import topovista.grid.*;



/**
 *  Fetches tiles via TVTP (TopoVista Tile Protocol) Version 1 over UDP.
 */

class TVTPReader extends Reader {

    private String hostname;		// server host name
    private int port;			// server port number
    private String path;		// path on server to tile directory

    private InetAddress ipa;		// server internet address
    private DatagramSocket udp;		// UDP socket for communication

    private Hashtable registry;		// registry of tiles seen


private static int DEFAULT_PORT = 4742;
private static int TVHEADER = 14;
private static int MAXRESPONSE = 8192;



/**
 *  Constructs a TVTPReader for fetching tiles from
 *  the directory of the TVTP pseudo-URL.
 */
public TVTPReader(String url) throws Exception {

    // parse the URL
    String hpa = url.substring(7);	// skip TVTP://
    int i = hpa.indexOf(':');
    int j = hpa.indexOf('/');
    if (j < 0) {
	throw new java.net.MalformedURLException(url);
    } if (i > 0 && i < j) {
	hostname = hpa.substring(0, i);
	port = Integer.parseInt(hpa.substring(i + 1, j));
    } else {
	hostname = hpa.substring(0, j);
	port = DEFAULT_PORT;
    }
    path = hpa.substring(j + 1);
    j = path.lastIndexOf('/');
    if (j >= 0) {
	path = path.substring(0, j);
    } else {
	path = "";
    }

    // construct a socket for sending and receiving data
    ipa = InetAddress.getByName(hostname);
    udp = new DatagramSocket();

    // initialize a registry for mapping records data to Tiles
    registry = new Hashtable();

    // spawn a background thread to receive and sort out input
    Thread t = new Thread() {
	public void run() {
	    try {
		listen();
	    } catch (Throwable t) {
		t.printStackTrace();
		System.exit(1);
	    }
	}
    };
    t.setDaemon(true);
    t.setPriority(Thread.currentThread().getPriority() + 1);
    t.start();
}



/**
 *  Listens for datagrams on the socket.
 */
private void listen() throws Exception {
    byte[] buf = new byte[MAXRESPONSE];
    for (;;) {
	DatagramPacket in = new DatagramPacket(buf, buf.length);
	udp.receive(in);
	int tid = ((buf[8] & 0xFF) << 24) | ((buf[9] & 0xFF) << 16) |
	    ((buf[10] & 0xFF) << 8) | (buf[11] & 0xFF);
	int[] vals = TileGrid.parseID(tid);
	int zone = vals[0];
	int xxx = vals[1];
	int yyy = vals[2];
	int k = vals[3];
	String s = TileGrid.makeName(zone, xxx, yyy, k);
	Tile t = (Tile) registry.get(s);
	if (t == null) {
	    t = new Tile(zone, xxx, yyy, k, ".???");
	    registry.put(s, t);
	    await(t);
	}
	if (buf[6] == 63) {		// if EOF
	    t.done();
	} else {
	    t.got(buf[TVHEADER + 4], in.getLength() - TVHEADER);
	}
    }
}



/**
 *  Spawns a process to wait for EOF on an unexpected tile, and then report it.
 */
private static void await(final Tile t) {
    TVSucker.announce("SURPRISE! receiving " + t);
    t.tResp = new Timestamp();
    new Thread() {
	public void run() {
	    synchronized(t) {
		while (t.tDone == null) {
		    try {
			t.wait();
		    } catch (InterruptedException unimportant) {
		    }
		}
	    }
	    TVSucker.report(t);
	}
    }.start();
}



/**
 *  Sends a TVTP request to load a tile, then waits for it.
 */
public void fetch(Tile t) throws Exception {
    String name = TileGrid.makeName(t.zone, t.xxx, t.yyy, t.k);
    int id = TileGrid.makeID(t.zone, t.xxx, t.yyy, t.k);
    String hexID = TileGrid.hexID(id);
    t.req();
    registry.put(name, t);
    byte[] req = ("GET " + hexID + " " + path + "/" + t.getName()).getBytes();
    udp.send(new DatagramPacket(req, req.length, ipa, port));
    t.resp();
    synchronized (t) {
	while (t.tDone == null) {
	    t.wait();
	}
    }
}



/**
 *  Returns a string representation of this reader.
 */
public String toString() {
    return "TVTPReader(" +  hostname + ":" + port + "/" + path + ")";
}



} // class TVFetch
