package topovista.util;

import java.io.*;

import topovista.util.*;



/**
 *  A TypedInputStream reads string, integer, floating, and null values
 *  produced by a TypedOutputStream.
 *
 *  @author Gregg Townsend, University of Arizona
 *  @version August, 2001
 */

public class TypedInputStream extends FilterInputStream {

    private long prevInt = 0;		// previous integral value seen



/**
 *  Constructs a TypedInputStream from an underlying InputStream.
 */
public TypedInputStream(InputStream in) {
    super(makeBuffered(in));
}



/**
 *  Makes an input stream buffered, if it is not already.
 */
private static InputStream makeBuffered(InputStream in) {
    if (in instanceof BufferedInputStream) {
	return in;
    } else {
	return new BufferedInputStream(in);
    }
}



/**
 *  Reads the next value, returning a String, Long, Double, or null
 *  depending on the type found.
 */
public Object readValue() throws IOException {
    int b = nextbyte();
    if (b < 0xE0) {
	return new Long(readLong(b));
    } else if (b < 0xF0) {
	return readString(b);
    } else if (b < 0xFB) {
	return new Double(readDouble(b));
    } else if (b == 0xFE) {
	return null;
    } else {
	throw new StreamCorruptedException
	    ("unrecognized byte 0x" + Integer.toHexString(b));
    }
}



/**
 *  Reads an integral value, returning an int.
 */
public int readInt() throws IOException {
    return (int) readLong(nextbyte());
}



/**
 *  Reads an integral value, returning a long.
 */
public long readLong() throws IOException {
    return readLong(nextbyte());
}



/**
 *  Reads an integral value, given the first byte.
 */
private long readLong(int b) throws IOException {

    if (b < 0x40) {
	return prevInt = b;		// 6-bit positive value
    }
    if (b < 0x80) {
	return prevInt = b - 0x80;	// 6-bit negative value
    }

    if (b < 0xC0) {
	long diff = -(b & 0x20) | (b & 0x1F);
	return prevInt += diff;		// 6-bit signed difference
    }

    if ((b & 0xF8) == 0xC0) {		// 11-bit signed value
	long v = ((- (b & 4)) | (b & 3)) << 8;
	return prevInt = v | nextbyte();
    }

    if ((b & 0xF8) == 0xC8) {		// 11-bit signed difference
	long v = ((- (b & 4)) | (b & 3)) << 8;
	return prevInt += v | nextbyte();
    }

    if ((b & 0xF0) != 0xD0) {
	throw new StreamCorruptedException("not an integer value");
    }
    if ((b & 8) == 0) {
	prevInt = 0;			// if value follows, not difference
    }

    int n = b & 7;			// get bytecount-1
    b = nextbyte();
    long v = (byte) b;			// sign extend first byte

    while (n-- > 0) {			// while more bits
	v = (v << 8) | nextbyte();	// squeeze them in
    }
    return prevInt += v;		// combine with previous value
}



/**
 *  Reads a floating-point value, returning a float.
 */
public float readFloat() throws IOException {
    return (float) readDouble(nextbyte());
}



/**
 *  Reads a floating-point value, returning a double.
 */
public double readDouble() throws IOException {
    return readDouble(nextbyte());
}


/**
 *  Reads an integral value, given the first byte.
 */
private double readDouble(int b) throws IOException {

    if (b == 0xF0) {
	return (double) readInt();
    } else if (b > 0xF0 && b <= 0xF7) {
	int nbytes = (b & 7) + 1;
	int shift = 56;
	long bits = 0;
	for (int i = 0; i < nbytes; i++) {
	    long c = nextbyte();
	    bits |= c << shift;
	    shift -= 8;
	}
	if (nbytes <= 4) {
	    return Float.intBitsToFloat((int) (bits >> 32));
	} else {
	    return Double.longBitsToDouble(bits);
	}

    } else if (b == 0xF8) {
	return 0.0;
    } else if (b == 0xF9) {
	return 1.0;
    } else if (b == 0xFA) {
	return -1.0;
    } else {					// not a floating value
	throw new StreamCorruptedException("not a floating value");
    }
}



/**
 *  Reads an encoded string or a null value.
 */
public String readString() throws IOException {
    return readString(nextbyte());
}



/**
 *  Reads an encoded string, or null, given the first byte.
 */
private String readString(int b) throws IOException {

    int len;				// saved string length
    int tsp;				// additional trailing spaces

    // crack the header byte

    if (b == 0xE0) {			// empty string
	return "";
    } else if (b > 0xE0 && b < 0xEF) {	// simple short string
	len = b & 0xF;
	tsp = 0;
    } else if (b == 0xEF) {		// arbitrary string
	tsp = this.readInt();
	if (tsp < 0) {			// if first int is <0, it counts spaces
	    tsp = -tsp;
	    len = this.readInt();
	} else {
	    len = tsp;			// positive or only int is the length
	    tsp = 0;
	}
    } else if (b == 0xFE) {		// null pointer
	return null;
    } else {				// not a string
	throw new StreamCorruptedException("not a string");
    }

    // read the characters of the string

    byte buf[] = new byte[len + tsp];		// allocate buffer
    int off = 0;
    while (off < len) {
	int r = read(buf, off, len - off);	// read characters
	if (r < 0) {
	    throw new EOFException();
	}
	off += r;
    }
    for (int i = 0; i < tsp; i++) {
	buf[len + i] = (byte) ' ';		// add trailing spaces
    }
    return new String(buf);
}



/**
 *  Reads the next byte, throwing an EOFException if there is none.
 */
private int nextbyte() throws IOException {
    int b = super.read();
    if (b < 0) {
	throw new EOFException();
    }
    return b;
}



/**
 *  Reads and decodes the stream from standard input,
 *  writing values to standard output.
 *  If a single argument "-x" is present, also shows hexadecimal
 *  encodings, except that strings are only partially dumped in hex.
 */

public static void main(String[] args) throws IOException {

    boolean wantHex = args.length > 0 && args[0].equals("-x");
    StringBuffer buf = new StringBuffer();
    String padding = "                             ";
    TypedInputStream ins =
	new TypedInputStream(new BufferedInputStream(System.in));

    try {
	for (;;) {
	    if (wantHex) {
		buf.setLength(0);
		ins.mark(10);
		ins.hexFormat(buf);
		ins.reset();
		buf.append(padding.substring(buf.length()));
		System.out.print(buf);
	    }
	    Object o = ins.readValue();
	    if (o == null) {
		System.out.println(o);
	    } else if (o instanceof String) {
		System.out.println("\"" + (String)o + "\"");
	    } else {
		System.out.println(o);
	    }
	}
    } catch (EOFException e) {
    }
}



/**
 *  Reads enough of the next value to format its header in hexadecimal.
 *  Leaves the input stream at an indeterminate position.
 */
private void hexFormat(StringBuffer buf) throws IOException {
    int b = nextbyte();
    hexAppend(buf, b);
    if (b < 0xC0 || b >= 0xF8) {	// if one-byte value
	// nothing more to do
    } else if (b < 0xD0) {		// if just one byte follows
	hexAppend(buf, nextbyte());
    } else if (b == 0xF0) {		// if to float a following int
	hexFormat(buf);
    } else if ((b & 0xF0) != 0xE0)  {	// if b+1 bytes follow
	b &= 7;
	while (b-- >= 0) {
	    hexAppend(buf, nextbyte());
	}
    } else if (b < 0xEF) {		// if one-byte string header
	buf.append("...");
	return;
    } else if (b == 0xEF) {		// if generalized string
	hexFormat(buf);			// show just first following int
	buf.append("...");
    } else {
	throw new Error("hexFormat: can't happen");
    }
}



/**
 *  Appends two hex characters of byte b, and a space, to a string buffer.
 */
private static void hexAppend(StringBuffer buf, int b) {
    buf.append("0123456789ABCDEF".charAt((b >> 4) & 0xF));
    buf.append("0123456789ABCDEF".charAt(b & 0xF));
    buf.append(' ');
}



} // class TypedInputStream
